Package: gDRimport
Type: Package
Title: Package for handling the import of dose-response data
Version: 1.11.4
Date: 2026-06-16
Authors@R: c(
    person("Arkadiusz", "Gladki", role=c("aut", "cre"), email="gladki.arkadiusz@gmail.com",
           comment = c(ORCID = "0000-0002-7059-6378")),
    person("Bartosz", "Czech", role="aut", 
           comment = c(ORCID = "0000-0002-9908-3007")),
    person("Marc", "Hafner", role="aut",
           comment = c(ORCID = "0000-0003-1337-7598")),
    person("Sergiu", "Mocanu", role="aut"),
    person("Dariusz", "Scigocki", role="aut"),
    person("Allison", "Vuong", role="aut"),
    person("Luca", "Gerosa", role="aut",
           comment = c(ORCID = "0000-0001-6805-9410")),
    person("Janina", "Smola", role="aut"))
Description: The package is a part of the gDR suite. It helps to prepare raw drug response data for downstream processing. It mainly contains helper functions for importing/loading/validating dose-response data provided in different file formats. 
License: Artistic-2.0
LazyLoad: yes
Depends: R (>= 4.2)
Imports: assertthat, BumpyMatrix, checkmate, CoreGx, PharmacoGx,
        data.table, futile.logger, gDRutils (>= 1.7.1), magrittr,
        methods, MultiAssayExperiment, readxl, rio, S4Vectors, stats,
        stringi, SummarizedExperiment, tibble, tools, utils, XML, yaml,
        openxlsx, qs2
Suggests: BiocStyle, gDRtestData (>= 1.7.1), gDRstyle (>= 1.7.1),
        knitr, purrr, testthat
URL: https://github.com/gdrplatform/gDRimport,
        https://gdrplatform.github.io/gDRimport/
BugReports: https://github.com/gdrplatform/gDRimport/issues
biocViews: Software, Infrastructure, DataImport
VignetteBuilder: knitr
ByteCompile: TRUE
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
SwitchrLibrary: gDRimport
DeploySubPath: gDRimport
Encoding: UTF-8
git_url: https://git.bioconductor.org/packages/gDRimport
git_branch: devel
git_last_commit: b9e01bd
git_last_commit_date: 2026-06-17
Repository: Bioconductor 3.24
Date/Publication: 2026-06-17
NeedsCompilation: no
Packaged: 2026-06-17 22:01:00 UTC; biocbuild
Author: Arkadiusz Gladki [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7059-6378>),
  Bartosz Czech [aut] (ORCID: <https://orcid.org/0000-0002-9908-3007>),
  Marc Hafner [aut] (ORCID: <https://orcid.org/0000-0003-1337-7598>),
  Sergiu Mocanu [aut],
  Dariusz Scigocki [aut],
  Allison Vuong [aut],
  Luca Gerosa [aut] (ORCID: <https://orcid.org/0000-0001-6805-9410>),
  Janina Smola [aut]
Maintainer: Arkadiusz Gladki <gladki.arkadiusz@gmail.com>

# Generated by roxygen2: do not edit by hand

export(check_metadata_names)
export(convert_LEVEL5_prism_to_gDR_input)
export(convert_LEVEL6_prism_to_gDR_input)
export(convert_MAE_to_PSet)
export(convert_pset_to_df)
export(detect_file_format)
export(getPSet)
export(get_exception_data)
export(get_test_D300_data)
export(get_test_EnVision_data)
export(get_test_Tecan_data)
export(get_test_data)
export(get_test_tsv_data)
export(import_D300)
export(is_readable_v)
export(load_data)
export(load_manifest)
export(load_results)
export(load_templates)
export(manifest_path)
export(parse_D300_xml)
export(read_excel_to_dt)
export(result_path)
export(setEnvForPSet)
export(standardize_record_values)
export(template_path)
exportClasses(gdr_test_data)
importFrom(data.table,":=")
importFrom(data.table,setDF)
importFrom(magrittr,"%>%")
importFrom(methods,new)
importFrom(methods,slotNames)
importFrom(tibble,tribble)
importFrom(utils,adist)

#' gDR Test Data object
#'
#' Object class `gdr_test_data` is build by function [gDRimport::get_test_data()]
#'
#' @slot manifest_path character, path to manifest file
#' @slot result_path character, path(s) to results file
#' @slot template_path character, path(s) to data.table with template data
#' @slot ref_m_df character, data.table with manifest data
#' @slot ref_r1_r2 character, path to reference file with raw data for treated & untreated
#' @slot ref_r1 character, path to reference file with raw data for treated
#' @slot ref_t1_t2 character, path to reference template file with treated & untreated data
#' @slot ref_t1 character, path to reference template file with treated data
#'
#' @return object class `gdr_test_data` with primary test data
#'
#' @docType class
#' @name gdr_test_data-class
#' @keywords classes test_data_class
#'
#' @export
setClass(
  Class = "gdr_test_data",
  package = "gDRimport",
  slots = list(
    manifest_path = "character",
    result_path = "character",
    template_path = "character",
    ref_m_df = "data.table",
    ref_r1_r2 = "character",
    ref_r1 = "character",
    ref_t1_t2 = "character",
    ref_t1 = "character"
  )
)

# Basic method show
setMethod("show", "gdr_test_data",
          function(object) {
            cat("class:", class(object)[1], "\nslots:", slotNames(object))
          }
)

#' Method manifest_path
#'
#' Method for object gdr_test_data - access to slot `manifest_path`
#' @param x object class gdr_test_data
#'
#' @return value of slot `manifest_path`
#'
#' @examples
#'  td <- get_test_data()
#'  manifest_file_path <- manifest_path(td)
#'
#' @docType methods
#' @rdname manifest_path-method
#' @keywords methods test_data_class
#'
#' @export
setGeneric("manifest_path", function(x) standardGeneric("manifest_path"))

#' @aliases manifest_path,gdr_test_data-method
#' @rdname manifest_path-method
#' @keywords test_data_class
setMethod("manifest_path", "gdr_test_data", function(x) x@manifest_path)

#' Method result_path
#'
#' Method for object gdr_test_data - access to slot `result_path`
#' @param x object class gdr_test_data
#'
#' @return value of slot `result_path`
#'
#' @examples
#'  td <- get_test_data()
#'  result_file_path <- result_path(td)
#'
#' @docType methods
#' @rdname result_path-method
#' @keywords methods test_data_class
#'
#' @export
setGeneric("result_path", function(x) standardGeneric("result_path"))

#' @aliases result_path,gdr_test_data-method
#' @rdname result_path-method
#' @keywords test_data_class
setMethod("result_path", "gdr_test_data", function(x) x@result_path)

#' Method template_path
#'
#' Method for object gdr_test_data - access to slot `template_path`
#' @param x object class gdr_test_data
#' @keywords test_data_class
#'
#' @return value of slot `template_path`
#'
#' @examples
#'  td <- get_test_data()
#'  template_file_path <- template_path(td)
#'
#' @docType methods
#' @rdname template_path-method
#' @keywords methods test_data_class
#'
#' @export
setGeneric("template_path", function(x) standardGeneric("template_path"))

#' @aliases template_path,gdr_test_data-method
#' @rdname template_path-method
setMethod("template_path", "gdr_test_data", function(x) x@template_path)

#' get primary test data
#'
#' @examples
#' get_test_data()
#'
#' @keywords test_data_class
#' @export
#'
#' @return object class "gdr_test_data" with with input data (manifest/template/result paths)
#' and related reference data (qs2 file paths)
get_test_data <- function() {

    ddir <- system.file(package = "gDRimport", "extdata", "data1")

    ## define manifest ref data
    bcode_tbl <- expand.grid(c("201904190", "201904197"), letters[seq_len(6)])
    bcode_v <- paste0(bcode_tbl$Var1, bcode_tbl$Var2)
    templates_v <- c("Template_Untreated.xlsx", "Template_7daytreated.xlsx")
    clids_n <- c(rep(11, 2), rep(12, 2), rep(13, 2), rep(14, 2), rep(15, 2), rep(18, 2))
    ref_m_df <-
      data.table::data.table(
        Barcode = bcode_v,
        Duration = rep(c(0, 168), 6),
        Template = rep(templates_v, 6),
        clid = paste0("CL000", clids_n)
      )

    new("gdr_test_data",
        manifest_path = file.path(ddir, "manifest.xlsx"),
        result_path =
          c(
            file.path(ddir, "RawData_day0.xlsx"),
            file.path(ddir, "RawData_day7.xlsx")
          ),
        template_path =
          c(
            file.path(ddir, "Template_7daytreated.xlsx"),
            file.path(ddir, "Template_Untreated.xlsx")
          ),
        ref_m_df = ref_m_df,
        ref_r1_r2 = file.path(ddir, "ref_RawData_day0_day7_xlsx.csv"),
        ref_r1 = file.path(ddir, "ref_RawData_day0_xlsx.csv"),
        ref_t1_t2 =  file.path(ddir, "ref_template_treated_untreated_xlsx.csv"),
        ref_t1 = file.path(ddir, "ref_template_treated_xlsx.csv")
    )
  }

#' get test Tecan data
#'
#' @examples
#' get_test_Tecan_data()
#
#' @keywords test_data
#' @export
#'
#' @return list with with input data (manifest/template/result paths)
#' and related reference data (qs2 file paths)
get_test_Tecan_data <- function() {
    ddir <- system.file(package = "gDRimport", "extdata", "data2")
    list(
      m_file = file.path(ddir, "manifest_Tecan_96_well_plates.xlsx"),
      r_files = file.path(ddir, "RawData_Tecan_96_well_plates.xlsx"),
      t_files =
        c(
          file.path(ddir, "D300_trt_Tecan_96_well_plates.xlsx")
        ),
      ref_m_df = file.path(ddir, "ref_manifest_Tecan_96_well_plates.qs2"),
      ref_r_df = file.path(ddir, "ref_RawData_Tecan_96_well_plates.qs2"),
      ref_t_df = file.path(ddir, "ref_D300_trt_Tecan_96_well_plates.qs2")
    )
  }

#' get test D300 data
#'
#' @examples
#' get_test_D300_data()
#
#' @keywords test_data
#' @export
#'
#' @return list with with input data (manifest/template/result paths)
#' and related reference data (qs2 file paths)
get_test_D300_data <- function() {
    ddir <- system.file(package = "gDRimport", "extdata", "data3")
    list(
      f_96w = list(
        d300 = file.path(ddir, "D300_96_well_plate_example.tdd"),
        Gnum = file.path(ddir, "Gnumber_D300_96_well_plate.xlsx"),
        dest_path = file.path(ddir, "output_files_96w", "output"),
        ref_d300 = file.path(ddir, "ref_D300_96_well_plate_example.qs2"),
        ref_Gnum = file.path(ddir, "ref_Gnumber_D300_96_well_plate.qs2"),
        ref_output_path = file.path(ddir, "output_files_96w", "reference")
      ),
      f_384w = list(
        d300 = file.path(ddir, "D300_384_well_plate_example.tdd"),
        Gnum = file.path(ddir, "Gnumber_D300_384_well_plate.xlsx"),
        ref_Gnum = file.path(ddir, "ref_Gnumber_D300_384_well_plate.qs2"),
        ref_d300 = file.path(ddir, "ref_D300_384_well_plate_example.qs2"),
        dest_path = file.path(ddir, "output_files_384w", "output"),
        ref_output_path = file.path(ddir, "output_files_384w", "reference")
      )
    )
  }

#' get test EnVision data
#'
#' @examples
#' get_test_EnVision_data()
#
#' @keywords test_data
#' @export
#'
#' @return list with with input data (manifest/template/result paths)
#' and related reference data (.qs2 file paths)
get_test_EnVision_data <- function() {
    ddir <- system.file(package = "gDRimport", "extdata", "data4")
    fls <- list.files(ddir, full.names = TRUE)
    list(
      m_file = grep("\\/Manifest.xlsx$", fls, value = TRUE),
      r_files = grep("\\/[H|K].+.csv$", fls, value = TRUE),
      t_files = grep("\\/Project40.+.xlsx$", fls, value = TRUE),
      ref_l_path = file.path(ddir, "ref_l.qs2")
    )
}


#' get test tsv data
#'
#' @examples
#' get_test_tsv_data()
#
#' @keywords test_data
#' @export
#'
#' @return list with with input data (manifest/template/result paths)
#' and related reference data (.qs2 file paths)
get_test_tsv_data <- function() {
  ddir <- system.file(package = "gDRimport", "extdata", "data5")
  fls <- list.files(ddir, full.names = TRUE)
  list(
    m_file = grep("\\/Manifest.tsv$", fls, value = TRUE),
    r_files = grep("\\RawData.tsv$", fls, value = TRUE),
    t_files = grep("\\Template.+.tsv$", fls, value = TRUE),
    ref_l_path = file.path(ddir, "ref_l.qs2")
  )
}

context("assert_utils")

test_that("assert_utils", {
  td1 <- get_test_data()
  expect_true(is_readable_v(manifest_path(td1)))
  expect_true(is_readable_v(result_path(td1)))
  err_msg1 <-
    "Following path(s) with no read permission found: '/non/existent/file'"
  expect_error(is_readable_v(c(manifest_path(td1), "/non/existent/file")), err_msg1, fixed = TRUE)

  td2 <- get_test_Tecan_data()
  expect_true(is_readable_v(td2$m_file))
  expect_true(is_readable_v(td2$r_files))
  expect_true(is_readable_v(td2$t_files))
  err_msg1 <-
    "Following path(s) with no read permission found: '/non/existent/file'"
  expect_error(is_readable_v(c(td2$m_file, "/non/existent/file")), err_msg1, fixed = TRUE)

})

## ----init-import, echo=FALSE--------------------------------------------------
library(gDRimport)
log_level <- futile.logger::flog.threshold("ERROR")

## ----load-test-datasets-------------------------------------------------------
# primary test data
td1 <- get_test_data()
summary(td1)
td1

# test data in Tecan format
td2 <- get_test_Tecan_data()
summary(td2)

# test data in D300 format
td3 <- get_test_D300_data()
summary(td3)

# test data obtained from EnVision
td4 <- get_test_EnVision_data()
summary(td4)

## ----execute-data-loading-----------------------------------------------------
ml <- load_manifest(manifest_path(td1))
summary(ml)

t_df <- load_templates(template_path(td1))
summary(t_df)

r_df <- suppressMessages(load_results(result_path(td1)))
summary(r_df)

l_tbl <-
  suppressMessages(
    load_data(manifest_path(td1), template_path(td1), result_path(td1)))
summary(l_tbl)

## ----process-prism-level5, echo=TRUE, eval=FALSE------------------------------
# convert_LEVEL5_prism_to_gDR_input("path_to_file")

## ----process-prism-level6, echo=TRUE, eval=FALSE------------------------------
# convert_LEVEL6_prism_to_gDR_input("prism_data_path", "cell_line_data_path", "treatment_data_path")

## ----session-info-------------------------------------------------------------
sessionInfo()

global_valid_notes_l <- list(
  list(
    length = 5,
    index_to_check = 2,
    text_to_check = "installed size is"
  ),
  list(
    length = 4,
    index_to_check = 2,
    text_to_check = "installed size is"
  ),
  list(
    length = 5,
    index_to_check = 2,
    text_to_check = "Found the following hidden files and directories"
  )
)

#' Convert MultiAssayExperiment to TreatmentResponseExperiment
#'
#' This function converts a MultiAssayExperiment generated by gDR into a TreatmentResponseExperiment for use in the
#' PharmacoGx package. The resulting PharmacoSet can be used for pharmacogenomic analysis of drug response.
#'
#' @param mae A MultiAssayExperiment object generated by gDR.
#' @param pset_name A character string specifying the name of the resulting PharmacoSet object.
#'
#' @keywords pset_conversion
#'
#' @return A PharmacoSet object.
#'
#' @examples
#' # Convert a MultiAssayExperiment object to a PharmacoSet object
#' m <- 20
#' n <- 10
#' rnames <- LETTERS[1:m]
#' cnames <- letters[1:n]
#' ref_gr_value <-  matrix(runif(m * n), nrow = m, ncol = n, dimnames = list(rnames, cnames))
#' se <- SummarizedExperiment::SummarizedExperiment(assays = list(RefGRvalue = ref_gr_value),
#'                                                  rowData = S4Vectors::DataFrame(rnames),
#'                                                  colData = S4Vectors::DataFrame(cnames))
#' mae <- MultiAssayExperiment::MultiAssayExperiment(experiments = list("single-agent" = se))
#' convert_MAE_to_PSet(mae, "my_pset")
#'
#' @export
convert_MAE_to_PSet <- function(mae,
                                pset_name) {
    checkmate::assert_class(mae, "MultiAssayExperiment")
    checkmate::assert_string(pset_name)

    # Get the unique assay names from all experiments combined
    assay_names <- unique(unlist(gDRutils::MAEpply(mae, function(x) SummarizedExperiment::assayNames(x))))

    # Create a list of data.tables, one for each assay name, with the data from all Summarized Experiments in the MAE.
    assay_list <- lapply(assay_names, function(x) {

        # Get list of all the Summarized Experiments with all the unique Assays across the entire MAE
        # If the summarized experiment does not have a given assay, create it
        se_list <- gDRutils::MAEpply(mae = mae, FUN = .extract_or_create_assay, unify = FALSE, x)

        # Get list of datatables with all assays with name=x converted to a data.table
        se_dt_list <- lapply(se_list, FUN = gDRutils::convert_se_assay_to_dt, x)

        # Combine all the data.tables into one data.table, add the name of the Summarized Experiment
        # as a "data_type" column so that users can identify which Summarized Experiment the data came from
        DT <- data.table::rbindlist(se_dt_list, fill = TRUE, idcol = "data_type")

        # Change column names
        data.table::setnames(DT, "rId", "treatmentid")
        data.table::setnames(DT, "cId", "sampleid")
    })

    # Set the names of the list of data.tables to the assay names so that they can be accessed by name later
    names(assay_list) <- assay_names

    # Combine all the data.tables into one data.table
    assay_dt <- data.table::rbindlist(assay_list, fill = TRUE, idcol = "assay_name")

    # Get the unique rownames from all experiments combined
    rownames <- unique(unlist(gDRutils::MAEpply(mae, FUN = rownames)))
    # convert the rowData of each Summarized Experiment to a data.table and combine them into one data.table
    rowdata_ <- gDRutils::MAEpply(
      mae = mae,
      FUN = function(x) data.table::as.data.table(data.frame(SummarizedExperiment::rowData(x)), keep.rownames = TRUE),
      unify = TRUE)
    # rename the rownames column to "treatmentid" (required for PharmacoSet Object)
    data.table::setnames(rowdata_, "rn", "treatmentid")


    # Get the unique colnames from all experiments combined
    colnames <- unique(unlist(gDRutils::MAEpply(mae, FUN = colnames)))
    # convert the colData of each Summarized Experiment to a data.table and combine them into one data.table
    coldata_ <- unique(gDRutils::MAEpply(
      mae = mae,
      FUN = function(x) data.table::as.data.table(data.frame(SummarizedExperiment::colData(x)), keep.rownames = TRUE),
      unify = TRUE))
    # rename the rownames column to "sampleid" (required for PharmacoSet Object)
    data.table::setnames(coldata_, "rn", "sampleid")

    # Create a list of rowname and column name identifiers for each assay name (required for
    # TreatmentResponseExperiment object)
    # in this case, all the identifiers are the same treatmentid & sampleid columns
    assayIDs <- stats::setNames(lapply(assay_names, function(x) c("treatmentid", "sampleid")), assay_names)

    tre <- CoreGx::TreatmentResponseExperiment(
        rowData = rowdata_,
        colData = coldata_,
        rowIDs = "treatmentid",
        colIDs = "sampleid",
        assays = assay_list,
        assayIDs = assayIDs,
        metadata = list()
    )

    # add batchid column to coldata_ (required for PSet)
    coldata_$batchid <- 1

    # Create PharmacoSet object (implicit return)
    # The molecularProfiles slot is empty because the gDR object does not contain molecular profiles
    # As of this version, curation slot is left empty.
    PharmacoGx::PharmacoSet2(
        name = pset_name,
        treatment = as.data.frame(rowdata_), # nolint
        sample = as.data.frame(coldata_), # nolint
        molecularProfiles =
            MultiAssayExperiment::MultiAssayExperiment(
                MultiAssayExperiment::ExperimentList(
                    emptySE = SummarizedExperiment::SummarizedExperiment(
                        assays = list(),
                        colData = coldata_
                ))),
        treatmentResponse = tre,
        perturbation = list(),
        curation = list(sample = data.frame(),
                        treatment = data.frame(),
                        tissue = data.frame()),
        datasetType = "sensitivity")

}


#' Extracts an assay from a SummarizedExperiment object or creates a new one if it does not exist
#'
#' This function takes a SummarizedExperiment object and an assay name as input.
#' If the specified assay already exists in the SummarizedExperiment object, it is returned.
#' Otherwise, a new assay with the specified name is created and added to the SummarizedExperiment object.
#' The new assay is initialized with NA values.
#' This is useful for when multiple Summarized Experiments in a given MAE do not have the same assays.
#' And it is necessary to have the same assays in all Summarized Experiments in order to convert the MAE to a PSet.
#'
#' @param SE A SummarizedExperiment object
#' @param assay_name A character string specifying the name of the assay to extract or create
#'
#' @keywords internal
#'
#' @return A SummarizedExperiment object with the specified assay
#'
.extract_or_create_assay <- function(SE,
                                     assay_name) {
    checkmate::assert_class(SE, "SummarizedExperiment")
    checkmate::assert_string(assay_name)
    if (assay_name %in% SummarizedExperiment::assayNames(SE)) {
        SE
    } else {
        SummarizedExperiment::assay(SE, assay_name) <-
            BumpyMatrix::splitAsBumpyMatrix(
                S4Vectors::DataFrame(x = rep(NA, prod(dim(SE)))),
                row = rownames(SE),
                column = colnames(SE)
            )
        SE
    }
}

test_that("convert_MAE_to_PSet works as expected", {
  m <- 20
  n <- 10
  rnames <- LETTERS[1:m]
  cnames <- letters[1:n]

  # Normal matrix.
  ref_gr_value <-  matrix(runif(m * n), nrow = m, ncol = n, dimnames = list(rnames, cnames))
  se <- SummarizedExperiment::SummarizedExperiment(assays = list(RefGRvalue = ref_gr_value),
                                                   rowData = S4Vectors::DataFrame(rnames),
                                                   colData = S4Vectors::DataFrame(cnames))

  mae <- MultiAssayExperiment::MultiAssayExperiment(experiments = list("single-agent" = se))

  pset <- convert_MAE_to_PSet(mae, "my_pset")
  expect_equal(class(pset@treatmentResponse)[1], "TreatmentResponseExperiment")
  expect_equal(sort(rownames(pset@treatmentResponse)),
               sort(unlist(lapply(MultiAssayExperiment::experiments(mae), rownames), use.names = FALSE)))
  expect_equal(pset@sample$sampleid, cnames)


  se1 <- SummarizedExperiment::SummarizedExperiment(assays = list(RefGRvalue = ref_gr_value[1:10, ]),
                                                   rowData = S4Vectors::DataFrame(rnames)[1:10, , drop = FALSE],
                                                   colData = S4Vectors::DataFrame(cnames))

  se2 <- SummarizedExperiment::SummarizedExperiment(assays = list(RefGRvalue = ref_gr_value[11:20, ]),
                                                    rowData = S4Vectors::DataFrame(rnames)[11:20, , drop = FALSE],
                                                    colData = S4Vectors::DataFrame(cnames))
  maeTwoExperiments <- MultiAssayExperiment::MultiAssayExperiment(experiments = list("single-agent" = se1,
                                                                                     "matrix" = se2))

  pset_mae2Exp <- convert_MAE_to_PSet(maeTwoExperiments, "my2exp")
  expect_equal(class(pset_mae2Exp@treatmentResponse)[1], "TreatmentResponseExperiment")
  expect_equal(sort(rownames(pset_mae2Exp@treatmentResponse)),
               sort(unlist(lapply(MultiAssayExperiment::experiments(maeTwoExperiments), rownames), use.names = FALSE)))
  expect_equal(pset_mae2Exp@sample$sampleid, cnames)
  expect_equal(pset_mae2Exp@treatment$treatmentid,
               c(
                 rownames(
                   SummarizedExperiment::assay(maeTwoExperiments, "single-agent")
                 ),
                 rownames(SummarizedExperiment::assay(maeTwoExperiments, "matrix"))
               ))
})

#' Convert a PharmacoSet to a data.table that is prepare for input into gDR pipeline
#'
#' @param pharmacoset PharmacoSet object
#' @param run_parallel logical, TRUE (default) if to run functions in Parallel, FALSE to run in serial
#' @param workers integer, number of workers defaults to 2L if run_parallel is TRUE
#'
#' @keywords pset_conversion
#'
#' @return data.table of PharmacoSet's dose response data with column names aligned with gDR standard
#'
#' @examples
#' pset <- suppressMessages(getPSet(
#'   "Tavor_2020",
#'   psetDir = system.file("extdata/pset", package = "gDRimport"),
#'   use_local_PSets_list = TRUE
#' ))
#' dt <- convert_pset_to_df(pset)
#' gDRutils::reset_env_identifiers()
#'
#' @author Jermiah Joseph -- collaboration with BHKLab
#'
#' @export
convert_pset_to_df <- function(pharmacoset,
                           run_parallel = TRUE,
                           workers = 2L) {
  assertthat::assert_that(is.logical(run_parallel),
                          msg = "run_parallel must be a logical.")
  assertthat::assert_that(inherits(pharmacoset, "PharmacoSet"),
                          msg = "pharmacoset paramater must inherit from PharmacoSet class.")
  assertthat::assert_that(is.integer(workers),
                          msg = "workers parameter must be an integer. Default = 2L")

  # FUNCTION TO MANIPULATE ENV IDENTIFIERS
  setEnvForPSet()

  # GET DOSE AND VIABILITY DATA & MELT INTO LARGE TABLE
  dose_response <- .extractDoseResponse(pset = pharmacoset)

  # ADD IN DURATION AND REFERENCE DIVISION TIME
  dose_response_duration_refdivtime <- .createPseudoData(dose_response)

  # REMOVE NEGATIVE VIABILITIES
  .removeNegatives(dose_response_duration_refdivtime)
}

#' Adjust environment variables to meet gDR standards
#'
#' @examples
#' setEnvForPSet()
#' gDRutils::reset_env_identifiers()
#'
#' @keywords pset_conversion
#'
#' @return \code{NULL}
#'
#' @export
setEnvForPSet <- function() {
  ## -- Set environment identifiers to map from our columns to gDR columns
  gDRutils::reset_env_identifiers()

  pgx_to_gdr_ids <- list(
      cellline = "Clid",
      cellline_name = "Clid",
      drug = "DrugName",
      drug_name = "DrugName",
      concentration = "Concentration",
      duration = "Duration",
      barcode = "Barcode"
  )

  invisible(Map(function(x, y) gDRutils::set_env_identifier(x, y),
                names(pgx_to_gdr_ids), pgx_to_gdr_ids))
}

#' Get PharmacoSet
#'
#' @param pset_name string with the name of the PharmacoSet
#' @param psetDir string with the temporary directory for the PharmacoSet
#' @param canonical logical flag indicating if the PSet canonical
#' @param timeout maximum number of seconds allowed for PSet download
#' @param use_local_PSets_list logical flag if PSets list should be used from
#' local. If FALSE PSets list will be taken from web.
#'
#' @keywords pset_conversion
#'
#' @examples
#' suppressMessages(getPSet(
#'   "Tavor_2020",
#'   psetDir = system.file("extdata/pset", package = "gDRimport"),
#'   use_local_PSets_list = TRUE
#' ))
#'
#' @return PharmacoSet object
#' @export
getPSet <- function(pset_name,
                    psetDir = getwd(),
                    canonical = FALSE,
                    timeout = 600,
                    use_local_PSets_list = FALSE) {

  assertthat::assert_that(is.character(pset_name),
                          msg = "pset_name parameter must be a character vector.")

  checkmate::assert_character(getwd())
  checkmate::assert_flag(canonical)
  checkmate::assert_numeric(timeout)

  availPSets <- if (use_local_PSets_list) {
    qs2::qs_read(system.file("extdata", "data_for_unittests", "PSets.qs2", package = "gDRimport"))
  } else {
    PharmacoGx::availablePSets(canonical = canonical)
  }

  pset_name_param <- if (pset_name %in% availPSets$"Dataset Name") {
    availPSets[availPSets$"Dataset Name" == pset_name, "PSet Name"]
    } else if (pset_name %in% availPSets$"PSet Name") {
      pset_name
      } else {
        stop(pset_name,
             " does not exist in the available PSets. Try one of the following:\n",
             paste(availPSets$`PSet Name`, collapse = "\n"))
        }

  # Check if PSet exists in directories where PSets are stored.
  # Read in if exists, download otherwise
  pset <- if (file.exists(file.path(psetDir, paste0(pset_name_param, ".qs2")))) {
    message("PSet exists in user-provided directory, reading .qs2 file")
    qs2::qs_read(file.path(psetDir, paste0(pset_name_param, ".qs2")))
    } else {
      message("PSet does not exist in user-provided directory, downloading from database.")
      PharmacoGx::downloadPSet(pset_name_param, saveDir = psetDir, timeout = timeout)
    }
  # Update PSet to new structure if not updated already.
  PharmacoGx::updateObject(pset)
}

#' Get dose and viability readouts and melt into large data table
#' @keywords internal
#'
#' @return data.table with dose-response data
#'
.extractDoseResponse <- function(pset) {
  checkmate::assert_class(pset, "PharmacoSet")

  tre <- pset@treatmentResponse
  raw_tr <- tre$raw
  info_dt <- data.table::as.data.table(tre$info, keep.rownames = TRUE)
  duration <- unique(tre$info$duration.hours)

  # use output of get_env_identifiers()
  env_ids <- gDRutils::get_env_identifiers()
  # Determine how many doses there are
  raw_tr_dims <- dim(raw_tr)
  if (raw_tr_dims[3] < 2) {
      stop("PharmacoSet treatmentResponse raw data does not have
           either/both of viability and concentration data.")
  } else {
      viability <- data.table::as.data.table(raw_tr[, seq_len(raw_tr_dims[2]), 2], TRUE)
      viability[[env_ids$untreated_tag[1]]] <- 100
      viability.m <- data.table::melt(viability,
                  measure.vars = c(2:length(viability)),
                  variable.name = "Dose",
                  value.name = "ReadoutValue")

      doses <- data.table::as.data.table(raw_tr[, seq_len(raw_tr_dims[2]), 1], TRUE)
      doses[[env_ids$untreated_tag[1]]] <- 0
      doses.m <- data.table::melt(doses,
                  measure.vars = c(2:length(doses)),
                  variable.name = "Dose",
                  value.name = env_ids$concentration)
  }
  # CHECK IF SAME SIZE and MERGE
  if (length(doses.m) == length(viability.m)) {
      merged_dt <- viability.m[doses.m, on = intersect(names(viability.m),
                                                       names(doses.m))]
  } else {
      sprintf("doses and viability data tables are not the same size")
  }
  treatment_cols <- c("sampleid", "treatmentid")
  selected_cols <- c(treatment_cols, "rn")
  merged_dt <- merged_dt[info_dt[, selected_cols, with = FALSE], on = "rn"]
  data.table::setnames(merged_dt, treatment_cols, c(env_ids$cellline, env_ids$drug_name))
  merged_dt[Dose == env_ids$untreated_tag[1], env_ids$drug_name := env_ids$untreated_tag[1]]
  merged_dt[, Dose := NULL]

  if (!is.null(duration)) {
    merged_dt[, (env_ids$duration) := duration]
  }
  merged_dt
}

#' Add in pseudo-data for duration and cell reference division time
#' @keywords internal
#'
#' @return data.table
.createPseudoData <- function(dt) {

  checkmate::assert_data_table(dt)

  barcode <- gDRutils::get_env_identifiers("barcode")[1]
  duration <- gDRutils::get_env_identifiers("duration")
  refDivTime <- gDRutils::get_env_identifiers("cellline_ref_div_time")


  if (!duration %in% names(dt)) {
    dt[, (duration) := NA_real_]
  }
  if (!refDivTime %in% names(dt)) {
    dt[, (refDivTime) := NA]
  }
  if (!barcode %in% names(dt)) {
    colnames(dt)[1] <- barcode
  }

  dt
}


#' Remove negative viabilities
#' @keywords internal
#'
#' @return data.table with positive values in column `ReadoutValue`
#'
.removeNegatives <- function(dataset) {
  checkmate::assert_data_table(dataset)
  dataset[dataset$ReadoutValue > 0]
}

context("PSets")

# in tests we used mocked availablePSets in case of lack of internet connection.
# to prepare new mock data:
# qs2::qs_save(PharmacoGx::availablePSets(canonical = FALSE),
#           system.file("extdata", "data_for_unittests", "PSets.qs2", package = "gDRimport"))

read_mocked_PSets <- function(canonical = FALSE) {
  qs2::qs_read(
    system.file("extdata", "data_for_unittests", "PSets.qs2", package = "gDRimport")
  )
}

test_that("getPSet works as expected", {
  pset <- testthat::with_mocked_bindings(
    availablePSets = read_mocked_PSets,
    .package = "PharmacoGx",
    suppressMessages(getPSet(
      "Tavor_2020", psetDir = system.file("extdata/pset", package = "gDRimport")
    ))
  )
  expect_equal(unname(dim(pset)), c(53, 46))
  expect_s4_class(pset, "PharmacoSet")
})


test_that("setEnvForPSet works as expected", {
  on.exit(gDRutils::reset_env_identifiers())
  setEnvForPSet()
  expect_equal(gDRutils::get_env_identifiers("drug"), "DrugName")
})

test_that(".extractDoseResponse works as expected", {
  pset <- testthat::with_mocked_bindings(
    availablePSets = read_mocked_PSets,
    .package = "PharmacoGx",
    suppressMessages(getPSet(
      "Tavor_2020", psetDir = system.file("extdata/pset", package = "gDRimport")
    ))
  )
  dt <- .extractDoseResponse(pset)
  expect_s3_class(dt, "data.table")
  expect_equal(names(dt), c("rn", "ReadoutValue", "Concentration", "clid", "DrugName",
                            "Duration"))
  expect_equal(dim(dt), c(34684, 6))
})

test_that(".extractDoseResponse, .removeNegatives, and .createPseudoData work as expected", {
  pset <- testthat::with_mocked_bindings(
    availablePSets = read_mocked_PSets,
    .package = "PharmacoGx",
    suppressMessages(getPSet(
      "Tavor_2020", psetDir = system.file("extdata/pset", package = "gDRimport")
    ))
  )
  dt <- .extractDoseResponse(pset)
  expect_s3_class(dt, "data.table")
  expect_equal(names(dt), c("rn", "ReadoutValue", "Concentration", "clid", "DrugName",
                            "Duration"))
  expect_equal(dim(dt), c(34684, 6))

  dt_positive <- .removeNegatives(dt)
  expect_equal(dim(dt_positive), c(34516, 6))

  dt_with_pseudodata <- .createPseudoData(dt_positive)
  expect_equal(dim(dt_with_pseudodata), c(34516, 7))
})

test_that("convert_pset_to_df works as expected", {
  on.exit(gDRutils::reset_env_identifiers())
  pset <- testthat::with_mocked_bindings(
    availablePSets = read_mocked_PSets,
    .package = "PharmacoGx",
    suppressMessages(getPSet(
      "Tavor_2020", psetDir = system.file("extdata/pset", package = "gDRimport")
    ))
  )
  dt <- convert_pset_to_df(pset)
  checkmate::assert_data_table(dt)
  expect_equal(dim(dt), c(34516, 7))
  expect_equal(names(dt), c("Barcode", "ReadoutValue", "Concentration", "Clid", "DrugName",
                            "Duration", "ReferenceDivisionTime"))
})

## ----knitr-init, include = FALSE----------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----load-deps----------------------------------------------------------------
library(PharmacoGx)
library(gDRimport)

## ----get-pset-example, eval = FALSE-------------------------------------------
# pset <- getPSet("Tavor_2020")
# pset

## ----load-mocked-psets, include = FALSE---------------------------------------
read_mocked_PSets <- function(canonical = FALSE) {
  qs2::qs_read(
    system.file("extdata", "data_for_unittests", "PSets.qs2", package = "gDRimport")
  )
}
pset <- testthat::with_mocked_bindings(
    availablePSets = read_mocked_PSets,
    .package = "PharmacoGx",
    suppressMessages(getPSet(
      "Tavor_2020", psetDir = system.file("extdata/pset", package = "gDRimport")
    ))
)
pset

## ----convert-pset-to-dt-------------------------------------------------------
# Store treatment response data in df_
dt <- convert_pset_to_df(pharmacoset = pset)
str(dt)

## ----subset-by-cell-line------------------------------------------------------
# example subset using only 1 cell line
subset_cl <- dt$Clid[1]
x <- dt[Clid == subset_cl]
x

## ----run-gdr-pipeline, eval = FALSE-------------------------------------------
# # RUN DRUG RESPONSE PROCESSING PIPELINE
# se <- gDRcore::runDrugResponseProcessingPipeline(x)
# se

## ----convert-se-to-dt, eval = FALSE-------------------------------------------
# # Convert Summarized Experiments to data.table
# # Available SEs : "RawTreatred", "Controls", "Normalized", "Averaged", "Metrics"
# 
# str(gDRutils::convert_se_assay_to_dt(se[[1]], "Averaged"))
# str(gDRutils::convert_se_assay_to_dt(se[[1]], "Metrics"))

## ----reset-identifiers, include = FALSE---------------------------------------
gDRutils::reset_env_identifiers()

## ----session-info-------------------------------------------------------------
sessionInfo()

#' @note To learn more about functions start with `help(package = "gDRimport")`
#' @keywords internal
#' @return package help page
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

#' standardize_record_values
#'
#' map values to a dictionary
#'
#' @param x a named array
#' @param dictionary a named array
#' @keywords utils
#'
#' @examples
#' standardize_record_values(c("Vehicle", "vehcle"))
#'
#' @return a named array with updated names
#'
#' @export
#'
standardize_record_values <- function(x, dictionary = DICTIONARY) {
  dnames <- names(dictionary)
  for (i in seq_along(dictionary)) {
    x[x == dnames[i]] <- dictionary[[i]]
  }
  x
}

#' read_ref_data
#'
#' Read reference data
#'
#' @param inDir a directory path of reference data
#' @param prefix a prefix of reference file names ('ref' by default)
#' @keywords utils
#'
#' @return a list of reference data
#'
read_ref_data <- function(inDir, prefix = "ref") {
  # Assertions:
  checkmate::assert_string(inDir)
  checkmate::assert_string(prefix)

  files <- list.files(inDir, paste0(prefix, "_.+\\.tsv$"), full.names = TRUE)
  lFiles <- lapply(files, function(x) {
    data.table::fread(x, sep = "\t", header = TRUE)
    })
  names(lFiles) <- gsub("\\.tsv", "", gsub(paste0("^", prefix, "_"), "", basename(files)))
  refKeys <- yaml::read_yaml(file.path(inDir, paste0(prefix, "_keys.yaml")))
  refRowMaps <- yaml::read_yaml(file.path(inDir, paste0(prefix, "_row_maps.yaml")))
  lFiles$ref_keys <- refKeys
  lFiles$ref_row_maps <- refRowMaps
  lFiles
}

#' Detect format of results data
#'
#' @param results_file path to results data
#' @keywords utils
#'
#' @examples
#' td2 <- get_test_Tecan_data()
#' detect_file_format(td2$r_files[1])
#'
#' @return string of the detected file format
#'
#' @export
detect_file_format <- function(results_file) {
  checkmate::assert_character(results_file)
  results_data <- rio::import(results_file, fill = TRUE)

  if (all(c("System", "User", "Plate", "Plate-ID (Stacker)") %in% results_data[, 1])) {
    "Tecan"
  } else if (any(grepl("Instrument Results from", c(names(results_data)[1],
                                                    results_data[, 1],
                                                    paste(results_data[1, ], collapse = " "))),
                 na.rm = TRUE)) {
    "EnVision_new"
  } else if ("Repeat Barcode" %in% c(
    do.call(paste, results_data[, 2:3]),
    paste(names(results_data[, 2:3]), collapse = " ")
  ) || any(grepl("Repeat,Barcode", results_data[, 1]))) {
    "EnVision"
  } else {
    "long_tsv"
  }
}

#' Read excel file and transorm it into data.table object
#'
#' @param path path to excel file
#' @param ... other arguments that should be passed into `readxl::read_excel`
#' @keywords utils
#'
#' @return data.table object with read excel file
#' @export
#'
#' @examples
#' datasets <- readxl::readxl_example("datasets.xlsx")
#' read_excel_to_dt(datasets)
read_excel_to_dt <- function(path, ...) {
  checkmate::assert_character(path)
  dt <- readxl::read_excel(path, ...)
  data.table::setDT(dt)
  dt
}

context("assert_utils")

test_that("assert_utils", {

  # real data
  td1 <- get_test_data()
  ref_tbl <- data.table::fread(td1@ref_t1)
  expect_equal(standardize_record_values(ref_tbl$Gnumber), ref_tbl$Gnumber)

  # toyset
  test_v <- c(letters[seq_len(4)], names(DICTIONARY))
  names(test_v) <- test_v
  res_v <- c(letters[seq_len(4)], as.character(DICTIONARY))
  names(res_v) <- names(test_v)
  expect_equal(standardize_record_values(test_v), res_v)
})

test_that("detect_file_format works as expected", {
  envision_path <- list.files(system.file(package = "gDRimport", "extdata", "data1"),
                           "^RawData", full.names = TRUE)
  expect_equal(unique(unlist(lapply(envision_path, detect_file_format))), "EnVision")

  tecan_path <- list.files(system.file(package = "gDRimport", "extdata", "data2"),
                              "^RawData", full.names = TRUE)
  expect_equal(unique(unlist(lapply(tecan_path, detect_file_format))), "Tecan")

  tsv_path <- list.files(system.file(package = "gDRimport", "extdata", "data1"),
                           "ref_RawData", full.names = TRUE)
  expect_equal(unique(unlist(lapply(tsv_path, detect_file_format))), "long_tsv")
})

test_that("read_excel_to_dt works as expected", {
  datasets <- readxl::readxl_example("datasets.xlsx")
  dt <- read_excel_to_dt(datasets)
  expect_s3_class(dt, "data.table")
})

#' get exception data
#'
#' @param status_code A numeric value
#'
#' @examples
#' get_exception_data(1)
#' get_exception_data()
#' @keywords correction_exception
#'
#' @return A data.table row with exception data or all exceptions
#' @export
get_exception_data <- function(status_code = NULL) {
  checkmate::assert_number(status_code, null.ok = TRUE)

  #nolint start
  exception_table <- tibble::tribble(
    ~status_code, ~title, ~sprintf_text, ~type, ~input_type,
    "1", "Error loading manifest", "There were errors loading manifest. Check error message below:\n```\n%s\n```", "error", "manifest",
    "2", "Manifest not found", "Please load manifest!", "error", "manifest",
    "3", "Error loading treatment", "Some of the treatment files entered in the manifest have not been uploaded:\n%s\nPlease check the uploaded files and their name (including extension)", "error", "treatment",
    "4", "Warning loading treatment", "Some uploaded files (**%s**) are not used", "warning", "treatment",
    "5", "Error loading treatment", "There were errors when loading the treatments. Check error message below:\n```\n%s\n```", "error", "treatment",
    "6", sprintf("Missing data for some %s", stringi::stri_flatten(gDRutils::get_env_identifiers("barcode"), collapse = ", ")), "Some plates (**%s**) found in the manifest are not reported in the data files and will be discarded", "warning", "manifest",
    "7", sprintf("Missing data for some %s", stringi::stri_flatten(gDRutils::get_env_identifiers("barcode"), collapse = ", ")), "Some plates (**%s**) found in the data files are not reported in the manifest and will be discarded", "warning", "manifest",
    "8", "Error loading raw data", "Txt file in an unsupported format. Please check if the file uses tab(\t) as separator", "error", "raw data",
    "9", "Error loading raw data", "There were errors when loading raw data. Check error message below:\n```\n%s\n```", "error", "raw data",
    "10", "Error during merging data", "There were errors when merging data. Check error message below:\n```\n%s\n```", "error", "merging data",
    "11", "Missing treatment files", "Some treatment files are missing: %s", "error", "treatment",
    "12", "Error reading Manifest", "There were errors reading the manifest file. Check error message below:\n```\n%s\n```", "error", "manifest",
    "13", "Manifest file format", "%s file format is not supported. Please convert your file to one of the following: %s", "error", "manifest",
    "14", "Manifest barcodes", "Barcodes in Manifest must be unique!", "error", "manifest",
    "15", "Treatment controls", "No untreated controls were found in the treatment. Please upload the appropriate treatment.", "error", "treatment",
    "16", "Instrument type", "Unrecognized instrument type", "error", "raw data",
    "17", "Treatment sheet name", "In untreated treatment file %s, sheet name must be %s", "error", "treatment",
    "18", "Treatment entry name", "In untreated treatment file %s, entries must be %s", "error", "treatment",
    "19", "Error loading treatment", "There were errors loading the treatment. Check error message below:\n```\n%s\n```", "error", "treatment",
    "20", "Error loading specific treatment", "There were errors loading treatment %s. Check error message below: \n%s", "error", "treatment",
    "21", "Error reading specific raw data", "Error reading %s", "error", "raw data",
    "22", "Error reading sheet in the raw data file", "Error reading %s, sheet %s", "error", "raw data",
    "23", "Error with readout values in plates", "In result file %s (sheet %s) readout values are misplaced for plate %s", "error", "raw data",
    "24", "Error with treatment headers", "Treatment does not contain all expected headers for a '%s'. '%s' is/are required. Please correct your treatment or change the identifiers.", "error", "treatment",
    "25", "Error with treatment file sheets", "Treatment file %s does not contain %s sheets. Please correct your treatment.", "error", "treatment",
    "26", "Metadata special characters", "Metadata field names for %s cannot contain special characters or start with a number: %s", "error", "metadata",
    "27", "Metadata invalid", "Metadata field name: %s in %s is not valid (reserved for output)", "error", "metadata",
    "28", "Raw Data delimiter", "Can't guess separator for the delimited files: %s", "error", "raw data",
    "29", "Raw Data EnVision", "Error reading %s: not an original EnVision .csv file", "error", "raw data",
    "30", "Raw Data Plate size", "Error reading %s: wrong plate size", "error", "raw data",
    "31", "Raw Data structure", "In result file %s (sheet %s) readout values are misplaced for plate %s.", "error", "raw data",
    "32", "Treatment File Gnumber and Concentration", "Treatment file(s) %s do/does not contain the same number of Gnumber_* and Concentration_* sheets. Gnumber_* and Concentration_* sheets are required. Please correct your treatment.", "error", "treatment",
    "33", "Error loading input files", "Some input files are missing. Please upload only files with names starting with Manifest_*, Treatment_* and RawData_*", "error", "input files",
    "34", "Missing drug annotation", "The following drugs IDs are missing: '%s'. <br>Don't worry, drug names will be used for them.", "warning", "submit tab",
    "35", "Missing cell line annotation", "The following cell line IDs are missing: '%s'. <br>Don't worry, cell line names will be used for them.", "warning", "submit tab",
    "36", "Invalid averaged data", "Averaged dose-response data for the selected cell line and drug: '%s' can be considered invalid. Please check your data in the module 'Manage Data' or contact gdrplatform team via 'gdr-support-d@gene.com'.", "error", "assay data",
    "37", "Raw Data header", "Invalid header in the result file: (%s)", "error", "raw data",
  )
  #nolint end

  res <- if (!is.null(status_code)) {
    checkmate::assert_choice(toString(status_code), exception_table$status_code)

    exception_table[exception_table$status_code == status_code, ]
  } else {
    exception_table
  }

  data.table::as.data.table(res)
}

library(testthat)

test_that("check get_exception_data works correctly", {
  obs <- gDRimport::get_exception_data(1)
  expect_equal(NROW(obs), 1)
  expect_equal(colnames(obs),
               c("status_code", "title", "sprintf_text", "type", "input_type"))
  expect_equal(obs$sprintf_text,
               "There were errors loading manifest. Check error message below:\n```\n%s\n```")

  expect_true(all(dim(gDRimport::get_exception_data()) >= 5))

  obs <- gDRimport::get_exception_data(10)
  expect_is(obs, "data.table")

  obs <- gDRimport::get_exception_data()
  expect_gt(NROW(obs), 1)
  expect_is(obs, "data.table")

  expect_error(
    gDRimport::get_exception_data("two"),
    "Assertion on 'status_code' failed: Must be of type 'number'",
    fixed = TRUE
  )

  expect_error(
    gDRimport::get_exception_data(0),
    paste0(
      "Assertion on \\'toString\\(status_code\\)\\' failed:.+but is \\'0\\'"
    )
  )
})

#' @importFrom utils adist
#' @importFrom magrittr %>%
#' @importFrom tibble tribble
#' @importFrom methods new slotNames
#' @importFrom data.table := setDF
NULL

DICTIONARY <- list(
  Untreated = "untreated",
  Vehicle = "vehicle",
  vehcle = "vehicle"
)

# Prevent R CMD check from complaining about the use of pipe expressions
# standard data.table variables
if (getRversion() >= "2.15.1") {
  utils::globalVariables(
    c(
      "..with",
      ".N",
      ".SD",
      "Dose",
      "column_name",
      "name",
      "new_name",
      "value",
      "ccle_name",
      ".",
      "CCLEName",
      "OncotreeLineage",
      "row_name",
      "ReadoutValue",
      "Elapsed",
      "BackgroundValue",
      "WellColumn",
      "pert_iname",
      "pert_dose",
      "pert2_iname",
      "pert2_dose"
    ),
    utils::packageName())
}

#' Get Excel sheets
#'
#' get sheets for given set of XLS files
#'
#' @param files charvec with file paths
#' @keywords correction_exception
#'
#' @return named list where names are the excel filenames
#' and the values are the sheets within each file
#'
get_xl_sheets <- function(files) {
  checkmate::assert_character(files)
  template_sheets <- lapply(files, readxl::excel_sheets)
  names(template_sheets) <- files
  template_sheets
}


#' Evaluate if template file with single sheet is present,
#' if the name of the sheet is correct and if it can be fixed
#'
#' get sheets for given set of XLS files
#' @param ts list with template sheets info
#' @keywords correction_exception
#'
#' @return logical flag
#'
.check_against_single_template_sheet <- function(ts) {

  checkmate::assert_list(ts)
  # edge case: 'untreated' template with (1) single sheet
  # and (2) improperly named (not 'idfs[['drug']])
  myv <- vapply(ts, function(x) {
    length(x) == 1 && x != gDRutils::get_env_identifiers("drug")
  }, logical(1))

  # one template file with
  status <- any(myv) && !all(myv)
  if (isTRUE(status)) {
    attributes(status) <- list(file = names(myv[myv]))
  }
  status
}


### XLS templates validation

#' Correct names of the template sheets (if required)
#'
#' Correct names of the template sheets (if required)
#'
#' @param tfiles charvec with paths to template files
#' @keywords correction_exception
#'
#' @return charvec with paths to corrected sheet names
correct_template_sheets <- function(tfiles) {

  checkmate::assert_character(tfiles)

  ts <- get_xl_sheets(tfiles)
  # no issues with templates sheets, return input data
  if (!are_template_sheets_valid(ts)) {
    # iterate through available fixes to try to fix data
    # spaces/bad capitalization
    ts <- fix_typos_with_reference(ts, unlist(gDRutils::get_env_identifiers()))

    # additional prefixes/postfixes in the optional sheets (e.g. "Concentration_2_3", "my_gnumber_2)"
    ts <- fix_typos_with_reference(ts, get_expected_template_sheets("optional"), method = "grepl")

    # additional prefixes/postfixes in the optional sheets and/or missing underscores
    ts <-
      fix_typos_with_reference(
        ts,
        get_expected_template_sheets("optional"),
        method = "grepl",
        fix_underscores = TRUE
      )
    # check one letter typos
    ts <- fix_typos_with_reference(
      ts,
      unlist(gDRutils::get_env_identifiers(gDRutils::get_required_identifiers(), simplify = FALSE)),
      method = "adist")

    # check if there is a template file with single, improperly named sheet
    # fix if possible
    st1 <- .check_against_single_template_sheet(ts)
    if (st1) {
      inv_file <- attributes(st1)[["file"]]
      ts[[inv_file]] <- gDRutils::get_env_identifiers("drug")
    }
  }
  stopifnot(are_template_sheets_valid(ts))
  ts
}

#' Get names of the sheets expected in templates xlsx
#'
#' Get names of the sheets expected in templates xlsx
#'
#' @param type charvec type of the sheets
#' @keywords correction_exception
#'
#' @return string with type of the sheets
#'
get_expected_template_sheets <-
  function(type = c("all", "core", "optional")) {

    type <- match.arg(type)

    ctype <- as.character(gDRutils::get_env_identifiers(c("drug",
                                                          "concentration"),
                                                        simplify = FALSE))
    otype <- as.character(gDRutils::get_env_identifiers(
      c("drug2",
        "drug3",
        "concentration2",
        "concentration3"),
      simplify = FALSE
    ))

    if (type == "all") {
      c(ctype, otype)
    } else if (type == "core") {
      ctype
    } else if (type == "optional") {
      otype
    } else {
      stop(sprintf("bad type: '%s", type))
    }
  }

#' are template sheet valid?
#'
#' are template sheet valid?
#'
#' @param ts list with (per file) template sheets
#' @keywords correction_exception
#'
#' @seealso get_xl_sheets
#'
#' @return logical flag
are_template_sheets_valid <- function(ts) {

  checkmate::assert_list(ts)

  cl <- vector("list", 10)
  drug <- gDRutils::get_env_identifiers("drug")

  # only allowed sheet names are present
  myv <- vapply(ts, function(x) {
    all(drug %in% x)
  }, logical(1))
  if (sum(!myv) == 1 && length(myv) > 1) { # allow for having only `Drug` sheet in case of untreated template
    myv[!myv] <- drug %in% ts[!myv][[1]]
  }
  cl[[length(cl) + 1]] <- all(myv)

 # at least idfs[['drug']] sheet is present in all files
  myv2 <- vapply(ts, function(x) {
    any(gDRutils::get_env_identifiers("drug") %in% x)
  }, logical(1))
  cl[[length(cl) + 1]] <- all(myv2)

 all(unlist(cl))
}


### TYPOS ###

#' Fix typos using reference data
#'
#' Fix typos using reference data
#' Evaluate given list of ids and try to update them
#  if they are similar to any id from 'ref' data
#'
#' @param data list of charvec(s) or charvec with data
#' @param ref charvec with reference data
#' @param method charvec type of the method to be used
#' 'exact' is used to find  identical entries from 'ref' in the
#' data (after corrections and uppercase'ing)
#' 'grepl' is used to find entries from 'ref' that might be
#' somehow pre- or post- fixed
#' @param fix_underscores logical flag fix the issues with underscores in data identfiers?
#' @keywords correction_exception
#'
#' @return list or charvec with corrected data
#'
fix_typos_with_reference <-
  function(data,
           ref,
           method = c("exact", "grepl", "adist"),
           fix_underscores = FALSE) {

    stopifnot(is.list(data) || is.character(data))
    checkmate::assert_character(ref)
    method <- match.arg(method)
    checkmate::assert_flag(fix_underscores)

    out <- if (is.list(data)) {
      lapply(data, function(x) {
        fix_typos_with_reference(x, ref, method, fix_underscores)
      })
    } else {
      # remove spaces at the beginning/end
      cdata <- vapply(data, function(x) gsub("^ +| +$", "", x), character(1))

      # replace spaces with "_"
      cdata <- vapply(cdata, function(x) gsub(" +", "_", x), character(1))

      # update to valid identifier (if found)
      # convert both v and valid identifiers to upper case before comparison
      # remove "_" from references if 'fix_underscores' enabled
      ref_uc <- if (fix_underscores) {
        gsub("_", "", toupper(ref))
      } else {
        toupper(ref)
      }

      cdata <- vapply(cdata, function(x) {
        idx <- if (method == "exact") {
          which(ref_uc %in% toupper(x))
        } else if (method == "adist") {
          which(adist(toupper(x), ref_uc) == 1)
        } else {
          which(mgrepl(ref_uc, toupper(x)))
        }
        if (length(idx) == 1) {
          ref[idx]
        } else {
          x
        }
      }, character(1))
      cdata
    }
    out
  }

#' grep wrapper to support multiple patterns
#'
#' @param patterns charvec with patterns to be checked
#' @param x charvec with data
#' @param do_unlist logical_flag unlist the final results?
#' @param ... additional argument
#' @keywords correction_exception
#'
#' @return list of charvec with grep output
#'
mgrepl <- function(patterns, x, do_unlist = TRUE,  ...) {
  checkmate::assert_character(patterns)
  checkmate::assert_character(x)
  checkmate::assert_flag(do_unlist)

  out <- lapply(patterns, function(p) {
    grepl(p, x, ...)
  })
  if (do_unlist) {
    unlist(out)
  }
}

context("correction")

test_that("fix_typos_with_reference", {

a <- c("GNUMBER 2", "Concentration  3")
b <- c("gnumber_2_3", "my_concentration_2")
c <- c("gnumber2", "my_concentration3")
d <- c(" gnumber_2", "GNUmber_3", " Concentration ", "concentration_3 ")

aobs <- as.character(fix_typos_with_reference(a, ref = get_expected_template_sheets("optional")))
aref <- as.character(gDRutils::get_env_identifiers(c("drug2", "concentration3"), FALSE))
expect_equal(aref, aobs)

bobs <- as.character(fix_typos_with_reference(b, ref = get_expected_template_sheets("optional"), method = "grepl"))
bref <- as.character(gDRutils::get_env_identifiers(c("drug2", "concentration2"), FALSE))

cobs <-
  as.character(
    fix_typos_with_reference(
      c,
      ref = get_expected_template_sheets("optional"),
      method = "grepl",
      fix_underscores = TRUE
    )
  )
cref <- as.character(gDRutils::get_env_identifiers(c("drug2", "concentration3"), FALSE))
expect_equal(cref, cobs)

dobs <- as.character(fix_typos_with_reference(d, ref = get_expected_template_sheets()))
dref <- as.character(gDRutils::get_env_identifiers(c("drug2", "drug3", "concentration", "concentration3"), FALSE))
expect_equal(dref, dobs)
})

test_that("fix_typos_with_reference", {

a <- c("GNUMBER 2", "Concentration  3")
b <- c("gnumber_2_3", "my_concentration_2")
c <- c("gnumber2", "my_concentration3")
d <- c(" gnumber_2", "GNUmber_3", " Concentration ", "concentration_3 ")
e <- c("Gnomber", "Concuntration")

aobs <- as.character(fix_typos_with_reference(a, ref = get_expected_template_sheets("optional")))
aref <- as.character(gDRutils::get_env_identifiers(c("drug2", "concentration3"), FALSE))
expect_equal(aref, aobs)

bobs <- as.character(fix_typos_with_reference(b, ref = get_expected_template_sheets("optional"), method = "grepl"))
bref <- as.character(gDRutils::get_env_identifiers(c("drug2", "concentration2"), FALSE))

cobs <-
  as.character(
    fix_typos_with_reference(
      c,
      ref = get_expected_template_sheets("optional"),
      method = "grepl",
      fix_underscores = TRUE
    )
  )
cref <- as.character(gDRutils::get_env_identifiers(c("drug2", "concentration3"), FALSE))
expect_equal(cref, cobs)

dobs <- as.character(fix_typos_with_reference(d, ref = get_expected_template_sheets()))
dref <- as.character(gDRutils::get_env_identifiers(c("drug2", "drug3", "concentration", "concentration3"), FALSE))
expect_equal(dref, dobs)

eobs <- as.character(fix_typos_with_reference(e, ref = get_expected_template_sheets(), method = "adist"))
eref <- as.character(gDRutils::get_env_identifiers(c("drug", "concentration"), FALSE))
expect_equal(eref, eobs)
})


test_that("correct_template_sheets works as expected", {
  tfiles <- list("template1.xlsx" = c("Gnumber", "Concentration", "Media"),
                 "template2.xlsx" = c("Gnomber", "Concentration"),
                 "template3.xlsx" = c("Gnomber", "Concentration", "Gnumber_2", "Concentration_2"),
                 "template_untr.xlsx" = c("Gnumber", "Media"))
  tfiles_empty <- lapply(tfiles, function(x) {
    listData <- lapply(x, function(y) data.table::data.table(x = runif(12)))
    names(listData) <- x
    listData
  })
  lapply(names(tfiles_empty), function(x) {
    openxlsx::write.xlsx(tfiles_empty[[x]], file.path(tempdir(), x))
  })
  correctedList <- correct_template_sheets(file.path(tempdir(), names(tfiles)))

  tfilesCorrect <- list("template1.xlsx" = c("Gnumber", "Concentration", "Media"),
                 "template2.xlsx" = c("Gnumber", "Concentration"),
                 "template3.xlsx" = c("Gnumber", "Concentration", "Gnumber_2", "Concentration_2"),
                 "template_untr.xlsx" = c("Gnumber", "Media"))
  testthat::expect_identical(lapply(unname(correctedList), unname), unname(tfilesCorrect))
})

##

# collaboration with BHKLab
# co-author: Jermiah Joseph
library(PharmacoGx)
library(gDR)
library(gDRcore)
library(gDRimport)
pset <- getPSet("CCLE_2015")
pset2 <- getPSet("FIMM_2016")
pset3 <- getPSet("NCI_60")
pset4 <- getPSet("PRISM_2020")

dt <- gDRimport::convert_pset_to_df(pharmacoset = pset,
                        run_parallel=TRUE,
                        workers = 4L)



# example subset to speed things up. Specific to CCLE Pset
x <- dt[Clid=="1321N1"]
x
# x$Barcode |> unique() |> length()
# x[,1:7] |> is.na() |> any()

#############################
# RUN DRUG RESPONSE PROCESSING PIPELINE
se <- gDRcore::runDrugResponseProcessingPipeline(x)
gDRutils::convert_se_assay_to_dt(se[[1]],"RawTreated")
gDRutils::convert_se_assay_to_dt(se[[1]],"Controls")

##########################################################
# RUNNING EACH STEP SEPARATELY IF NEEDED.
##########################################################

#############################
# CREATE SUMMARIZED EXPERIMENT
se2 <- gDRcore::create_SE(x,  data_type = "single-agent")
gDRutils::convert_se_assay_to_dt(se2, "RawTreated")
gDRutils::convert_se_assay_to_dt(se2, "Controls")

#############################
# NORMALIZE
se2_norm <- gDRcore::normalize_SE(se = se2, data_type = "single-agent")
gDRutils::convert_se_assay_to_dt(se2_norm, "Normalized")

#############################
# AVERAGE
se2_avg <- gDRcore::average_SE(se2_norm, data_type = "single-agent")
gDRutils::convert_se_assay_to_dt(se2_avg, "Averaged")

#############################
# FIT
se2_fit <- gDRcore::fit_SE(se2_avg, data_type = "single-agent")
gDRutils::convert_se_assay_to_dt(se2_fit, "Metrics")

#' Load data
#'
#' This functions loads and checks the data file(s)
#'
#' @param manifest_file character, file path(s) to manifest(s)
#' @param df_template_files data.table, with datapaths and names of results file(s)
#' or character with file path of templates file(s)
#' @param results_file  data.table, with datapaths and names of results file(s)
#' or character with file path of results file(s)
#' @param instrument character
#' @keywords load_files
#'
#' @examples
#'  td <- get_test_data()
#'  l_tbl <- load_data(manifest_path(td), template_path(td), result_path(td))
#'
#' @export
#'
#' @return a list with three data.tables for manifest/treatment and results
#'
load_data <-
  function(manifest_file,
           df_template_files,
           results_file,
           instrument = "EnVision") {

    assertthat::assert_that(is.character(manifest_file), msg = "'manifest_file' must be a character vector")
    assertthat::assert_that(assertthat::is.readable(manifest_file), msg = "'manifest_file' must be a readable path")
    assertthat::assert_that(is.character(df_template_files) || data.table::is.data.table(df_template_files),
                msg = "'df_template_files' must be a character vector or data.table")
    assertthat::assert_that(is.character(results_file) || data.table::is.data.table(results_file),
                msg = "'results_file' must be a character vector or data.table")
    assertthat::assert_that(assertthat::is.string(instrument), msg = "'instrument' must be a character vector")

    if (data.table::is.data.table(df_template_files)) {
      template_file <- df_template_files$datapath # for the shiny app
      is_readable_v(template_file)
      template_filename <- df_template_files$name
    } else {
      template_filename <- df_template_files
      is_readable_v(template_filename)
    }

    manifest_list <- load_manifest(manifest_file)
    manifest <- manifest_list$data
    headers <- manifest_list$headers
    treatments <- load_templates(df_template_files)
    data <- load_results(results_file, instrument, headers = headers)

    # check the all template files are available
    if (!all(unique(manifest[[headers[["template"]]]][manifest[[headers[["barcode"]]]] %in%
                                                      data[[headers[["barcode"]]]]])
             %in% basename(template_filename))) {
      exception_data <- get_exception_data(11)
      stop(sprintf(
        exception_data$sprintf_text,
        toString(setdiff(
          unique(manifest[[headers[["template"]]]][manifest[[headers[["barcode"]]]] %in%
                                                     data[[headers[["barcode"]]]]]),
          basename(template_filename)
        ))
      ))
    }
    return(list(
      manifest = manifest,
      treatments = treatments,
      data = data
    ))
  }


#' Load manifest
#'
#' This functions loads and checks the manifest file(s)
#'
#' @param manifest_file character, file path(s) to manifest(s)
#' @keywords load_files
#'
#' @examples
#'  td <- get_test_data()
#'  ml <- load_manifest(manifest_path(td))
#'
#' @export
#'
#' @return list with manifest data.table and headers
#'
load_manifest <- function(manifest_file) {
  # manifest_file is a string or a vector of strings
  available_formats <- c("text/tsv",
                         "text/tab-separated-values",
                         "xlsx",
                         "tsv")
  assertthat::assert_that(is.character(manifest_file), msg = "'manifest_file' must be a character vector")
  checkmate::assert_file_exists(manifest_file)


  manifest_data <- read_in_manifest_file(manifest_file, available_formats)
  headers <- gDRutils::validate_identifiers(do.call(rbind, manifest_data), req_ids = "barcode")

  # Set character type to barcode
  manifest_data <- lapply(manifest_data, function(x) {
    x[, (headers[["barcode"]]) := lapply(.SD, as.character), .SDcols = headers[["barcode"]]]
  })


  # check default headers are in each df
  dump <- lapply(seq_along(manifest_file),
                 function(i) {
                   check_metadata_names(
                     colnames(manifest_data[[i]]),
                     df_name = manifest_file[[i]],
                     df_type = "manifest"
                   )})

  cat_manifest_data <- data.table::rbindlist(manifest_data)
  colnames(cat_manifest_data) <-
    check_metadata_names(colnames(cat_manifest_data), "manifest")

  # check that barcodes are unique
  exception_data <- get_exception_data(14)
  if (dim(cat_manifest_data)[1] != length(unique(cat_manifest_data[[headers[["barcode"]]]])))
    stop(exception_data$sprintf_text)

  cat_manifest_data[[headers[["template"]]]] <- basename(cat_manifest_data[[headers[["template"]]]])

  futile.logger::flog.info("Manifest loaded successfully")
  return(list(data = cat_manifest_data,
              headers = headers))
}

#' read manifest files
#'
#' @param manifest_file character, file path(s) to manifest(s)
#' @param available_formats charvec with available file formats
#' @keywords load_files
#'
#' @return a data.table with manifest data
#'
read_in_manifest_file <- function(manifest_file, available_formats) {

  manifest_data <- lapply(manifest_file, function(x) {
    manifest_ext <- tools::file_ext(x)
    if (manifest_ext == "xlsx") {
      df <- tryCatch({
        read_excel_to_dt(x, col_names = TRUE)
      }, error = function(e) {
        exception_data <- get_exception_data(12)
        stop(sprintf(
          exception_data$sprintf_text,
          e
        ))
      })
    } else if (manifest_ext %in% c("text/tsv",
                                   "text/tab-separated-values",
                                   "tsv")) {

      df <- tryCatch({
        stats::na.omit(data.table::fread(
          x, sep = "\t", header = TRUE, na.strings = c("", "NA")))
      }, error = function(e) {
        exception_data <- get_exception_data(12)
        stop(sprintf(
          exception_data$sprintf_text,
          e
        ))
      })
    } else {
      exception_data <- get_exception_data(13)
      stop(
        sprintf(
          exception_data$sprintf_text,
          manifest_ext,
          stringi::stri_flatten(available_formats, collapse = ", ")
        )
      )
    }
  })

  # replace synonyms, e.g. 'time => Duration' (backwards compatibility)
  manifest_data <- lapply(manifest_data, function(x) {
    colnames(x) <- gDRutils::update_idfs_synonyms(colnames(x))
    x
  })
  manifest_data
}

#' Load templates
#'
#' This functions loads and checks the template file(s)
#'
#' @param df_template_files data.table, with datapaths and names of results file(s)
#' or character with file path of templates file(s)
#' @keywords load_files
#'
#' @examples
#'  td <- get_test_data()
#'  t_df <- load_templates(template_path(td))
#'
#' @export
#'
#' @return data.table with templates data
#'
load_templates <- function(df_template_files) {
  # Assertions:
  stopifnot(any(inherits(df_template_files, "data.table"), checkmate::test_character(df_template_files)))

  # template_file is a string or a vector of strings
  if (data.table::is.data.table(df_template_files)) {
    # for the shiny app
    template_file <- df_template_files$datapath
    template_filename <- df_template_files$name
  } else {
    template_file <- df_template_files
    template_filename <- basename(template_file)
  }

  checkmate::assert_file_exists(template_file)

  all_templates <- data.table::data.table()
  if (any(grepl("\\.xlsx?$", template_filename))) {
    idx <- grepl("\\.xlsx?$", template_filename)
    futile.logger::flog.info("Reading %s with load_templates_xlsx", template_filename[idx])
    all_templates_1 <-
      load_templates_xlsx(template_file[idx], template_filename[idx])
    all_templates <- rbind(all_templates, all_templates_1)
  }
  if (any(grepl("\\.[ct]sv$", template_filename))) {
    idx <- grepl("\\.[ct]sv$", template_filename)
    futile.logger::flog.info("Reading %s with load_templates_tsv", template_filename[idx])
    all_templates_2 <-
      load_templates_tsv(template_file[idx], template_filename[idx])
    all_templates <- rbind(all_templates, all_templates_2)
  }

  all_templates[[gDRutils::get_env_identifiers("drug")]] <-
    standardize_record_values(all_templates[[gDRutils::get_env_identifiers("drug")]], dictionary = DICTIONARY)

  if (all(!gDRutils::get_env_identifiers("untreated_tag") %in% unlist(all_templates))) {
    exception_data <- get_exception_data(15)
    stop(exception_data$sprintf_text)
  }

  return(all_templates)

}


#' Load results
#'
#' This functions loads and checks the results file(s)
#'
#' @param df_results_files  data.table, with datapaths and names of results file(s)
#' or character with file path of results file(s)
#' @param instrument character
#' @param headers list of headers identified in the manifest file
#' @keywords load_files
#'
#' @examples
#'  td <- get_test_data()
#'  r_df <- load_results(result_path(td))
#'
#' @export
#'
#' @return data.table with results' data
#'
load_results <-
  function(df_results_files, instrument = "EnVision", headers = gDRutils::get_env_identifiers()) {

    stopifnot(any(inherits(df_results_files, "data.table"), checkmate::test_character(df_results_files)))
    checkmate::assert_string(instrument, pattern = "^EnVision$|^long_tsv$|^Tecan$|^EnVision_new$|^Incucyte$")
    checkmate::assert_list(headers, null.ok = TRUE)

    if (data.table::is.data.table(df_results_files)) {
      # for the shiny app
      results_file <- df_results_files$datapath
      results_filename <- df_results_files$name
    } else {
      results_file <- df_results_files
      results_filename <- basename(results_file)
    }
    checkmate::assert_file_exists(results_file)

    if (all(endsWith(results_file, ".tsv"))) {
      instrument <- "long_tsv"
    }

    if (instrument == "EnVision") {
      all_results <-
        load_results_EnVision(results_file, headers = headers)
    } else if (instrument == "long_tsv") {
      all_results <-
        load_results_tsv(results_file, headers = headers)
    } else if (instrument == "Tecan") {
      all_results <-
        load_results_Tecan(results_file, headers = headers)
    } else if (instrument == "Incucyte") {
      all_results <-
        load_results_Incucyte(results_file, headers = headers)
    } else if (instrument == "EnVision_new") {
      all_results <-
        load_results_EnVision_new(results_file, headers = headers)
    } else {
      exception_data <- get_exception_data(16)
      stop(exception_data$sprintf_text)
    }
    return(unique(all_results))
  }


# individual functions

#' Load templates from tsv
#'
#' This functions loads and checks the template file(s)
#'
#' @param template_file character, file path(s) to template(s)
#' @param template_filename character, file name(s)
#' @keywords load_files
#'
#' @return data.table with template data
#'
load_templates_tsv <-
  function(template_file,
           template_filename = NULL) {
    # Assertions:
    checkmate::assert_character(template_file)
    checkmate::assert_character(template_filename, null.ok = TRUE)


    if (is.null(template_filename))
      template_filename <- basename(template_file)

    # read columns in files
    templates <- lapply(template_file, function(x) {
      stats::na.omit(data.table::fread(x,
                                       sep = "\t",
                                       header = TRUE,
                                       na.strings = c("", "NA")))
    })
    names(templates) <- template_filename
    # check WellRow/WellColumn is present in each df
    dump <- lapply(seq_along(template_file),
                   function(i) {
                     if (!(all(
                       gDRutils::get_env_identifiers("well_position") %in% colnames(templates[[i]])
                     ))) {
                       futile.logger::flog.info("%s missing, %s as header",
                                                template_filename[[i]],
                                                gDRutils::get_env_identifiers("well_position"))
                     }})
    # check drug_identifier is present in each df
    dump <- lapply(seq_along(template_file),
                   function(i) {
                     check_metadata_names(
                       setdiff(colnames(templates[[i]]), gDRutils::get_env_identifiers("well_position")),
                       df_name = template_filename[[i]],
                       df_type = "template"
                     )})

    all_templates <- read_in_tsv_template_files(template_file, template_filename, templates)
    futile.logger::flog.info("Templates loaded successfully!")
    all_templates
  }

#' read in tsv template files
#'
#' @param template_file character, file path(s) to template(s)
#' @param template_filename character, file name(s)
#' @param templates list with templates data
#' @keywords load_files
#'
#' @return data.table with templates data
#'
read_in_tsv_template_files <- function(template_file, template_filename, templates) {
    metadata_fields <- NULL
    tmpl_l <- lapply(seq_along(template_file), function(iF) {
      futile.logger::flog.info("Loading %s", template_filename[iF])
      # 1) check that the sheet names are ok and 2) identify drug_identifier sheet (case insensitive)
      Gnumber_idx <- grep(paste0(gDRutils::get_env_identifiers("drug"), "$"),
                          colnames(templates[[iF]]), ignore.case = TRUE)
      Conc_idx <-
        grepl("Concentration", colnames(templates[[iF]]), ignore.case = TRUE)
      # case of untreated plate
      if (sum(Conc_idx) == 0) {
        if (length(Gnumber_idx) == 0) {
          exception_data <- get_exception_data(17)
          stop(sprintf(exception_data$sprintf_text, template_file[[iF]], gDRutils::get_env_identifiers("drug")))
        }
        df <- templates[[iF]][, gDRutils::get_env_identifiers("drug")]
        if (!(all(toupper(df)[!is.na(df)]) %in% toupper(gDRutils::get_env_identifiers("untreated_tag")))) {
          exception_data <- get_exception_data(18)
          stop(sprintf(exception_data$sprintf_text, template_file[[iF]],
            paste(gDRutils::get_env_identifiers("untreated_tag"), collapse = " or ")))
        }
      } else {
        # normal case
        check_metadata_names(colnames(templates[[iF]]),
                             df_name = template_filename[iF],
                             df_type = "template_treatment")
      }
      df_template <- templates[[iF]]
      for (iS in colnames(df_template)) {
        # check if metadata field already exist and correct capitalization if needed
        if (!(iS %in% metadata_fields)) {
          if (!is.null(metadata_fields) &&
              toupper(iS) %in% toupper(metadata_fields)) {
            oldiS <- iS
            iS <-
              metadata_fields[toupper(iS) == toupper(metadata_fields)]
            futile.logger::flog.info("%s corrected to match case with %s", oldiS, iS)
            colnames(df_template)[colnames(df_template) == oldiS] <-
              iS
          } else {
            metadata_fields <- c(metadata_fields, iS)
          }
        }
      }
      df_template$Template <- template_filename[iF]
      colnames(df_template) <-
        check_metadata_names(colnames(df_template), df_name = template_filename[iF])
      df_template
    })
    do.call(rbind, tmpl_l)
}

#' Load templates from xlsx
#'
#' This functions loads and checks the template file(s)
#'
#' @param template_file character, file path(s) to template(s)
#' @param template_filename character, file name(s)
#' @keywords load_files
#'
#' @return data.table with templates data
#'
load_templates_xlsx <-
  function(template_file,
           template_filename = NULL) {
    # Assertions:
    checkmate::assert_character(template_file)
    checkmate::assert_character(template_filename, null.ok = TRUE)

    if (is.null(template_filename))
      template_filename <- basename(template_file)

    # validate template sheets
    template_sheets <- correct_template_sheets(template_file)
    # check drug_identifier is present in each df
    dump <- lapply(seq_along(template_file),
                   function(i) {
                     check_metadata_names(
                       template_sheets[[i]],
                       df_name = template_file[[i]],
                       df_type = "template"
                     )})

    all_templates <- read_in_template_xlsx(template_file, template_filename, template_sheets)

    # standardize untreated values
    all_templates <- .standardize_untreated_values(all_templates)
    futile.logger::flog.info("Templates loaded successfully!")
    return(all_templates)
  }

#' Read in xlsx template files
#'
#' @param template_file character, file path(s) to template(s)
#' @param template_filename character, file name(s)
#' @param template_sheets template sheet names
#' @keywords load_files
#'
#' @return data.table with templates data
#'
read_in_template_xlsx <- function(template_file, template_filename, template_sheets) {

    metadata_fields <- NULL
    all_templates <- data.table::data.table()
    for (iF in seq_along(template_file)) {
      futile.logger::flog.info("Loading %s", template_filename[iF])
      Gnumber_idx <- which(template_sheets[[iF]] %in% gDRutils::get_env_identifiers("drug"))

      # first check that the sheet names are ok
      # identify drug_identifier sheet
      validate_template_xlsx(template_file, template_filename, template_sheets, iF)

      # read the different sheets and check for plate size
      # enforce range to avoid skipping empty rows at the beginning
      plate_info <- get_plate_info_from_template_xlsx(template_file, Gnumber_idx, iF)
      df_template <-
        read_in_template_sheet_xlsx(template_file,
                                    template_sheets,
                                    iF,
                                    plate_info)
      df_template$Template <- template_filename[iF]
      colnames(df_template) <-
        check_metadata_names(colnames(df_template),
                             df_name = template_filename[iF])

      all_templates <- data.table::rbindlist(list(all_templates, df_template), fill = TRUE)

    }
    all_templates
}

#' Read in data from xlsx template sheet
#'
#' @param template_file character, file path(s) to template(s)
#' @param template_sheets template sheet names
#' @param idx template file index
#' @param plate_info list with plate info
#' @keywords load_files
#'
#' @return data.table with template data
#'
read_in_template_sheet_xlsx <- function(template_file, template_sheets, idx, plate_info) {
  metadata_fields <- NULL
  # need to adapt for 1536 well plates
  df_template <- data.table::CJ(WellRow = LETTERS[seq_len(plate_info$n_row)],
                                WellColumn = as.character(seq_len(plate_info$n_col)))
  for (iS in seq_along(template_sheets[[idx]])) {
    sheetName <- template_sheets[[idx]][[iS]]
    tryCatch({
      df <- read_excel_to_dt(
          template_file[[idx]],
          sheet = iS,
          col_names = paste0("x", seq_len(plate_info$n_col)),
          range = plate_info$plate_range
        )
    }, error = function(e) {
      exception_data <- get_exception_data(20)
      stop(sprintf(exception_data$sprintf_text,
                   template_file[[idx]],
                   e))
    })
    df$WellRow <- LETTERS[seq_len(plate_info$n_row)]
    nonCharvecCols <- df[, colnames(.SD), .SDcols = !is.character]
    df[, (nonCharvecCols) := lapply(.SD, as.character), .SDcols = nonCharvecCols]
    df_melted <- data.table::melt(df, id.vars = "WellRow")
    # check if metadata field already exist and correct capitalization if needed
    if (!(sheetName %in% metadata_fields)) {
      if (!is.null(metadata_fields) &&
          toupper(sheetName) %in% toupper(metadata_fields)) {
        oldiS <- sheetName
        sheetName <-
          metadata_fields[toupper(sheetName) == toupper(metadata_fields)]
        futile.logger::flog.info("%s corrected to match case with %s", oldiS, sheetName)
      } else {
        metadata_fields <- c(metadata_fields, sheetName)
      }
    }
    colnames(df_melted)[3] <- sheetName
    colnames(df_melted)[colnames(df_melted) == "variable"] <-
      "WellColumn"
    df_melted$WellColumn <-
      gsub("x", "", df_melted$WellColumn)
    df_template <-
      df_template[df_melted, on = c("WellRow", "WellColumn")]
  }
  data.table::setorder(df_template)
  df_template
}

#' Get plate info from template xlsx
#'
#' @param template_file character, file path(s) to template(s)
#' @param Gnumber_idx index with Gnumber data
#' @param idx template file index
#' @keywords load_files
#'
#' @return list with plate info
#'
get_plate_info_from_template_xlsx <- function(template_file, Gnumber_idx, idx) {

      tryCatch({
        df <-
          read_excel_to_dt(
            template_file[[idx]],
            sheet = Gnumber_idx,
            col_names = paste0("x", seq_len(48)),
            range = "A1:AV32", # 32x48 plate (largest plate size)
            col_types = "text"
          )
      }, error = function(e) {
        exception_data <- get_exception_data(5)
        stop(sprintf(exception_data$sprintf_text, e))
      })

      # get the plate size
      n_row <-
        2 ^ ceiling(log2(max(which(
          apply(!is.na(df), 1, any)
        ))))
      n_col <-
        1.5 * 2 ^ ceiling(log2(max(which(
          apply(!is.na(df), 2, any)
        )) / 1.5))
      n_row <- max(n_row, n_col / 1.5)
      n_col <- max(1.5 * n_row, n_col)
      plate_range_val <- if (n_col < 26) {
        paste0("A1:", LETTERS[n_col], n_row)
      } else {
        "A1:AV32"
      }
      list(
        plate_range = plate_range_val,
        n_row = n_row,
        n_col = n_col
      )
}

#' Validate template xlsx data
#'
#' @param template_file character, file path(s) to template(s)
#' @param template_filename character, file name(s)
#' @param template_sheets template sheet names
#' @param idx template file index
#' @keywords load_files
#'
#' @return \code{NULL} invisibly.
#'
validate_template_xlsx <- function(template_file, template_filename, template_sheets, idx) {

      # first check that the sheet names are ok
      # identify drug_identifier sheet
      Gnumber_idx <- which(template_sheets[[idx]] %in% gDRutils::get_env_identifiers("drug"))
      Conc_idx <-
        grepl("Concentration", template_sheets[[idx]], ignore.case = TRUE)
      # case of untreated plate
      if (sum(Conc_idx) == 0) {
        if (length(Gnumber_idx) == 0) {
          exception_data <- get_exception_data(17)
          stop(sprintf(
            exception_data$sprintf_text,
            template_file[[idx]],
            gDRutils::get_env_identifiers("drug")
          ))
        }
        tryCatch({
          df <-
            read_excel_to_dt(
              template_file[[idx]],
              sheet = Gnumber_idx,
              col_names = paste0("x", seq_len(48)),
              range = "A1:AV32"
            )
        }, error = function(e) {
          exception_data <- get_exception_data(5)
          stop(sprintf(exception_data$sprintf_text, e))
        })
        if (!(all(toupper(unlist(df)[!is.na(unlist(df))]) %in%
                  toupper(gDRutils::get_env_identifiers(
                    "untreated_tag"
                  ))))) {
          exception_data <- get_exception_data(18)
          stop(sprintf(
            exception_data$sprintf_text,
            template_file[[idx]],
            paste(gDRutils::get_env_identifiers("untreated_tag"), collapse = " or ")
          ))
        }
      } else {
        # normal case
        dump <- check_metadata_names(template_sheets[[idx]],
                                     df_name = template_filename[idx],
                                     df_type = "template_treatment")
      }
}

#' Load results from tsv
#'
#' This functions loads and checks the results file(s)
#'
#' @param results_file character, file path(s) to template(s)
#' @param headers list of headers identified in the manifest
#' @keywords load_files
#'
#' @return data.table with results data
#'
load_results_tsv <-
  function(results_file, headers) {

    # results_file is a string or a vector of strings
    results_filename <- basename(results_file)

    all_results <- read_in_result_files(results_file, results_filename, headers)

    cols_subset <- intersect(c(headers[["barcode"]], gDRutils::get_env_identifiers("well_position")),
                             names(all_results))

    if (dim(unique(all_results[, cols_subset, with = FALSE]))[1] !=
        dim(all_results[, cols_subset, with = FALSE])[1]) {
      futile.logger::flog.error("Multiple rows with the same Barcode and Well across all files")
    }

    return(all_results)
  }

#' Read in results files
#'
#' @param results_file  data.table, with datapaths and names of results file(s)
#' or character with file path of results file(s)
#' @param results_filename charvect with file names
#' @param headers list of headers identified in the result files
#' @keywords load_files
#'
#' @return data.table with results data
#'
read_in_result_files <- function(results_file, results_filename, headers) {

    # read all files
    res_l <- lapply(seq_along(results_file), function(iF) {
      futile.logger::flog.info("Reading file %s", results_file[iF])
      tryCatch({
        df <-
        stats::na.omit(data.table::fread(
          results_file[iF], sep = "\t", header = TRUE, na.strings = c("", "NA")))
      }, error = function(e) {
        exception_data <- get_exception_data(21)
        stop(sprintf(exception_data$sprintf_text, results_file[[iF]]))
      })
      # skip_empty_rows flag needs to be TRUE even if it ends up not skipping empty rows
      if (NCOL(df) == 1) {
        tryCatch({
          # likely a csv file
          df <-
            stats::na.omit(data.table::fread(
              results_file[iF], header = TRUE, na.strings = c("", "NA")))
        }, error = function(e) {
          exception_data <- get_exception_data(21)
          stop(sprintf(exception_data$sprintf_text, results_file[[iF]]))
        })
      }
      for (coln in c(headers[["barcode"]],
                     gDRutils::get_env_identifiers("well_position"),
                     "ReadoutValue")) {
        if (!(coln %in% colnames(df))) {
          futile.logger::flog.error("%s needs to be a column of %s", coln, results_filename[iF])
        }
      }

      cols_subset <- intersect(c(headers[["barcode"]], gDRutils::get_env_identifiers("well_position")),
                               names(df))
      if (dim(unique(df[, cols_subset, with = FALSE]))[1] !=
          dim(df[, cols_subset, with = FALSE])[1]) {
        futile.logger::flog.error("Multiple rows with the same Barcode and Well in %s",
                                  results_filename[iF])
      }
      if (!("BackgroundValue" %in% colnames(df))) {
        df$BackgroundValue <- 0
        futile.logger::flog.info("File %s read; %d wells", results_filename[iF], NROW(df))
        futile.logger::flog.info("File done")
      }
      df
    })
    all_results <- do.call(rbind, res_l)
}

#' Load EnVision results from xlsx
#'
#' This functions loads and checks the results file(s)
#'
#' @param results_file character vector containing file path(s) to results file(s)
#' @param headers list of headers identified in the manifest
#' @keywords load_files
#'
#' @return data.table with results data
#'
load_results_EnVision <-
  function(results_file,
           headers = gDRutils::get_env_identifiers()) {
    checkmate::assert_character(results_file)

    # read all files and sheets
    results_sheets <- get_excel_sheet_names(results_file)
    all_results <- data.table::data.table()

    for (iF in seq_along(results_file)) {
      for (iS in results_sheets[[iF]]) {

        futile.logger::flog.info("Reading file %s, sheet %s", results_file[[iF]], iS)
        is_XLSX <- iS != 0L
        if (!is_XLSX) {
          fInfo <- read_EnVision_delim(results_file[[iF]])
          df <- fInfo$df
          isEdited <- fInfo$isEdited
        } else {
          isEdited <- TRUE
          df <- read_EnVision_xlsx(results_file[[iF]], iS)
        }

        barcode_col <- grep(paste(headers[["barcode"]], collapse = "|"), df)[1]

        if (isEdited) {
          # get the expected plate size
          plate_dim <- .get_plate_size(df)
          n_row <- plate_dim[1]
          n_col <- plate_dim[2]

          df <- enhance_raw_edited_EnVision_df(df, barcode_col, headers)
          barcode_idx <- which(unlist(df[, barcode_col, with = FALSE])
                               %in% headers[["barcode"]])
          df_results <-
            get_df_from_raw_edited_EnVision_df(df,
                                               barcode_idx,
                                               barcode_col,
                                               n_row,
                                               n_col,
                                               results_file[[iF]],
                                               iS,
                                               headers)
          all_results <- rbind(all_results, df_results)
        } else {
          # proper original EnVision file
          df_results <- get_df_from_raw_unedited_EnVision_df(df, fInfo$n_row, fInfo$n_col, barcode_col)
          all_results <- rbind(all_results, df_results)
        }
      }
      futile.logger::flog.info("File done")
    }

    all_results
  }

#' Load results from EnVision_new (CSV and XLSX)
#'
#' This functions loads and checks the results file(s) from a new Envision instrument
#' in the CSV or XLSX format. Supports multiple plates in a single file or multiple
#' sheets in an Excel file by robustly checking the file structure.
#'
#' @param results_file character, file path(s) to result file(s)
#' @param headers list of headers identified in the manifest
#' @keywords load_files
#'
#' @return data.table with results data
#'
load_results_EnVision_new <- function(results_file, headers = gDRutils::get_env_identifiers()) {

  checkmate::assert_character(results_file)
  checkmate::assert_list(headers, null.ok = TRUE)

  all_results <- data.table::data.table()

  for (iF in seq_along(results_file)) {
    current_file <- results_file[iF]
    is_excel <- grepl("\\.xlsx?$", current_file, ignore.case = TRUE)

    lines_list <- list()

    if (is_excel) {
      futile.logger::flog.info("Reading EnVision_new Excel file %s", current_file)
      sheets <- readxl::excel_sheets(current_file)

      for (sheet in sheets) {
        dt <- read_excel_to_dt(current_file, sheet = sheet, col_names = FALSE)
        dt_char <- dt[, lapply(.SD, function(x) {
          char_x <- as.character(x)
          char_x[is.na(char_x)] <- ""
          char_x
        })]
        lines_list[[sheet]] <- do.call(paste, c(dt_char, sep = ","))
      }
    } else {
      futile.logger::flog.info("Reading EnVision_new CSV file %s", current_file)
      lines_list[["csv"]] <- readLines(current_file, warn = FALSE)
    }

    for (sheet_name in names(lines_list)) {
      lines <- lines_list[[sheet_name]]

      if (is_excel) {
        futile.logger::flog.info("Processing sheet '%s' from %s", sheet_name, current_file)
      }

      clean_lines <- gsub("^\"|\"$", "", lines)
      clean_lines <- gsub("\"", "", clean_lines)

      data_header_idx <- grep("^[;, \t]*1[;, \t]+2[;, \t]+3", clean_lines)

      if (length(data_header_idx) == 0) {
        if (is_excel) {
          futile.logger::flog.warn("Could not find data matrix header in file: %s, sheet: '%s'. Skipping.",
                                   current_file, sheet_name)
          next
        } else {
          stop(sprintf("Could not find data matrix header (e.g., ';1;2;3...') in file: %s", current_file))
        }
      }

      for (idx in seq_along(data_header_idx)) {
        data_start_line <- data_header_idx[idx]

        if (data_start_line > 1 && grepl("Plate Map", clean_lines[data_start_line - 1], ignore.case = TRUE)) {
          futile.logger::flog.info("Skipping 'Plate Map' matrix at line %d in file '%s'", data_start_line, current_file)
          next
        }

        barcode <- NA
        search_limit <- max(1, data_start_line - 15)

        for (r in seq(data_start_line - 1, search_limit, by = -1)) {
          if (grepl("^Plate Barcode[;,]Loop", clean_lines[r], ignore.case = TRUE)) {
            barcode_line <- clean_lines[r + 1]
            barcode <- strsplit(barcode_line, ";|,")[[1]][1]
            break
          }
        }

        if (is.na(barcode) || barcode == "") {
          for (r in seq(data_start_line - 1, search_limit, by = -1)) {
            if (grepl("^Plate Barcode[;,]", clean_lines[r], ignore.case = TRUE)) {
              parts <- strsplit(clean_lines[r], ";|,")[[1]]
              vals <- parts[parts != "" & toupper(parts) != "PLATE BARCODE"]
              if (length(vals) > 0) {
                barcode <- vals[1]
                break
              }
            }
          }
        }

        if (is.na(barcode) || barcode == "") {
          futile.logger::flog.info("Skipping matrix at line %d in file '%s':
                                   no associated 'Plate Barcode' found within 15 lines.",
                                   data_start_line, current_file)
          next
        }

        n_rows <- 0
        for (r in (data_start_line + 1):length(clean_lines)) {
          if (grepl("^[A-Za-z]+[;,]", clean_lines[r])) {
            n_rows <- n_rows + 1
          } else {
            break
          }
        }

        if (n_rows == 0) n_rows <- 16

        data_lines <- lines[data_start_line:(data_start_line + n_rows)]

        tryCatch({
          raw_data <- data.table::fread(
            text = data_lines,
            header = TRUE,
            colClasses = "character",
            blank.lines.skip = FALSE
          )
        }, error = function(e) {
          exception_data <- get_exception_data(21)
          stop(sprintf(exception_data$sprintf_text, current_file))
        })

        data.table::setnames(raw_data, old = names(raw_data)[1], new = "WellRow")
        raw_data <- raw_data[, lapply(.SD, as.character)]

        melted_data <- data.table::melt(
          raw_data,
          id.vars = "WellRow",
          variable.name = "WellColumn",
          value.name = "ReadoutValue"
        )

        melted_data[, WellColumn := gsub("^\"|\"$", "", WellColumn)]
        invalid_cols <- !grepl("^[0-9]+$", melted_data$WellColumn)
        melted_data <- melted_data[!invalid_cols]

        melted_data[, WellColumn := as.integer(WellColumn)]

        melted_data[, (headers[["barcode"]]) := barcode]
        melted_data[, BackgroundValue := 0]

        melted_data[, ReadoutValue := trimws(ReadoutValue)]

        is_empty_or_na <- is.na(melted_data$ReadoutValue) |
          melted_data$ReadoutValue == "" |
          toupper(melted_data$ReadoutValue) %in% c("NA", "NAN", "INF", "-INF")

        num_regex <- "^[-+]?([0-9]+(\\.[0-9]*)?|\\.[0-9]+)([eE][-+]?[0-9]+)?$"
        valid_num_idx <- grepl(num_regex, melted_data$ReadoutValue)

        invalid_idx <- !valid_num_idx & !is_empty_or_na

        if (any(invalid_idx)) {
          futile.logger::flog.warn("Non-numeric readout values found and coerced to NA in plate %s of %s",
                                   barcode, current_file)
          melted_data[invalid_idx, ReadoutValue := NA_character_]
        }

        melted_data[is_empty_or_na, ReadoutValue := NA_character_]
        melted_data[, ReadoutValue := as.numeric(ReadoutValue)]

        futile.logger::flog.info("Plate %s read; %d wells", barcode, NROW(melted_data))

        all_results <- rbind(all_results, melted_data)
      }
    }
  }

  std_cols <- c(headers[["barcode"]], "WellRow", "WellColumn", "ReadoutValue", "BackgroundValue")
  data.table::setcolorder(all_results, intersect(std_cols, names(all_results)))

  return(unique(all_results))
}


#' Read in single xlsx data from EnVision
#'
#' @param results_file character, file path(s) to results file(s)
#' @param results_sheet results sheet names
#' @keywords load_files
#'
#' @return data.table with results data
#'
read_EnVision_xlsx <- function(results_file, results_sheet) {

  # expect an Excel spreadsheet
  exception_data <- get_exception_data(22)
  if (length(results_sheet) > 1) {
    # if multiple sheets, assume 1 plate per sheet
    tryCatch({
      df <-
        read_excel_to_dt(
          results_file,
          sheet = results_sheet,
          col_names = paste0("x", seq_len(48)),
          range = "A1:AV32"
        )
    }, error = function(e) {
      stop(sprintf(exception_data$sprintf_text, results_file, results_sheet))
    })
  } else {
    tryCatch({
      df <- read_excel_to_dt(results_file,
                             sheet = results_sheet,
                             col_names = FALSE)
    }, error = function(e) {
      stop(sprintf(exception_data$sprintf_text, results_file, results_sheet))
    })
    colnames(df) <-
      col_names <- paste0("x", seq_len(NCOL(df)))
  }
  # Find rows with data and drop empty columns
  colsRange <-
    grep("Plate information", unlist(df[, 1]))[1] + 10
  colsToKeep <- which(colSums(is.na(df[seq_len(colsRange), ])) != colsRange)
  df <-
    df[, colsToKeep, with = FALSE]
}

#' get Excel sheets names for a charvec of files
#' for non-Excel files return 0
#'
#' @param fls charvec with file pathsa
#' @keywords load_files
#'
#' @return list with one element per file with
#' sheet names or 0 (for non-Excel file)
#'
get_excel_sheet_names <- function(fls) {

  # results_file is a string or a vector of strings
  # test if the result files are .tsv or .xls(x) files
  isExcel <- vapply(fls, function(x) {
    return(tools::file_ext(x) == "xlsx")
  },
  logical(1))

  # read sheets in files; warning if more than one sheet (unexpected but can be handled)
  e_sheets <- vector("list", length(fls))
  e_sheets[!isExcel] <- 0
  e_sheets[isExcel] <-
    lapply(fls[isExcel], readxl::excel_sheets)
  if (any(lapply(e_sheets, length) > 1)) {
    futile.logger::flog.warn("Multiple sheets in result file: %s",
                             fls[lapply(e_sheets, length) > 1])
  }
  e_sheets
}

#' Get final results (as a data.table)
#' from raw edited EnVision data.table
#'
#' @param df raw data.table
#' @param barcode_idx numeric vector with barcode indices
#' @param barcode_col column number for barcode data
#' @param n_row number of rows
#' @param n_col number of columns
#' @param fname file name
#' @param sheet_name name of the Excel sheet
#' @param headers list with the headersa
#' @keywords load_files
#'
#' @return data.table derived from EnVision data
#'
get_df_from_raw_edited_EnVision_df <-
  function(df, barcode_idx, barcode_col, n_row, n_col, fname, sheet_name, headers) {

    res_l <- lapply(barcode_idx, function(iB) {
      # two type of format depending on where Background information is placed

      if (any(unlist(df[iB + (seq_len(4)), 1]) %in% "Background information")) {
        ref_bckgrd <- which(unlist(df[iB + (seq_len(4)), 1]) %in% "Background information")
        readout_offset <- 1 + ref_bckgrd
        stopifnot(as.character(df[iB + ref_bckgrd, 4]) %in% "Signal")
        BackgroundValue <- as.numeric(df[iB + ref_bckgrd + 1, 4])
      } else {
        # export without background, case with: " Exported with EnVision Workstation version 1.13.3009.1409 "
        ref_bckgrd <- 0
        readout_offset <- 1
        BackgroundValue <- 0
      }
      # check the structure of file is ok
      .check_file_structure(df, fname, sheet_name, readout_offset, n_row, n_col, iB, barcode_col)

      Barcode <- as.character(df[iB + 1, barcode_col, with = FALSE])
      if (is.na(Barcode) && ref_bckgrd > 0) {
        bg_row <- iB + ref_bckgrd + 1
        Barcode <- as.character(df[bg_row, barcode_col, with = FALSE])
      }
      if (is.na(Barcode)) return(NULL)
      readout <- as.matrix(df[iB + ref_bckgrd + seq_len(n_row) + 1,
                              seq_len(n_col), with = FALSE])
      stopifnot(dim(readout) == c(n_row, n_col))

      # check that the plate size is consistent and contains values
      if (any(is.na(readout))) {
        exception_data <- get_exception_data(23)
        stop(sprintf(
          exception_data$sprintf_text,
          basename(fname),
          sheet_name,
          as.character(df[iB + 1, barcode_col, with = FALSE])
        ))
      }

      df_results <- data.table::data.table(
        Barcode = Barcode,
        WellRow = LETTERS[seq_len(n_row)],
        WellColumn = as.vector(t(matrix(seq_len(n_col), n_col, n_row))),
        ReadoutValue = as.numeric(as.vector(readout)),
        BackgroundValue = BackgroundValue
      )
      names(df_results)[1] <- headers[["barcode"]][1]
      futile.logger::flog.info("Plate %s read; %d wells",
                               as.character(df[iB + 1, barcode_col, with = FALSE]),
                               dim(df_results)[1])
      df_results
    })
    data.table::rbindlist(res_l)
  }

#' Enhance raw edited EnVision data.table
#'
#' @param df raw data.table
#' @param barcode_col column number for barcode data
#' @param headers list with the headersa
#' @keywords load_files
#'
#' @return data.table derived from EnVision data
#'
enhance_raw_edited_EnVision_df <- function(df, barcode_col, headers) {

  # not empty rows; before discarding the rows: move ''Background information'' in the next row

  bckd_info_idx <- which(unlist(df[, 1]) %in% "Background information")
  if (length(bckd_info_idx) > 0) {
    df[bckd_info_idx + 1, 1] <- df[bckd_info_idx, 1]
    df[bckd_info_idx, 1] <- ""
  }

  # get the expected plate size
  plate_dim <- .get_plate_size(df)
  n_row <- plate_dim[1]
  n_col <- plate_dim[2]

  # manually add full rows
  plate_rows <- which(do.call(paste, df[, c(2, 3)]) %in% "Repeat Barcode") - 1
  spacer_rows <- grep("[[:alpha:]]", unlist(df[, 1]))
  standardized_bckd_info <-
    if (length(bckd_info_idx) == 0) {
      0
    } else {
      bckd_info_idx
    }
  actual_plate_rows <- pmax(plate_rows, standardized_bckd_info)
  data_rows <- unlist(lapply(actual_plate_rows,
                             function(x) {
                               (x + 4):(x + 4 + n_row - 1)}))
  # find full numeric rows
  df <- .fill_empty_wells(df, plate_rows, data_rows, n_row, n_col)

  # need to do some heuristic to find where the data is
  df_to_check <- df[, -6:-1]
  full_rows <- Reduce(union, lapply(df, function(x) grep("^\\d+$", x)))
  Barcode_idx <- which(unlist(df[, barcode_col, with = FALSE])
                       %in% headers[["barcode"]])
  additional_rows <- c(Barcode_idx, bckd_info_idx + 1)
  full_rows_index <-
    unique(sort(c(
      additional_rows,
      additional_rows + 1,
      setdiff(full_rows, spacer_rows))))

  # don't consider the first columns as these may be metadata
  # if big gap, delete what is at the bottom (Protocol information)
  gaps <- min(max(data_rows), NROW(df))
  df <- df[full_rows_index[full_rows_index <= gaps], ]

  # add empty column to complete plate (assume left column is #1)
  if (NCOL(df) < n_col) df[, (NCOL(df) + 1):n_col] <- NA
  df
}

#' Get final results (as a data.table)
#' from raw unedited EnVision data.table
#'
#' @param df raw data.table
#' @param barcode_col column number for barcode data
#' @param n_row number of rows
#' @param n_col number of columns
#' @keywords load_files
#'
#' @return data.table derived from EnVision data
#'
get_df_from_raw_unedited_EnVision_df <-
  function(df, n_row, n_col, barcode_col) {

    # proper original EnVision file
    Barcode <- df[3, barcode_col, with = FALSE]
    selected_cols <- seq_len(n_col)
    readout <-
      as.matrix(df[4 + seq_len(n_row), selected_cols, with = FALSE])

    if (any(df[, 1] %in% "Background information")) {
      ref_bckgrd <-
        which(df[, 1] %in% "Background information")
      BackgroundValue <-
        as.numeric(df[ref_bckgrd + 1, 4])
    } else {
      # export without background information
      BackgroundValue <- 0
    }

    df_results <- data.table::data.table(
      Barcode = unlist(Barcode),
      WellRow = LETTERS[seq_len(n_row)],
      WellColumn = as.vector(t(matrix(
        seq_len(n_col), n_col, n_row
      ))),
      ReadoutValue = as.numeric(as.vector(readout)),
      BackgroundValue = BackgroundValue
    )
    futile.logger::flog.info("Plate %s read; %d wells",
                             as.character(Barcode),
                             dim(df_results)[1])
    df_results
  }

#' Load tecan results from xlsx
#'
#' This functions loads and checks the results file
#'
#' @param results_file string, file path to a result file
#' @param headers list of headers identified in the manifest
#' @keywords load_files
#'
#' @return data.table derived from Tecan data
#'
load_results_Tecan <-
  function(results_file, headers = gDRutils::get_env_identifiers()) {
    # Assertions:
    checkmate::assert_string(results_file)
    # check the result files is a .xls(x) file
    isExcel <- tools::file_ext(results_file) == "xlsx"
    if (!isExcel) {
      futile.logger::flog.error("Results file for Tecan importer must be a 'xls(x)' file: %s",
                                results_file)
    }
    # read sheets in files; in the Tecan format each sheet is a plate
    results_sheets <- readxl::excel_sheets(results_file)
    if (length(results_sheets) < 1) {
      futile.logger::flog.error("No data sheet found in: %s",
                                results_file)
    }
    # # read all plates
    all_results <- read_in_results_Tecan(results_file, results_sheets, headers)
    return(all_results)
  }

#' read in Tecan data
#'
#' @param results_file string, file path to a result file
#' @param results_sheets template sheet names
#' @param headers list of headers identified in the manifest
#' @keywords load_files
#'
#' @return data.table derived from Tecan data
#'
read_in_results_Tecan <- function(results_file, results_sheets, headers) {
  all_results <- data.table::data.table()
  for (iS in seq_along(results_sheets)) {
    futile.logger::flog.info("Reading file %s, sheet %s", results_file, results_sheets[[iS]])
    # read the content of each plate
    tryCatch({
      df <- read_excel_to_dt(results_file,
                             sheet = results_sheets[[iS]],
                             col_names = FALSE)
    }, error = function(e) {
      exception_data <- get_exception_data(22)
      stop(sprintf(exception_data$sprintf_text, results_file, iS))
    })

    # find the indicator ("<>") that identifies where plate readings are
    ind <- which(df == "<>", arr.ind = TRUE)
    dfm <- df[(ind[1]):NROW(df), ind[2]:NCOL(df), with = FALSE] # remove text above "<>"
    # remove text after data matrix ends, as identified by first na value
    ind <- which(is.na(dfm), arr.ind = TRUE)[1]
    dfm <- dfm[seq_len(ind) - 1, seq_len(NCOL(dfm)), with = FALSE]

    # rows and columns in data matrix with row and col names
    n_row <- NROW(dfm)
    n_col <- NCOL(dfm)
    readout <- dfm[2:n_row, 2:n_col, with = FALSE]
    # get well identifiers (numbers and letters) from layout
    WellRow <- as.character(t(dfm[2:n_row, 1]))
    WellColumn <- strtoi(as.character(dfm[1, 2:n_col]))
    # results data table for plate
    df_results <- data.table::data.table(
      Barcode = results_sheets[iS],
      WellRow = WellRow,
      WellColumn =  as.vector(t(matrix(
        WellColumn, NCOL(readout), NROW(readout)
      ))),
      ReadoutValue = as.numeric(as.vector(as.matrix(readout))),
      BackgroundValue = 0 ## Tecan users report negligible background readings, usually background is not recorded
    )
    names(df_results)[1] <- headers[["barcode"]]
    all_results <- rbind(all_results, df_results)
  }
  all_results
}

.find_header <- function(header_dt, header_name, error_msg) {
  idx <- which(header_dt[[1]] %in% header_name)
  if (length(idx) == 0) {
    exception_data <- get_exception_data(37)
    stop(sprintf(exception_data$sprintf_text, error_msg))
  }
  idx
}

#' Load incucyte results from plain text
#'
#' This functions loads incucyte time-course cell count file
#'
#' @param results_file list of strings: file paths to result paths from individual plates
#' @param headers list of headers identified in the manifest
#' @keywords load_files
#'
#' @return data.table derived from Incucyte data
#'
load_results_Incucyte <-
  function(results_file,
           headers = gDRutils::get_env_identifiers()) {

    # identifiers
    bcode_name <- headers$barcode[1]
    bcode_names <- c(bcode_name, paste0(bcode_name, ":"))
    dur_name <- headers$duration

    # Use lapply instead of for loop for better performance and idiomatic R style
    all_data_list <- lapply(results_file, function(iP) {

      header_dt <- if (grepl(".xlsx$", iP)) {
        tryCatch({
          read_excel_to_dt(iP, n_max = 20)
        }, error = function(e) {
          exception_data <- get_exception_data(22)
          stop(sprintf(exception_data$sprintf_text, iP))
        })
      } else {
        tryCatch({
          data.table::fread(iP, nrows = 20, header = TRUE)
        }, error = function(e) {
          exception_data <- get_exception_data(21)
          stop(sprintf(exception_data$sprintf_text, iP))
        })
      }

      dstart_idx <- .find_header(header_dt, "Date Time", "missing 'Date Time' column")
      barcode_idx <- .find_header(header_dt, bcode_names, sprintf("missing '%s' column", bcode_name))
      barcode <- header_dt[barcode_idx, 2][[1]]

      dt_input <- if (grepl(".xlsx$", iP)) {
        read_excel_to_dt(iP, skip = dstart_idx)
      } else {
        data.table::fread(iP, skip = dstart_idx, header = TRUE)
      }

      dt_input <- data.table::melt(
        dt_input[, -1],
        id.vars = 1,
        variable.name = "Well",
        value.name = "ReadoutValue"
      )

      dt_input[[bcode_name]] <- barcode

      return(dt_input)
    })

    all_data <- data.table::rbindlist(all_data_list)

    all_data[, (dur_name) := as.numeric(Elapsed)]
    all_data[, ReadoutValue := as.numeric(ReadoutValue)]
    all_data <- all_data[!is.na(get(dur_name)) & !is.na(ReadoutValue)]

    well_rname <- headers$well_position[1]
    well_cname <- headers$well_position[2]
    all_data[[well_rname]] <- gsub("([A-Za-z]+).*", "\\1", all_data$Well)
    all_data[[well_cname]] <- gsub("[A-Za-z]+(.*)", "\\1", all_data$Well)

    # Cleanup intermediate columns
    cols_to_remove <- c("Well", "Elapsed")
    all_data[, (cols_to_remove) := NULL]

    return(all_data)
  }


#' check_metadata_names
#'
#' Check whether all metadata names are correct
#'
#' @param col_df a character with colnames of df
#' @param df_name a name of data.table ("" by default)
#' @param df_type a type of a data.table (NULL by default)
#' @keywords correction_exception
#'
#' @examples
#'  td <- get_test_data()
#'  m_file <- manifest_path(td)
#'  m_data <- read_excel_to_dt(m_file)
#'  result <- check_metadata_names(col_df = colnames(m_data))
#'
#' @return a charvec with corrected colnames of df
#'
#' @export
#'
check_metadata_names <-
  function(col_df,
           df_name = "",
           df_type = NULL) {
    # Assertions:
    checkmate::assert_character(col_df)
    checkmate::assert_character(df_name)
    checkmate::assert_character(df_type, null.ok = TRUE)

    check_metadata_req_col_names(col_df, df_name, df_type)
    corrected_names <- col_df
    corrected_names <- check_metadata_against_spaces(corrected_names, df_name)
    corrected_names <- check_metadata_field_names(corrected_names, df_name)
    corrected_names <- check_metadata_headers(corrected_names, df_name)
    corrected_names
  }

#' Check metadata for required column names
#
#' @param col_df a charvec with corrected colnames of df
#' @param df_name a name of data.table ("" by default)
#' @param df_type a type of a data.table (NULL by default)
#' @keywords correction_exception
#'
#'
#' @return \code{NULL} invisibly.
#'
check_metadata_req_col_names <- function(col_df, df_name, df_type) {
  # first check for required column names
  if (!is.null(df_type)) {
    if (df_type == "manifest") {
      expected_headers <- gDRutils::get_header("manifest")
    } else if (df_type == "template") {
      expected_headers <- gDRutils::get_env_identifiers("drug")
    } else if (df_type == "template_treatment") {
      expected_headers <-
        c(gDRutils::get_env_identifiers("drug"),
          "Concentration")
    }
    upperHeaders <- lapply(expected_headers, toupper)
    headersOK <-
      vapply(upperHeaders, function(x) {
        any(x %in% toupper(col_df))}, FUN.VALUE = logical(1))
    if (any(!headersOK)) {
      exception_data <- get_exception_data(24)
      stop(sprintf(
        exception_data$sprintf_text, df_type,
        toString(expected_headers[!(expected_headers %in% col_df)])))
    }
    if (df_type == "template_treatment") {
      # assess if multiple drugs and proper pairing
      n_drug <- agrep(gDRutils::get_env_identifiers("drug"), col_df, ignore.case = TRUE)
      n_conc <- agrep("Concentration", col_df, ignore.case = TRUE)
      if (length(n_drug) != length(n_conc)) {
        exception_data <- get_exception_data(32)
        stop(sprintf(exception_data$sprintf_text, df_name))
      }
      if (length(n_drug) > 1) {
        trt_sheets <- c(
          paste0(
            gDRutils::get_env_identifiers("drug"), "_", 2:length(n_drug)),
          paste0("Concentration_", 2:length(n_conc))
        )
        if (!(all(toupper(trt_sheets) %in% toupper(col_df)))) {
          exception_data <- get_exception_data(25)
          stop(sprintf(
            exception_data$sprintf_text, df_name,
            toString(trt_sheets[!(toupper(trt_sheets) %in% toupper(col_df))])))
        }
      }
    }
  }

}

#' Check metadata field names
#
#' @param corrected_names a charvec with corrected colnames of df
#' @param df_name a name of data.table ("" by default)
#' @keywords correction_exception
#'
#' @return a charvec with corrected colnames of df
#'
check_metadata_field_names <- function(corrected_names, df_name) {
  # check for wrong metadata field names (including dash, starting with number, ... )
  bad_names <-
    regexpr("\\W", corrected_names) > 0 |
    regexpr("\\d", corrected_names) == 1
  if (any(bad_names)) {
    exception_data <- get_exception_data(26)
    stop(sprintf(
      exception_data$sprintf_text,
      df_name,
      toString(corrected_names[bad_names])
    ))
  }
  corrected_names
}

#' Check metadata against spaces
#
#' @param corrected_names a charvec with corrected colnames of df
#' @param df_name a name of data.table ("" by default)
#' @keywords correction_exception
#'
#' @return a charvec with corrected colnames of df
#'
check_metadata_against_spaces <- function(corrected_names, df_name) {
  # remove spaces and convert to WordUppercase
  names_spaces <- regexpr("\\s", corrected_names) > 0
  if (any(names_spaces)) {
    for (i in which(names_spaces)) {
      s <- strsplit(corrected_names[i], " ")[[1]]
      corrected_names[i] <-
        paste0(toupper(substring(s, 1, 1)),
               substring(s, 2),
               collapse = "")
    }
    futile.logger::flog.warn(
      "Metadata field names for %s cannot contain spaces --> corrected to: %s",
      df_name,
      toString(corrected_names[names_spaces])
    )
  }
  corrected_names
}

#' Check whether metadata headers are correct and make fixes if needed
#
#' @param corrected_names a charvec with corrected colnames of df
#' @param df_name a name of data.table ("" by default)
#' @keywords correction_exception
#'
#' @return a charvec with corrected colnames of df
#'
check_metadata_headers <- function(corrected_names, df_name) {

  # common headers that are written in a specific way
  # throw warning if close match and correct upper/lower case for consistency
  controlled_headers <- gDRutils::get_header("controlled")
  for (i in seq_along(controlled_headers)) {
    grep_pattern <- paste0(controlled_headers[[i]], "$", collapse = "|")
    exact_match_grep <- grep(grep_pattern, corrected_names)

    # To avoid cases when grep compare 'PLATE' to 'temPLATE'
    case_match <- setdiff(grep(grep_pattern, corrected_names, ignore.case = TRUE),
                          exact_match_grep)
    if (isTRUE(length(exact_match_grep) == 1 ||
               corrected_names[case_match] %in% unlist(controlled_headers)))
      next

    if (length(case_match) > 0) {
      corrected_names[case_match] <- controlled_headers[[i]]
      futile.logger::flog.warn("Header %s in %s corrected to %s",
                               corrected_names[case_match],
                               df_name,
                               controlled_headers[[i]])
    }
  }

  # check for headers that are reserved for downstream analyses
  check_headers <-
    setdiff(
      gDRutils::get_header("reserved"),
      gDRutils::get_env_identifiers("well_position")
    )
  if (any(corrected_names %in% check_headers)) {
    exception_data <- get_exception_data(27)
    stop(sprintf(
      exception_data$sprintf_text,
      toString(intersect(check_headers, corrected_names)),
      df_name
    ))
  }
  corrected_names
}

#' Read EnVision delimited text files
#'
#' This function reads file from the EnVision Workstation
#'
#' @param file string to path of input file from EnVision scanner
#' @param nrows maximum number of file rows to be processed
#' @param seps potential field separators of the input file
#' @keywords load_files
#'
#' @return a list containing the data table, n_col, n_row, and if is edited
#'
read_EnVision_delim <- function(file,
                         nrows = 10000,
                         seps = c(",", "\t")) {
  # Assertions:
  checkmate::assert_string(file)
  checkmate::assert_number(nrows)
  checkmate::assert_character(seps)

  results.list <- read_in_EnVision_file(file, nrows, seps)

  if (which(vapply(results.list, function(x) {
    grepl("Plate information", x[1])}, logical(1))) != 1) {
    # fails if not the right header
    exception_data <- get_exception_data(29)
    stop(sprintf(exception_data$sprintf_text, file))
  }
  # identify if original csv file or not (isEdited + n_row/n_col)
  ep <- get_EnVision_properties(results.list, basename(file))

  # pad the lines with NA if not full before creating the data.table
  results.list <- lapply(results.list, function(x) {
    if (length(x) < ep[["n_col"]]) {
      x <- c(x, array(NA, ep[["n_col"]] - length(x)))
    } else if (length(x) > ep[["n_col"]]) {
      x <- x[seq_len(ep[["n_col"]])]
    }
    x[x == "" & !is.na(x)] <- NA
    as.list(x)
  })

  df_ <- data.table::rbindlist(results.list)
  data.table::setnames(df_, paste0("x", seq_len(ep[["n_col"]])))

  return(list(df = df_, n_col = ep[["n_col"]], n_row = ep[["n_row"]], isEdited = ep[["isEdited"]]))
}

#' Read EnVision file
#'
#' This function reads file from the EnVision Workstation
#'
#' @param file  input file from EnVision
#' @param nrows maximum number of file rows to be processed
#' @param seps potential field separators of the input file
#' @keywords load_files
#'
#' @return list with one element per EnVisoin input file
#'
read_in_EnVision_file <- function(file, nrows, seps) {

  con <- file(file)
  open(con)

  results.list <- list()
  current.line <- 1

  while (length(line <- readLines(con, n = 1, warn = FALSE)) > 0 && current.line < nrows) {
    cleaned.line <- gsub("=\"([^\"]*)\"", "\\1", line)
    results.list[[current.line]] <- cleaned.line
    current.line <- current.line + 1
  }
  close(con)

  # Try to identify the delimiter.
  n_sep <- colSums(vapply(seps, function(sep) {
    vapply(results.list[seq_len(10)], function(line) {
      length(unlist(strsplit(line, split = sep)))},
      integer(length = 1))
  }, integer(length = 10)))
  sep <- n_sep[n_sep == max(n_sep)]
  if (sep < 30) {
    exception_data <- get_exception_data(28)
    stop(sprintf(exception_data$sprintf_text, file))
  }
  sep <- names(sep)
    lapply(results.list, function(line) {
      unlist(strsplit(line, split = sep))})
}

#' Get properties of EnVision data
#'
#' This function return properties of EnVision data
#'
#' @param results.list  list with EnVision data
#' @param fname name of the input file
#' @keywords load_files
#'
#' @return list with EnVision propertiesa
#'
get_EnVision_properties <- function(results.list, fname) {

  # has been open/saved in a spreadsheet software --> first line is padded with empty columns
  isEdited <- (length(results.list[[1]]) > 1) |
    !any(vapply(results.list, function(x) {
      grepl("EnVision Workstation", x[1])}, logical(1)))

  if (isEdited) {
    # Identify closest larger plate size; may be altered and miss columns.
    n_col <- max(vapply(results.list[5:10], length, integer(length = 1))) # Assume first row of data lies in rows 5:10
    n_col <- 1.5 * 2 ^ ceiling(log2((n_col - 1) / 1.5))
    n_row <- which(vapply(results.list, function(x) {
               grepl("Basic assay information", x[1]) | grepl("Plate information", x[1])},
             logical(1)))
    if (length(n_row) == 1 && n_row == 1) {
      # only the top line with "Plate information" was found
      n_row <- length(results.list) - 4
    } else {
      # scrap a few rows above "Basic assay information"
      n_row <- min(n_row[n_row > 1]) - 7
    }
    if (length(n_row) != 1 ||
        abs(log2(1.5 * n_row / n_col)) > 0.5)  {
      exception_data <- get_exception_data(30)
      stop(sprintf(exception_data$sprintf_text, fname))
    }
    n_row <- n_col / 1.5
  } else {
    n_col <- max(vapply(results.list[5:10], length, integer(length = 1)))
    n_row <- sum(vapply(results.list[5:which(vapply(results.list, function(x) {
      grepl("Basic assay information", x[1])}, logical(1)))],
      length, integer(1)) == n_col)
    if (log2(1.5 * n_row / n_col) != 0)  {
      exception_data <- get_exception_data(30)
      stop(sprintf(exception_data$sprintf_text, fname))
    }
  }
  list(n_col = n_col, n_row = n_row, isEdited = isEdited)
}

#' Get plate size
#'
#' @details
#' All plate sizes assume 1.5x nrows = ncolumns.
#' @keywords internal
#'
#' @return charvec with plate dims
#'
.get_plate_size <- function(df) {
  n_col <- 1.5 * 2 ^ ceiling(log2((NCOL(df) - 2) / 1.5))
  n_row <- n_col / 1.5
  c(n_row, n_col)
}


#' Check the structure of raw data
#' @keywords internal
#'
#' @return \code{NULL} invisibly.
.check_file_structure <- function(df,
                                  filename,
                                  sheet_name,
                                  readout_offset,
                                  n_row,
                                  n_col,
                                  bcode_idx,
                                  bcode_col) {

  # check the structure of file is ok
  check_values <-
    as.matrix(df[bcode_idx + readout_offset + c(0, 1, n_row, n_row + 1),
                 n_col, with = FALSE])
  if (is.na(check_values[2])) {
    exception_data <- get_exception_data(31)
    stop(
      sprintf(
        exception_data$sprintf_text,
        filename,
        sheet_name,
        as.character(df[bcode_idx + 1, bcode_col])
      )
    )
  }
}


#' Correct plates with not fully filled readout values
#' @keywords internal
#'
#' @return data.table with corrected plates data
#'
.fill_empty_wells <- function(df, plate_rows, data_rows, exp_row, exp_col, numeric_regex = "^\\d+$") {
  all_rows <- Reduce(intersect, lapply(df, function(x) grep(numeric_regex, x)))
  if (NCOL(df) < exp_col) {
    new_cols <- exp_col - NCOL(df)
    df <- cbind(df, matrix("", ncol = new_cols))
    df[data_rows, rev(names(df))[seq_len(new_cols)]] <- NA
  }
  if (length(all_rows) / exp_row != length(plate_rows)) {
    fill_rows <- intersect(which(apply(df, 1, function(x) all(is.na(x)))), data_rows)
    df[fill_rows, ] <- "0"

    #fill up data_rows
    df[data_rows, ] <- lapply(df[data_rows, ], function(x) ifelse(is.na(x), "0", x))
  }
  df
}


#' Standardize untreated values to ignore cases
#' @keywords internal
#'
#' @return data.table with standardized untreated values
#'
.standardize_untreated_values <- function(df) {
  untreated_tags <- gDRutils::get_env_identifiers("untreated_tag")

  data.table::as.data.table(lapply(df, function(x) {
    if (is.factor(x)) x <- as.character(x)
    x[toupper(x) %in% toupper(untreated_tags)] <- untreated_tags[[1]]
    return(x)
    }))
}

context("load_files")

test_that("load_manifest works as expected", {

  td1 <- get_test_data()

  m_df <- load_manifest(manifest_path(td1))
  expect_identical(td1@ref_m_df, m_df$data)

  td2 <- get_test_Tecan_data()

  m_df <- load_manifest(td2$m_file)
  ref_m_df <- qs2::qs_read(td2$ref_m_df)
  expect_equal(m_df, ref_m_df)

  err_msg1 <- "Assertion on 'manifest_file' failed: File does not exist: '/non/existent_file'."
  expect_error(load_manifest("/non/existent_file"), err_msg1)

  err_msg2 <- "'manifest_file' must be a character vector"
  expect_error(load_manifest(c(2, 3)), err_msg2)

  err_msg3 <- "Barcodes in Manifest must be unique!"
  expect_error(load_manifest(c(manifest_path(td1), manifest_path(td1))), err_msg3)

})

test_that("load_results works as expected", {

  td1 <- get_test_data()

  headers <- gDRutils::get_env_identifiers()
  headers$barcode <- headers$barcode[[1]]

  res_tbl <- load_results(df_results_files = c(result_path(td1)), headers = headers)
  ref_tbl <- data.table::fread(td1@ref_r1_r2)
  expect_equal(res_tbl, ref_tbl)

  df_results <- data.table::data.table(datapath = result_path(td1), name = basename(result_path(td1)))
  res_df_tbl <- load_results(df_results, headers = headers)
  expect_equal(res_df_tbl, ref_tbl)

  res_tbl2 <- load_results(df_results_files = c(result_path(td1)[1]), headers = headers)
  ref_tbl2 <- data.table::fread(td1@ref_r1)
  expect_equal(res_tbl2, ref_tbl2)

  td2 <- get_test_Tecan_data()

  res_tbl3 <- load_results(df_results_files = c(td2$r_files), instrument = "Tecan", headers = headers)
  ref_tbl3 <- qs2::qs_read(td2$ref_r_df)
  expect_equal(res_tbl3, ref_tbl3)

  err_msg1 <- "Assertion on 'results_file' failed: File does not exist: '/non/existent_file'."
  expect_error(load_results(c(result_path(td1)[1], "/non/existent_file")), err_msg1)

  err_msg_2a <- "Assertion on 'instrument' failed: "
  err_msg2b <- "Must comply to pattern '^EnVision$|^long_tsv$|^Tecan$|^EnVision_new$|^Incucyte$'."
  err_msg2 <- paste0(err_msg_2a, err_msg2b)
  expect_error(load_results(result_path(td1), "invalid_instrument"), err_msg2, fixed = TRUE)
})

test_that("load_templates works as expected", {

  td1 <- get_test_data()

  t_tbl <- load_templates(df_template_files = c(template_path(td1)))
  ref_tbl <- data.table::fread(td1@ref_t1_t2, colClasses = "character")
  expect_equal(t_tbl, ref_tbl)

  df_templates <- data.table::data.table(datapath = template_path(td1), name = basename(template_path(td1)))
  res_t_tbl <- load_templates(df_templates)
  expect_equal(res_t_tbl, ref_tbl)

  td2 <- get_test_Tecan_data()

  res_t_tbl3 <- load_templates(df_template_files = c(td2$t_files))
  ref_tbl3 <- .standardize_untreated_values(qs2::qs_read(td2$ref_t_df))
  expect_equal(res_t_tbl3, ref_tbl3)

  err_msg1 <- "Assertion on 'template_file' failed: File does not exist: '/non/existent_file'."
  expect_error(load_templates(c(template_path(td1)[1], "/non/existent_file")), err_msg1)
})

test_that("load_templates with no untreated conditions works as expected", {
  err_msg <- "No untreated controls were found in the treatment. Please upload the appropriate treatment."
  expect_error(load_templates(system.file("extdata/data_for_unittests/Template_7daytreated.xlsx",
                                          package = "gDRimport")), err_msg)

})

test_that("load_data works as expected", {

  td1 <- get_test_data()

  l_tbl <- load_data(manifest_path(td1), template_path(td1), result_path(td1))
  expect_identical(td1@ref_m_df, l_tbl$manifest)
  ref_tbl <- .standardize_untreated_values(data.table::fread(td1@ref_t1_t2))
  expect_equal(l_tbl$treatments, ref_tbl)
  ref_tbl <- data.table::fread(td1@ref_r1_r2)
  expect_equal(l_tbl$data, ref_tbl)

  td2 <- get_test_Tecan_data()

  l_tbl2 <- load_data(td2$m_file, td2$t_files, td2$r_files, instrument = "Tecan")
  ref_m_df <- qs2::qs_read(td2$ref_m_df)
  expect_equal(ref_m_df$data, l_tbl2$manifest)
  ref_t_df <- .standardize_untreated_values(qs2::qs_read(td2$ref_t_df))
  expect_equal(ref_t_df, l_tbl2$treatments)
  ref_r_df <- qs2::qs_read(td2$ref_r_df)
  expect_equal(l_tbl2$data, ref_r_df)

  td4 <- get_test_EnVision_data()
  l_tbl4 <-
    load_data(td4$m_file, td4$t_files, td4$r_files, instrument = "EnVision")
  ref_l <- qs2::qs_read(td4$ref_l_path)
  expect_equal(ref_l$manifest, l_tbl4$manifest)
  expect_equal(ref_l$treatments, l_tbl4$treatments)
  expect_equal(ref_l$data, l_tbl4$data)

  err_msg1 <- "'manifest_file' must be a readable path"
  expect_error(load_data("/non/existent_file", template_path(td1), result_path(td1)), err_msg1)

  err_msg2 <- "'manifest_file' must be a character vector"
  expect_error(load_data(c(2, 3), template_path(td1), result_path(td1)), err_msg2)

  err_msg3 <- "Barcodes in Manifest must be unique!"
  expect_error(load_manifest(c(manifest_path(td1), manifest_path(td1))), err_msg3)

  err_msg4 <- "Following path(s) with no read permission found: '/non/existent_file'"
  expect_error(load_data(manifest_path(td1), c(result_path(td1)[1], "/non/existent_file"),
                         result_path(td1)), err_msg4, fixed = TRUE)

  err_msg_5a <- "Assertion on 'instrument' failed: "
  err_msg5b <- "Must comply to pattern '^EnVision$|^long_tsv$|^Tecan$|^EnVision_new$|^Incucyte$'."
  err_msg5 <- paste0(err_msg_5a, err_msg5b)
  expect_error(load_data(manifest_path(td1), template_path(td1), result_path(td1), "invalid_instrument"),
               err_msg5, fixed = TRUE)

  err_msg6 <- "Assertion on 'results_file' failed: File does not exist: '/non/existent_file'."
  expect_error(load_data(manifest_path(td1), template_path(td1), c(result_path(td1)[1], "/non/existent_file")),
               err_msg6)

  td5 <- get_test_tsv_data()
  l_tbl5 <-
    load_data(td5$m_file, td5$t_files, td5$r_files)
  ref_l <- qs2::qs_read(td5$ref_l_path)
  expect_equal(ref_l$manifest, l_tbl5$manifest)
  expect_equal(ref_l$treatments, l_tbl5$treatments)
  expect_equal(ref_l$data, l_tbl5$data)
})

test_that(".get_plate_size works as expected", {
  df <- read_excel_to_dt(system.file("extdata/data1/RawData_day7.xlsx", package = "gDRimport"))
  size <- .get_plate_size(df)
  expect_length(size, 2)
  expect_equal(prod(size), 384)
})

test_that(".check_file_structure works as expected", {
  df <- read_excel_to_dt(system.file("extdata/data1/RawData_day7.xlsx", package = "gDRimport"))
  df <-
    df[, !apply(df[seq_len(35), ], 2, function(x) {
      all(is.na(x))}), with = FALSE]
  Bckd_info_idx <-
    which(df[[1]] %in% "Background information")
  if (length(Bckd_info_idx) > 0) {
    df[Bckd_info_idx + 1, 1] <- df[Bckd_info_idx, 1]
    df[Bckd_info_idx, 1] <- ""
  }
  size <- .get_plate_size(df)
  n_row <- size[1]
  n_col <- size[2]
  df_to_check <- df[, c(-6:-1), with = FALSE]
  ml <- lapply(df_to_check, function(x) {
    is.na(suppressWarnings(as.numeric(x)))
  })
  full_rows <- rowSums(do.call(cbind, ml)) != NCOL(df_to_check)
  plate_row <- which(df[[1]] %in% "Plate information")
  spacer_rows <- unlist(lapply(plate_row, function(x) c(x + 1, x + 2, x + 4 + n_row)))
  data_rows <- unlist(lapply(plate_row, function(x) (x + 4):(x + 4 + n_row - 1)))
  df[data_rows, ] <- lapply(df[data_rows, ], function(x) ifelse(is.na(x), "0", x))
  full_rows_index <- sort(union(spacer_rows, data_rows))
  gaps <-
    min(which(full_rows)[(diff(which(full_rows)) > 20)] + 1, dim(df)[1])
  df <-
    df[full_rows_index[full_rows_index <= gaps], ]
  if (NCOL(df) < n_col) {
    df[, (NCOL(df) + 1):n_col] <- NA
  }
  iB <- 1
  iF <- 1
  iS <- 1
  results_filename <- "data"
  ref_bckgrd <- 4
  readout_offset <- 1 + ref_bckgrd
  barcode_col <- 3
  expect_null(.check_file_structure(df, basename(results_filename[[iF]]),
                                    iS, readout_offset, n_row, n_col, iB, barcode_col))

  df2 <- read_excel_to_dt(system.file("extdata/data1/RawData_day7.xlsx", package = "gDRimport"))
  expect_error(.check_file_structure(df2, basename(results_filename[[iF]]),
                                     iS, readout_offset, n_row, n_col, iB, barcode_col))
})

test_that(".fill_empty_wells works as expected", {
  df <- read_excel_to_dt(system.file("extdata/data1/RawData_day7.xlsx", package = "gDRimport"), col_names = FALSE)
  df <-
    df[, !apply(df[seq_len(35), ], 2, function(x) {
      all(is.na(x))}), with = FALSE]
  Bckd_info_idx <-
    which(df[[1]] %in% "Background information")
  if (length(Bckd_info_idx) > 0) {
    df[Bckd_info_idx + 1, 1] <- df[Bckd_info_idx, 1]
    df[Bckd_info_idx, 1] <- ""
  }
  size <- .get_plate_size(df)
  n_row <- size[1]
  n_col <- size[2]
  df_to_check <- df[, -6:-1, with = FALSE]
  ml <- lapply(df_to_check, function(x) {
    is.na(suppressWarnings(as.numeric(x)))
  })
  full_rows <- rowSums(do.call(cbind, ml)) != NCOL(df_to_check)
  plate_row <- which(df[[1]] %in% "Plate information")
  spacer_rows <- unlist(lapply(plate_row, function(x) c(x + 1, x + 2, x + 4 + n_row)))
  data_rows <- unlist(lapply(plate_row, function(x) (x + 4):(x + 4 + n_row - 1)))
  df2 <- .fill_empty_wells(df, plate_row, data_rows, n_row, n_col)
  expect_identical(df, df2)
  df_modified <- df
  df_modified[9, ] <- NA
  df_modified <- .fill_empty_wells(df_modified, plate_row, data_rows, n_row, n_col)
  expect_true(all(df_modified[9, ] == 0))
})

test_that(".standardize_untreated_values works as expected", {
  untreated_tags <- gDRutils::get_env_identifiers("untreated_tag")

  df_test <- data.table::data.table(a = c(untreated_tags[[1]], untreated_tags[[2]],
                                          toupper(untreated_tags[[1]]), tolower(untreated_tags[[2]])))
  df_corrected <- .standardize_untreated_values(df_test)
  expect_true(all(unlist(df_corrected) == untreated_tags[[1]]))
})

test_that("check_metadata_names works as expected", {

  td1 <- get_test_data()
  m_file <- manifest_path(td1)
  m_data <- read_excel_to_dt(m_file)

  result <- check_metadata_names(col_df = colnames(m_data))
  expect_equal(result, colnames(m_data))

  result <- check_metadata_names(col_df = colnames(m_data), m_file, df_type = "manifest")
  expect_equal(result, colnames(m_data))

  t_file <- template_path(td1)[[1]]
  t_data <- correct_template_sheets(t_file)
  result <- check_metadata_names(col_df = t_data[[1]])
  expect_equal(result, t_data[[1]])
  result <- check_metadata_names(col_df = t_data[[1]], df_type = "template")
  expect_equal(result, t_data[[1]])
  result <- check_metadata_names(col_df = t_data[[1]], df_type = "template_treatment")
  expect_equal(result, t_data[[1]])

  expect_error(check_metadata_names(col_df = t_data[[1]][-1], df_type = "template_treatment"))
  expect_error(check_metadata_names(col_df = t_data[[1]][-3], df_type = "template_treatment"), "Treatment file")

  t_data_with_space <- t_data[[1]]
  t_data_with_space[1] <- "Gnumber "
  expect_no_error(check_metadata_names(col_df = t_data_with_space))

  t_data_with_number <- t_data[[1]]
  t_data_with_number[5] <- "1"
  expect_error(
    check_metadata_names(col_df = t_data_with_number),
    "cannot contain special characters or start with a number"
  )

  t_data_with_duplication <- t_data[[1]]
  t_data_with_duplication[5] <- "Gnumber"
  t_data_with_duplication[6] <- "GnumbeR"
  expect_no_error(check_metadata_names(col_df = t_data_with_duplication))

  t_data_with_resticted_name <- t_data[[1]]
  t_data_with_resticted_name[5] <- "Tissue"
  expect_error(check_metadata_names(col_df = t_data_with_resticted_name), "Metadata field name")

  expect_error(check_metadata_names(col_df = NULL), "Assertion on 'col_df' failed")
  expect_error(check_metadata_names(col_df = colnames(m_data), df_name = NULL), "Assertion on 'df_name'")
  expect_error(check_metadata_names(col_df = colnames(m_data), df_type = 5), "Assertion on 'df_type'")

})

test_that("load_results_Incucyte works as expected", {

  headers <- gDRutils::get_env_identifiers()
  bcode_name <- headers$barcode[1]
  d_name <- headers$duration
  well_rname <- headers$well_position[1]
  well_cname <- headers$well_position[2]

  file_csv_1_path <- tempfile(fileext = ".csv")
  file_csv_2_na_path <- tempfile(fileext = ".csv")
  file_tsv_3_path <- tempfile(fileext = ".tsv")
  file_xlsx_4_path <- tempfile(fileext = ".xlsx")
  file_bad_header_path <- tempfile(fileext = ".csv")
  file_bad_barcode_path <- tempfile(fileext = ".csv")
  file_custom_header_path <- tempfile(fileext = ".csv")
  file_colon_header_path <- tempfile(fileext = ".csv")

  writeLines(
    c(
      "Vessel Name: Test Plate 1,,,",
      "Metric: Red Object Count,,,",
      "Notes:,,,",
      sprintf("Barcode,%s", "PLATE_001_CSV,,"),
      "Some extra metadata line,,,",
      "Another metadata line,,,",
      "Date Time,Elapsed,A1,A2",
      "8/18/25 3:02,0,100,200",
      "8/18/25 9:02,6,110,210"
    ),
    file_csv_1_path
  )

  writeLines(
    c(
      "Vessel Name: Test Plate 2,,,",
      sprintf("Barcode,%s", "PLATE_002_CSV_NA,,"),
      "Cell Type: MCF-7,,,",
      "Date Time,Elapsed,B1,B2",
      "8/18/25 3:02,0,300,400",
      "8/18/25 9:02,6,310,410",
      "8/18/25 15:02,12,320,NA",
      "8/18/25 21:02,NA,330,430"
    ),
    file_csv_2_na_path
  )

  writeLines(c(
    paste("Vessel Name\tTest Plate 3\t\t"),
    paste("Barcode\tPLATE_003_TSV\t\t"),
    paste("Cell Type\tA549\t\t"),
    paste("Date Time", "Elapsed", "C1", "C2", sep = "\t"),
    paste("8/18/25 3:02", "0", "500", "600", sep = "\t"),
    paste("8/18/25 9:02", "6", "510", "610", sep = "\t")
  ),
  file_tsv_3_path)

  xlsx_df <- data.frame(
    V1 = c(
      "Vessel Name",
      "Barcode",
      "Notes",
      "Date Time",
      "8/18/25 3:02",
      "8/18/25 9:02"
    ),
    V2 = c("Test Plate 4", "PLATE_004_XLSX", NA, "Elapsed", "0", "6"),
    V3 = c(NA, NA, NA, "D1", "700", "710"),
    V4 = c(NA, NA, NA, "D2", "800", "810"),
    stringsAsFactors = FALSE
  )
  openxlsx::write.xlsx(xlsx_df, file_xlsx_4_path, colNames = FALSE)

  writeLines(
    c(
      "Vessel Name: Bad Plate,,",
      sprintf("Barcode,%s", "PLATE_005_BAD,"),
      "Elapsed,A1,A2",
      "0,100,200"
    ),
    file_bad_header_path
  )

  writeLines(
    c(
      "Vessel Name: Bad Barcode Plate,,,",
      "Notes: No barcode line,,,",
      "Date Time,Elapsed,A1,A2",
      "8/18/25,0,100,200"
    ),
    file_bad_barcode_path
  )

  writeLines(
    c(
      "Vessel,PLATE_CUSTOM",
      "Notes: custom test",
      "Date Time,Elapsed,A1",
      "8/18/25 3:02,0,900",
      "8/18/25 9:02,6,950"
    ),
    file_custom_header_path
  )

  writeLines(
    c(
      "Vessel Name: Colon Test,,",
      sprintf("%s:,%s", bcode_name, "PLATE_COLON_TEST,"),
      "Date Time,Elapsed,A1",
      "8/18/25 3:02,0,999"
    ),
    file_colon_header_path
  )

  on.exit(unlink(
    c(
      file_csv_1_path,
      file_csv_2_na_path,
      file_tsv_3_path,
      file_xlsx_4_path,
      file_bad_header_path,
      file_bad_barcode_path,
      file_custom_header_path
    )
  ))

  dt_csv <- load_results_Incucyte(file_csv_1_path, headers)

  expect_s3_class(dt_csv, "data.table")
  expect_equal(NROW(dt_csv), 4)

  expect_true(all(
    c(bcode_name, d_name, well_rname, well_cname, "ReadoutValue") %in% names(dt_csv)
  ))
  expect_false(any(c("Well", "Elapsed") %in% names(dt_csv)))

  expect_equal(unique(dt_csv[[bcode_name]]), "PLATE_001_CSV")
  expect_equal(sort(unique(dt_csv[[d_name]])), c(0, 6))
  expect_equal(sort(dt_csv$ReadoutValue), c(100, 110, 200, 210))
  expect_equal(sort(unique(dt_csv[[well_rname]])), "A")
  expect_equal(sort(unique(dt_csv[[well_cname]])), c("1", "2"))

  dt_tsv <- load_results_Incucyte(file_tsv_3_path, headers)

  expect_s3_class(dt_tsv, "data.table")
  expect_equal(NROW(dt_tsv), 4)

  expect_equal(unique(dt_tsv[[bcode_name]]), "PLATE_003_TSV")
  expect_equal(sort(dt_tsv$ReadoutValue), c(500, 510, 600, 610))
  expect_equal(sort(unique(dt_tsv[[well_rname]])), "C")
  expect_equal(sort(unique(dt_tsv[[well_cname]])), c("1", "2"))

  dt_xlsx <- load_results_Incucyte(file_xlsx_4_path, headers)

  expect_s3_class(dt_xlsx, "data.table")
  expect_equal(NROW(dt_xlsx), 4)

  expect_equal(unique(dt_xlsx[[bcode_name]]), "PLATE_004_XLSX")
  expect_equal(sort(dt_xlsx$ReadoutValue), c(700, 710, 800, 810))
  expect_equal(sort(unique(dt_xlsx[[well_rname]])), "D")
  expect_equal(sort(unique(dt_xlsx[[well_cname]])), c("1", "2"))

  all_files <-
    c(file_csv_1_path,
      file_csv_2_na_path,
      file_tsv_3_path,
      file_xlsx_4_path)
  dt_all <- load_results_Incucyte(all_files, headers)

  expect_s3_class(dt_all, "data.table")

  expect_equal(NROW(dt_all), 17)

  expect_equal(
    sort(unique(dt_all[[bcode_name]])),
    c(
      "PLATE_001_CSV",
      "PLATE_002_CSV_NA",
      "PLATE_003_TSV",
      "PLATE_004_XLSX"
    )
  )

  expect_false(any(is.na(dt_all$ReadoutValue)))
  expect_false(any(is.na(dt_all[[d_name]])))

  missing_file <- "/non/existent/file.csv"
  expect_error(
    load_results_Incucyte(missing_file, headers),
    "Error reading /non/existent/file.csv"
  )

  expect_error(
    load_results_Incucyte(file_bad_header_path, headers),
    "Invalid header in the result file: (missing 'Date Time' column)",
    fixed = TRUE
  )

  expect_error(
    load_results_Incucyte(file_bad_barcode_path, headers),
    "Invalid header in the result file: (missing 'Barcode' column)",
    fixed = TRUE
  )

  dt_colon <- load_results_Incucyte(file_colon_header_path, headers)

  expect_s3_class(dt_colon, "data.table")
  expect_equal(unique(dt_colon[[bcode_name]]), "PLATE_COLON_TEST")
  expect_equal(dt_colon$ReadoutValue, 999)

})

test_that("load_results_EnVision_new works as expected", {
  headers <- gDRutils::get_env_identifiers()
  bcode_name <- headers$barcode[1]

  file_envision_multi <- tempfile(fileext = ".csv")
  file_envision_error <- tempfile(fileext = ".csv")
  file_envision_xlsx <- tempfile(fileext = ".xlsx")

  writeLines(
    c(
      "Instrument Results from;;;;;;",
      "Protocol Name;CyQuant 96w;;;;;",
      ";;;;;;",
      "Result of Fluorescence Intensity Filter 1;;;;;;",
      "Plate Barcode;Loop;Repeat;Point X;Point Y;Exc WL[nm];",
      "P01;;1;;;485 / 20;",
      ";;;;;;",
      ";1;2;3",
      "A;10;20;30",
      "B;40;OVR;60",
      "C;70;;90",
      ";;;;;;",
      "Measurement Information;;;;;;",
      "Measurement GUID;;747f1a39-507e-4049;;;;",
      "Result of Fluorescence Intensity Filter 1;;;;;;",
      "Plate Barcode;Loop;Repeat;Point X;Point Y;Exc WL[nm];",
      "P02;;1;;;485 / 20;",
      ";;;;;;",
      ";1;2;3",
      "A;11;21;31",
      "B;41;51;61",
      "C;71;81;91",
      rep(";;;;;;", 16),
      "Plate Map;;;;;;",
      ";1;2;3",
      "A;-;-;-",
      "B;-;-;-",
      "C;-;-;-"
    ),
    file_envision_multi
  )

  writeLines(
    c(
      "Instrument Results from;;;;;;",
      "Plate Barcode;;ERROR_PLATE;;;;"
    ),
    file_envision_error
  )

  xlsx_df <- data.frame(
    V1 = c("Instrument Results from", "Protocol Name", NA, "Result of Fluorescence",
           "Plate Barcode", "P03", NA, NA, "A", "B", "C"),
    V2 = c(NA, "CyQuant 96w", NA, NA, "Loop", NA, NA, "1", "12", "42", "72"),
    V3 = c(NA, NA, NA, NA, "Repeat", "1", NA, "2", "22", "52", "82"),
    V4 = c(NA, NA, NA, NA, "Point X", NA, NA, "3", "32", "62", "92"),
    V5 = c(NA, NA, NA, NA, "Point Y", NA, NA, NA, NA, NA, NA),
    V6 = c(NA, NA, NA, NA, "Exc WL[nm]", "485 / 20", NA, NA, NA, NA, NA),
    stringsAsFactors = FALSE
  )
  openxlsx::write.xlsx(xlsx_df, file_envision_xlsx, colNames = FALSE)

  on.exit(unlink(c(file_envision_multi, file_envision_error, file_envision_xlsx)))

  dt_multi <- expect_warning(
    load_results_EnVision_new(file_envision_multi, headers),
    NA
  )

  expect_s3_class(dt_multi, "data.table")
  expect_equal(NROW(dt_multi), 18)

  expect_equal(unique(dt_multi[[bcode_name]]), c("P01", "P02"))

  expect_true(all(c(bcode_name, "WellRow", "WellColumn", "ReadoutValue", "BackgroundValue") %in% names(dt_multi)))

  plate1 <- dt_multi[dt_multi[[bcode_name]] == "P01", ]
  expect_equal(plate1[WellRow == "A" & WellColumn == 1, ReadoutValue], 10)
  expect_true(is.na(plate1[WellRow == "B" & WellColumn == 2, ReadoutValue]))
  expect_true(is.na(plate1[WellRow == "C" & WellColumn == 2, ReadoutValue]))

  plate2 <- dt_multi[dt_multi[[bcode_name]] == "P02", ]
  expect_false(any(is.na(plate2$ReadoutValue)))
  expect_equal(plate2[WellRow == "A" & WellColumn == 1, ReadoutValue], 11)
  expect_equal(plate2[WellRow == "C" & WellColumn == 3, ReadoutValue], 91)

  dt_xlsx <- expect_warning(
    load_results_EnVision_new(file_envision_xlsx, headers),
    NA
  )

  expect_s3_class(dt_xlsx, "data.table")
  expect_equal(NROW(dt_xlsx), 9)
  expect_equal(unique(dt_xlsx[[bcode_name]]), "P03")

  plate3 <- dt_xlsx[dt_xlsx[[bcode_name]] == "P03", ]
  expect_equal(plate3[WellRow == "A" & WellColumn == 1, ReadoutValue], 12)
  expect_equal(plate3[WellRow == "C" & WellColumn == 3, ReadoutValue], 92)

  expect_error(
    load_results_EnVision_new(file_envision_error, headers),
    "Could not find data matrix header"
  )
})

#' is_readable_v
#' Check if all paths in vector are readable
#'
#' @param paths a character with path(s)
#' @keywords utils
#'
#' @examples
#' td2 <- get_test_Tecan_data()
#' is_readable_v(td2$r_files)
#'
#' @export
#'
#' @return \code{NULL} invisibly.
#'
is_readable_v <- function(paths) {
  checkmate::assert_character(paths)
  missing_path_string <- paste(paths[as.logical(-file.access(paths, 4))], collapse = ", ", sep = "   ")
  message <- paste0("Following path(s) with no read permission found: '", missing_path_string, "'")
  assertthat::assert_that(sum(file.access(paths, 4)) == 0, msg = message)
}

#' Load, convert and process the level 5 PRISM data into a gDR input
#'
#' @param prism_data_path path to PRISM LEVEL5 csv file with data
#' @param meta_data_path path to metadata file describing all cancer models/cell lines
#' which are referenced by a dataset contained within the DepMap portal
#' @param readout_min minimum ReadoutValue
#'
#' @return \code{data.table} object with input data for gDR pipeline
#' @keywords prism_conversion
#'
#' @examples
#'  prism_data <- system.file("testdata/prism_sa.csv", package = "gDRimport")
#'  prism_meta <- system.file("testdata/prism_model.csv", package = "gDRimport")
#'  convert_LEVEL5_prism_to_gDR_input(prism_data, prism_meta)
#'
#' @export
convert_LEVEL5_prism_to_gDR_input <- function(prism_data_path,
                                              meta_data_path,
                                              readout_min = 1.03) {

  checkmate::check_file_exists(prism_data_path)
  gDRutils::reset_env_identifiers()
  idfs <- gDRutils::get_env_identifiers()

  data <- data.table::fread(prism_data_path)
  meta <- data.table::fread(meta_data_path)

  checkmate::assert_names(names(meta),
                          must.include = c("ModelID",
                                           "CCLEName",
                                           "OncotreeLineage"))

  # Define the mapping for old column names
  column_mappings <- list(
    LFC_cb = c("LFC_cb", "LFC.cb", "LFC", "l2fc"),
    pert_iname = c("pert_iname", "pert_name"),
    pert_time = c("pert_time", "day"),
    pert2_iname = c("pert2_name", "pert2_iname", "drug2", "compound2"),
    pert2_dose = c("pert2_dose", "concentration2", "dose2")
  )

  # Rename columns based on mapping if default column is missing
  for (col in names(column_mappings)) {
    if (!col %in% names(data)) {
      for (old_col in column_mappings[[col]]) {
        if (old_col %in% names(data)) {
          data.table::setnames(data, old_col, col)
          break
        }
      }
    }
  }

  if (!"ccle_name" %in% names(data)) {
    data$ccle_name <- meta$CCLEName[match(data$depmap_id, meta$ModelID)]
  }

  checkmate::assert_names(names(data), must.include = c("ccle_name",
                                                        "pert_iname",
                                                        "pert_dose",
                                                        "pert_time",
                                                        "LFC_cb"))

  data <- data[data$ccle_name != "", ]

  if ("pert2_iname" %in% names(data) && "pert2_dose" %in% names(data)) {
    data[, (idfs$drug) := pert_iname]
    data[, (idfs$concentration) := as.numeric(pert_dose)]

    data[, (idfs$drug2) := pert2_iname]
    data[, (idfs$concentration2) := as.numeric(pert2_dose)]

  } else {
    if (any(grepl("\\|", data$pert_iname))) {
      separator <- "|"
    } else {
      separator <- "_"
    }

    data[, unlist(idfs[c("drug", "drug2")]) :=
           data.table::tstrsplit(data$pert_iname, separator, fixed = TRUE)]

    data[, unlist(idfs[c("concentration", "concentration2")]) :=
           data.table::tstrsplit(data$pert_dose, separator, fixed = TRUE, type.convert = TRUE)]
  }

  data <- meta[, .SD, .SDcols =  c("ModelID",
                                   "CCLEName",
                                   "OncotreeLineage")][data, on = .(CCLEName = ccle_name)]

  data[, unlist(idfs[c("cellline_parental_identifier", "cellline_subtype", "cellline_ref_div_time")]) :=
         list("unknown", "unknown", as.numeric(NA))]

  raw_data <- data.table::data.table(clid = data$CCLEName,
                                     CellLineName = data$CCLEName,
                                     Tissue = ifelse(is.na(data$OncotreeLineage),
                                                     "unknown",
                                                     data$OncotreeLineage),
                                     parental_identifier = data$parental_identifier,
                                     subtype = data$subtype,
                                     ReferenceDivisionTime = data$ReferenceDivisionTime,
                                     Duration = as.numeric(ifelse(grepl("d|^\\d+$", data$pert_time),
                                                                  as.numeric(gsub("d", "", data$pert_time)) * 24,
                                                                  gsub("H", "", data$pert_time))),
                                     ReadoutValue = pmin(readout_min, 2 ^ data$LFC_cb),
                                     BackgroundValue = 0,
                                     Gnumber = data[[idfs$drug]],
                                     Gnumber_2 = data[[idfs$drug2]],
                                     Concentration = data[[idfs$concentration]],
                                     Concentration_2 = data[[idfs$concentration2]],
                                     masked = FALSE)
  # rename columns of raw data
  data.table::setnames(raw_data, c("clid", "CellLineName", "Tissue", "parental_identifier",
                                   "subtype", "ReferenceDivisionTime", "Duration", "Gnumber",
                                   "Gnumber_2", "Concentration", "Concentration_2", "masked"),
                       unlist(idfs[c("cellline", "cellline_name", "cellline_tissue",
                                     "cellline_parental_identifier", "cellline_subtype",
                                     "cellline_ref_div_time", "duration", "drug", "drug2",
                                     "concentration", "concentration2", "masked_tag")]))
  # combo columns
  if (all(raw_data[[idfs$drug]] == raw_data[[idfs$drug2]]) &&
      all(raw_data[[idfs$concentration]] == raw_data[[idfs$concentration2]])) {
    raw_data[,  unlist(idfs[c("drug2", "concentration2")]) := NULL]
  }

  # control data
  dt_ctrl <- data.table::data.table(clid = unique(raw_data$clid),
                                    Gnumber = gDRutils::get_env_identifiers("untreated_tag")[1],
                                    Gnumber_2 = gDRutils::get_env_identifiers("untreated_tag")[1],
                                    Duration = unique(raw_data$Duration),
                                    Concentration = 0,
                                    Concentration_2 = 0,
                                    ReadoutValue = 1,
                                    BackgroundValue = 0,
                                    masked = FALSE)
  # merge raw and control
  dt_ctrl <- merge(dt_ctrl,
                   unique(raw_data[, unlist(idfs[c("cellline", "cellline_name", "cellline_tissue",
                                                   "cellline_parental_identifier", "cellline_subtype",
                                                   "cellline_ref_div_time")]), with = FALSE]), all.x = TRUE)

  # rename columns of control data
  data.table::setnames(dt_ctrl, c("clid", "Duration", "Gnumber", "Gnumber_2",
                                  "Concentration", "Concentration_2", "masked"),
                       unlist(idfs[c("cellline", "duration", "drug", "drug2",
                                     "concentration", "concentration2", "masked_tag")]))
  # combo columns
  if (!all(unlist(idfs[c("drug2", "concentration2")]) %in% names(raw_data))) {
    dt_ctrl[, unlist(idfs[c("drug2", "concentration2")]) := NULL]
  }
  # final
  rbind(raw_data, dt_ctrl)
}



#' Load, convert and process the level 6 PRISM data into a gDR input
#'
#' @param prism_data_path path to PRISM LEVEL6 csv file with collapsed
#' log fold change data
#' @param cell_line_data_path path to cell line info data
#' @param treatment_data_path path to collapsed treatment info data
#' @param meta_data_path path to metadata file describing all cancer models/cell lines
#' which are referenced by a dataset contained within the DepMap portal
#' @param readout_min minimum ReadoutValue
#'
#' @return \code{data.table} object with input data for gDR pipeline
#'
#' @keywords prism_conversion
#'
#' @examples
#'  prism_data_path <- system.file("testdata/prism_collapsed_LOGFC.csv", package = "gDRimport")
#'  cell_line_data_path <- system.file("testdata/prism_cell_lines.csv", package = "gDRimport")
#'  treatment_data_path <- system.file("testdata/prism_treatment.csv", package = "gDRimport")
#'  prism_meta <- system.file("testdata/prism_model.csv", package = "gDRimport")
#'  convert_LEVEL6_prism_to_gDR_input(prism_data_path, cell_line_data_path, treatment_data_path, prism_meta)
#'
#' @export
convert_LEVEL6_prism_to_gDR_input <- function(prism_data_path,
                                              cell_line_data_path,
                                              treatment_data_path,
                                              meta_data_path,
                                              readout_min = 1.03) {

  checkmate::assert_file_exists(prism_data_path)
  checkmate::assert_file_exists(cell_line_data_path)

  checkmate::assert_file_exists(treatment_data_path)
  checkmate::assert_file_exists(meta_data_path)

  gDRutils::reset_env_identifiers()
  idfs <- gDRutils::get_env_identifiers()

  cell_lines <- data.table::fread(cell_line_data_path)
  treatment <- data.table::fread(treatment_data_path)
  res <- data.table::fread(prism_data_path)
  meta <- data.table::fread(meta_data_path)

  checkmate::assert_names(names(meta),
                          must.include = c("ModelID",
                                           "CCLEName",
                                           "OncotreeLineage"))
  # rename cell_lines and treatment
  if ("depmap_id" %in% names(cell_lines)) {
    data.table::setnames(cell_lines, "depmap_id", "row_name", skip_absent = TRUE)
  } else {
    data.table::setnames(cell_lines, "row_id", "row_name", skip_absent = TRUE)
  }
  data.table::setnames(treatment,
                       c("profile_id", "SampleID", "CompoundName", "GeneSymbolOfTargets"),
                       c("column_name", "column_name", "name", "moa"),
                       skip_absent = TRUE)

  checkmate::assert_names(names(cell_lines), must.include = "row_name")
  checkmate::assert_names(names(treatment), must.include = "column_name")

  if (!"LFC" %in% names(res)) {
    checkmate::assert_names(names(res), must.include = "V1")
    res <- data.table::melt(res, id.vars = "V1")
    data.table::setnames(res,
                         c("V1", "variable"),
                         c("row_name", "column_name"))

    if (all(grepl("::", res$column_name)) && !"dose" %in% names(treatment)) {
      res[, c("column_name", "dose") := data.table::tstrsplit(column_name, "::", keep = c(1, 2))]
    }
  } else {
    data.table::setnames(res, c("row_id", "profile_id"), c("row_name", "column_name"))
  }

  # add meta data to cell_line
  cell_lines <- meta[, .SD, .SDcols = c("ModelID",
                                        "CCLEName",
                                        "OncotreeLineage")][cell_lines, on = .(ModelID = row_name)]
  cell_lines[, unlist(idfs[c("cellline_parental_identifier", "cellline_subtype", "cellline_ref_div_time")]) :=
               list("unknown", "unknown", as.numeric(NA))]
  cell_lines[, (idfs[["cellline_name"]]) := CCLEName]
  cell_lines[, OncotreeLineage := data.table::fcoalesce(OncotreeLineage, "unknown")]
  # rename cell_lines
  data.table::setnames(cell_lines,
                       c("CCLEName", "OncotreeLineage"),
                       unlist(idfs[c("cellline", "cellline_tissue")]))

  # merge results with cell_line data
  res$row_name <- gsub("::.*", "", res$row_name)

  full_data <- merge(res,
                     unique(cell_lines[, c("ModelID",
                                           unlist(idfs[c("cellline",
                                                         "cellline_name",
                                                         "cellline_tissue",
                                                         "cellline_parental_identifier",
                                                         "cellline_subtype",
                                                         "cellline_ref_div_time")])), with = FALSE]),
                     all.x = TRUE,
                     by.x = "row_name",
                     by.y = "ModelID")


  # merge results with treatment data
  full_data <- merge(full_data,
                     unique(treatment[, intersect(names(treatment),
                                                  c("column_name", "name", "dose", "broad_id", "moa")),
                                      with = FALSE]),
                     all.x = TRUE,
                     by = "column_name")
  value_col <- if ("LFC" %in% names(full_data)) {
    "LFC"
  } else {
    "value"
  }
  full_data$value <- pmin(readout_min, 2 ^ full_data[[value_col]])
  full_data <- full_data[!(is.na(name) | is.na(value))]

  # data for conc = 0
  untrt_tag <- gDRutils::get_env_identifiers("untreated_tag")[1]
  dt_ctrl <- data.table::data.table(clid = unique(full_data$clid),
                                    Gnumber = untrt_tag,
                                    DrugName = untrt_tag,
                                    drug_moa = untrt_tag,
                                    Duration = NA_real_,
                                    Concentration = 0,
                                    ReadoutValue = 1,
                                    masked = FALSE)
  # control data
  dt_ctrl <- merge(dt_ctrl,
                   unique(full_data[, unlist(idfs[c("cellline", "cellline_name", "cellline_tissue",
                                                    "cellline_parental_identifier", "cellline_subtype",
                                                    "cellline_ref_div_time")]), with = FALSE]), all.x = TRUE)

  # data for treatment
  ls_col <- intersect(c("clid", "name", "broad_id", "moa", "dose", "value",
                        unlist(idfs[c("cellline", "cellline_name", "cellline_tissue",
                                      "cellline_parental_identifier", "cellline_subtype",
                                      "cellline_ref_div_time")])),
                      colnames(full_data))
  dt_trt <- data.table::copy(full_data[, c(ls_col), with = FALSE])
  if (is.null(dt_trt$moa)) dt_trt$moa <- "unknown"
  data.table::setnames(dt_trt,
                       old = c("broad_id", "name", "moa", "dose", "value"),
                       new = c("Gnumber", "DrugName", "drug_moa", "Concentration", "ReadoutValue"))
  dt_trt$Duration <- NA_real_
  dt_trt$masked <- FALSE
  data.table::setcolorder(dt_trt, neworder = colnames(dt_ctrl))

  # merge treatment and control
  merged_data <- rbind(dt_trt, dt_ctrl)
  data.table::setnames(merged_data,
                       c("clid", "Gnumber", "DrugName", "drug_moa",
                         "Duration", "Concentration", "masked"),
                       unlist(idfs[c("cellline", "drug", "drug_name",
                                     "drug_moa", "duration", "concentration",
                                     "masked_tag")]))

  # final
  merged_data
}

#' Import D300
#'
#' This functions takes a D300 file and generates corresponding template files
#'
#' @param D300_file character, file path to D300 file
#' @param destination_path character, path to folder where template
#' files will be generated
#' @param metadata_file character, file path to file with mapping from D300 names to Gnumbers. Defaults to NULL.
#' @param day0 logical, if TRUE, creates a template file for Day 0 data filled with vehicles in addition to
#' the standard plates. Defaults to FALSE.
#' @keywords D300
#'
#' @examples
#' td3 <- get_test_D300_data()[["f_96w"]]
#' o_path <- file.path(tempdir(), "td3")
#' dir.create(o_path)
#' import_D300(td3$d300, o_path, td3$Gnum)
#' list.files(o_path)
#' unlink(o_path, recursive = TRUE)
#'
#' @return
#' Create one Excel file per plate. Each sheet in each plate file describes
#' the drugs and corrresponding concentrations of what was tested in each well.
#' @details
#' For example, wells treated with 2 drugs in combination will result in 4 sheets per plate.
#' \itemize{
#'  \item{Sheet 1: Drug 1}
#'  \item{Sheet 2: Conc of Drug 1}
#'  \item{Sheet 3: Drug 2}
#'  \item{Sheet 4: Conc of Drug 2}
#' }
#'
#' @export
#'
import_D300 <- function(D300_file,
                        destination_path,
                        metadata_file = NULL,
                        day0 = FALSE) {
  assertthat::assert_that(is.character(destination_path), msg = "'destination_path' must be a character vector")
  assertthat::assert_that(assertthat::is.readable(destination_path),
                          msg = "'destination_path' must be a readable path")

  # Parse the D300 file first
  D300 <- parse_D300_xml(D300_file)
  D300 <- fill_NA(D300, from = "D300_Barcode", with = "D300_Plate_N")

  idfs <- list(
    untreated_tags = gDRutils::get_env_identifiers("untreated_tag"),
    drug_identifier = gDRutils::get_env_identifiers("drug"),
    conc_identifier = gDRutils::get_env_identifiers("concentration")) #standard identifiers

  has_meta <- !is.null(metadata_file)

  # Conditionally process metadata if provided
  if (has_meta) {
    Gnums <- parse_D300_metadata_file(metadata_file)
    treatment <- merge_D300_w_metadata(D300, Gnums)
  } else {
    # Use the original identifiers straight from the D300 file
    treatment <- D300
    treatment[[idfs$drug_identifier]] <- treatment$Name
  }

  req_cols <- c("Row", "Col")
  if (!all(present <- req_cols %in% colnames(treatment))) {
    stop(sprintf("missing required columns from D300 file: '%s'", toString(req_cols[!present])))
  }

  if (day0) {
    # Safely extract actual plate dimensions from the XML Dimension tag (e.g. "(8,12)")
    # This works reliably regardless of whether metadata was merged or not.
    dim_str <- treatment$Dimension[1]
    dims <- as.integer(strsplit(gsub("\\(|\\)", "", dim_str), ",")[[1]])
    nrow_plate <- dims[1]
    ncol_plate <- dims[2]

    # Fallback to absolute max if dimension string is somehow corrupt
    if (is.na(nrow_plate) || is.na(ncol_plate)) {
      nrow_plate <- max(as.numeric(treatment$Row), na.rm = TRUE)
      ncol_plate <- max(as.numeric(treatment$Col), na.rm = TRUE)
    }

    wb <- openxlsx::createWorkbook()

    # Initialize with empty strings
    drug_mat <- matrix("", nrow = nrow_plate, ncol = ncol_plate)
    conc_mat <- matrix("", nrow = nrow_plate, ncol = ncol_plate)

    for (m in seq_len(nrow_plate)) {
      for (n in seq_len(ncol_plate)) {
        if (has_meta) {
          # Legacy logic: fill full bounding box (Unit Tests expect this)
          drug_mat[m, n] <- idfs$untreated_tags[[1]]
          conc_mat[m, n] <- 0.0
        } else {
          # New logic: explicitly respect the outer edge blanking rule
          is_edge <- (m == 1 || m == nrow_plate || n == 1 || n == ncol_plate)
          if (is_edge) {
            drug_mat[m, n] <- ""
            conc_mat[m, n] <- ""
          } else {
            drug_mat[m, n] <- idfs$untreated_tags[[1]]
            conc_mat[m, n] <- 0.0
          }
        }
      }
    }

    openxlsx::addWorksheet(wb, idfs$drug_identifier)
    openxlsx::writeData(wb, sheet = 1, data.table::data.table(drug_mat), colNames = FALSE)

    openxlsx::addWorksheet(wb, idfs$conc_identifier)
    openxlsx::writeData(wb, sheet = 2, data.table::data.table(conc_mat), colNames = FALSE)

    fname <- "trt_day0.xlsx"
    openxlsx::saveWorkbook(wb, file.path(destination_path, fname), overwrite = TRUE)
  }

  # Sort only the plate list numerically to ensure trt_1, trt_2 files generate in chronological order
  uplates <- sort(as.numeric(unique(treatment$D300_Plate_N)))

  existing_files <- list.files(destination_path, pattern = "^trt_P\\d+\\.xlsx$")

  # Calculate the starting offset
  max_idx <- 0
  if (length(existing_files) > 0) {
    nums <- as.numeric(gsub("trt_P|\\.xlsx", "", existing_files))
    max_idx <- max(nums, na.rm = TRUE)
  }

  for (i in seq_along(uplates)) {
    wb <- openxlsx::createWorkbook()
    idx <- treatment$D300_Plate_N == uplates[i] # Filter to 1 plate.
    trt_filt <- treatment[idx, ]

    # create a list with Gnumber and Concentration
    trt_filt$gn_conc <- apply(trt_filt, 1, function(x) list(x[idfs$drug_identifier], x[idfs$conc_identifier]))
    trt_gnumber_conc <- data.table::dcast(trt_filt, Row ~ Col,
                                          value.var = c("gn_conc"),
                                          fun.aggregate = list)
    rownames_trt_gnumber_conc <- trt_gnumber_conc$Row
    trt_gnumber_conc <- trt_gnumber_conc[, setdiff(colnames(trt_gnumber_conc), "Row"), with = FALSE]

    # count number of drugs,conc in each well
    trt_n_drugs <- apply(trt_gnumber_conc, c(1, 2), function(x) length(x[[1]]))

    # Extract actual plate dimensions from the XML Dimension tag (e.g. "(8,12)")
    dim_str <- trt_filt$Dimension[1]
    dims <- as.integer(strsplit(gsub("\\(|\\)", "", dim_str), ",")[[1]])

    trt_info <- list(
      max_drugs_per_well =  max(trt_n_drugs),
      col_idx = strtoi(colnames(trt_gnumber_conc)),
      row_idx = strtoi(rownames_trt_gnumber_conc),
      plate_nrow = dims[1],
      plate_ncol = dims[2],
      has_metadata = has_meta
    )
    save_drug_info_per_well(trt_info, trt_gnumber_conc, wb, idfs)
    current_file_num <- max_idx + i
    fname <- sprintf("trt_P%d.xlsx", current_file_num)

    openxlsx::saveWorkbook(wb, file.path(destination_path, fname), overwrite = TRUE)
  }
}

#' for each drug create a Gnumber and Concentration information for each well
#'
#' @param trt_info list with treatment info
#' @param trt_gnumber_conc list with treatment data
#' @param wb pointer to xlsx workbook
#' @param idfs charvec with identifiers
#' @keywords D300
#'
#' @return \code{NULL} invisibly.
#'
save_drug_info_per_well <-
  function(trt_info, trt_gnumber_conc, wb, idfs) {

    # Toggle dimensions based on metadata presence to preserve legacy unit tests
    n_rows <- if (trt_info$has_metadata) max(trt_info$row_idx) else trt_info$plate_nrow
    n_cols <- if (trt_info$has_metadata) max(trt_info$col_idx) else trt_info$plate_ncol
    nwells <- n_rows * n_cols

    for (j in seq_len(trt_info$max_drugs_per_well)) {

      drug_sname <- idfs$drug_identifier
      conc_sname <- idfs$conc_identifier
      if (j != 1L) {
        drug_sname <- paste0(drug_sname, "_", j)
        conc_sname <- paste0(conc_sname, "_", j)
      }

      # Initialize with empty strings to guarantee cells are created in Excel
      conc_mat <- matrix("", nrow = n_rows, ncol = n_cols)
      drug_mat <- matrix("", nrow = n_rows, ncol = n_cols)

      if (trt_info$has_metadata) {
        # -------------------------------------------------------------
        # LEGACY LOGIC: Used for unit tests and when Metadata is supplied
        # -------------------------------------------------------------
        for (m in seq_along(trt_info$row_idx)) {
          for (n in seq_along(trt_info$col_idx)) {
            drug_entry <- trt_gnumber_conc[[m, n]]
            if (length(drug_entry) >= j) {
              drug <- drug_entry[[j]][[1]]
              conc <- drug_entry[[j]][[2]]
              if (drug %in% idfs$untreated_tags) conc <- 0.0
            } else {
              drug <- idfs$untreated_tags[[1]]
              conc <- 0.0
            }
            conc_mat[trt_info$row_idx[m], trt_info$col_idx[n]] <- conc
            drug_mat[trt_info$row_idx[m], trt_info$col_idx[n]] <- drug
          }
        }
      } else {
        # -------------------------------------------------------------
        # NEW LOGIC: Full 96-well expansion when Metadata is NULL
        # -------------------------------------------------------------
        for (m in seq_len(n_rows)) {
          for (n in seq_len(n_cols)) {
            r_idx <- which(trt_info$row_idx == m)
            c_idx <- which(trt_info$col_idx == n)

            if (length(r_idx) > 0 && length(c_idx) > 0) {
              drug_entry <- trt_gnumber_conc[[r_idx, c_idx]]
            } else {
              drug_entry <- list()
            }

            if (length(drug_entry) >= j) {
              drug <- drug_entry[[j]][[1]]
              conc <- drug_entry[[j]][[2]]
              if (drug %in% idfs$untreated_tags) conc <- 0.0
            } else {
              # Custom rule: inner gaps = vehicle, outer edges = empty
              is_edge <- (m == 1 || m == n_rows || n == 1 || n == n_cols)
              if (is_edge) {
                drug <- ""
                conc <- ""
              } else {
                drug <- idfs$untreated_tags[[1]]
                conc <- 0.0
              }
            }
            conc_mat[m, n] <- conc
            drug_mat[m, n] <- drug
          }
        }
      }

      drug_data <- data.table::data.table(drug_mat)
      conc_data <- data.table::data.table(conc_mat)

      openxlsx::addWorksheet(wb, drug_sname)
      openxlsx::writeData(wb, sheet = (j * 2) - 1, drug_data, colNames = FALSE)
      openxlsx::addWorksheet(wb, conc_sname)
      openxlsx::writeData(wb, sheet = (j * 2), conc_data, colNames = FALSE)
    }
  }


merge_D300_w_metadata <- function(D300, Gnums) {
  validate_columns <- function(col, df) {
    if (!col %in% colnames(df)) {
      stop(sprintf("missing required column: '%s' in '%s'", col, quote(df)))
    }
    invisible(NULL)
  }

  merge_trt_col <- "Name"
  validate_columns(merge_trt_col, D300)

  merge_metadata_col <- "D300_Label"
  validate_columns(merge_metadata_col, Gnums)

  # Restored default sorting (sort = TRUE implicit) to maintain multi-drug combination order
  merge(D300, Gnums, by.x = merge_trt_col, by.y = merge_metadata_col, all.x = TRUE)
}


#########
# D300
#########

#' Parse D300
#'
#' This function parses a D300 *.tdd file (XML format) into a data.table
#'
#' @param D300_file string, file path to D300 .tdd file
#' @keywords D300
#'
#' @return data.table representing input \code{D300_file}.
#'
#' @examples
#' td3 <- get_test_D300_data()
#' fs <- td3[["f_96w"]]
#' dose_df <- parse_D300_xml(fs[["d300"]])
#
#' @export
#'
parse_D300_xml <- function(D300_file) {
  assertthat::assert_that(is.character(D300_file), msg = "'D300_file' must be a character vector")
  assertthat::assert_that(assertthat::is.readable(D300_file), msg = "'D300_file' must be a readable path")

  # Open D300 XML format.
  D300_xml.tree <- XML::xmlTreeParse(D300_file, useInternal = TRUE)
  top <- XML::xmlRoot(D300_xml.tree)

  # Safely retrieve units (prevents UseMethod error if node is missing).
  node_vol <- top[["VolumeUnit"]]
  vol_unit  <- if (!is.null(node_vol)) XML::xmlValue(node_vol) else NA

  node_conc <- top[["ConcentrationUnit"]]
  conc_unit <- if (!is.null(node_conc)) XML::xmlValue(node_conc) else NA

  node_mol <- top[["MolarityConcentrationUnit"]]
  mol_conc_unit <- if (!is.null(node_mol)) XML::xmlValue(node_mol) else NA

  # Handle missing ConcentrationUnit in newer D300 software versions
  if (is.na(conc_unit)) {
    conc_unit <- mol_conc_unit
  }

  # Assertions.
  if (!is.na(conc_unit) && !is.na(mol_conc_unit)) {
    assertthat::assert_that(conc_unit == mol_conc_unit,
                            msg = "Mismatch between the units for ConcentrationUnit and MolarityConcentrationUnit")
  }

  # if there is DMSO backfill defined throw a warning, support not yet implemented
  backfills <- XML::xpathSApply(top, ".//Backfills/Backfill")
  if (length(backfills) > 0) {
    warning("Backfill identified in D300 but not supported.")
  }

  id_col <- "ID"
  df_drug <- get_D300_xml_drugs(top, id_col)
  df_trt <- get_D300_xml_treatments(top, id_col, vol_unit, conc_unit)

  # Restored default sorting (sort = TRUE implicit) to maintain multi-drug combination order
  df_D300 <- merge(df_trt, df_drug, by.x = id_col, by.y = id_col, all.x = TRUE)
  df_D300
}


get_D300_xml_drugs <-
  function(xml_tree_root, id_col = "ID") {

    drug_cols <- c(id_col, "Name", "Stock_Conc", "Stock_Unit")

    # Safely extract information for every fluid (i.e. drugs) using XPath
    fluids <- XML::xpathSApply(xml_tree_root, ".//Fluids/Fluid")
    nfluids <- length(fluids)
    df_drug <- vector("list", nfluids)

    for (fi in seq_len(nfluids)) {
      fluid <- fluids[[fi]]
      id <- XML::xmlAttrs(fluid)[[id_col]]

      node_name <- fluid[["Name"]]
      name <- if (!is.null(node_name)) XML::xmlValue(node_name) else ""

      node_stock <- fluid[["Concentration"]]
      stock_conc <- if (!is.null(node_stock)) XML::xmlValue(node_stock) else NA

      node_conc_unit <- fluid[["ConcentrationUnit"]]
      fluid_conc_unit <- if (!is.null(node_conc_unit)) XML::xmlValue(node_conc_unit) else NA

      df_drug[[fi]] <- data.table::data.table(t(c(id, name, stock_conc, fluid_conc_unit)))
      colnames(df_drug[[fi]]) <- drug_cols
    }
    data.table::rbindlist(df_drug)
  }


get_plate_info <- function(plate, vol_unit) {

  rows_plate <- XML::xmlValue(plate[["Rows"]])
  cols_plate <- XML::xmlValue(plate[["Cols"]])
  plate_dim <- sprintf("(%s,%s)", rows_plate, cols_plate)
  assay_vol <- XML::xmlValue(plate[["AssayVolume"]])
  desired_unit <- get_muL()
  assay_vol_conv <- convert_units(assay_vol, from = vol_unit, to = desired_unit)

  node_name <- plate[["Name"]]
  barcode_plate <- if (!is.null(node_name)) XML::xmlValue(node_name) else ""
  if (is.na(barcode_plate)) barcode_plate <- ""

  # check if the plate is randomized; should probably be changed
  node_rand <- plate[["Randomize"]]
  randomize <- if (!is.null(node_rand)) XML::xmlValue(node_rand) else ""
  if (!is.na(randomize) && randomize != "") {
    warning("Randomization of D300 plate possibly detected, but not supported yet.")
  }

  list(
    plate_dim = plate_dim,
    desired_unit = desired_unit,
    assay_vol_conv = assay_vol_conv,
    barcode_plate = barcode_plate
  )
}


get_D300_xml_treatments <-
  function(xml_tree_root, id_col = "ID", vol_unit, conc_unit) {

    # define treatment columns
    trt_cols <- c("D300_Plate_N", "D300_Barcode", "Dimension", "Row", "Col",
                  "Volume", "Volume_Unit", id_col, "Concentration", "Unit")

    # extract drug dispensing information for each plate
    plates <- XML::xpathSApply(xml_tree_root, ".//Plates/Plate")

    pl <- lapply(seq_along(plates), function(pli) {
      plate <- plates[[pli]]
      pl_info <- get_plate_info(plate, vol_unit) # plate info

      # extract drug dispensing information for each well using XPath
      wells <- XML::xpathSApply(plate, ".//Wells/Well")

      wl <- lapply(wells, function(well) {

        well_attr <- XML::xmlAttrs(well)
        # D300 files are always 0-indexed. Hardcoding the +1 shift to match Excel matrices.
        row_well <- strtoi(well_attr[["Row"]]) + 1
        col_well <- strtoi(well_attr[["Col"]]) + 1

        # extract information each fluid delivered in well
        fluids <- XML::xpathSApply(well, ".//Fluid")
        if (length(fluids) == 0) return(NULL)

        res <- vapply(fluids, function(fluid) {
          id_fluid <- XML::xmlAttrs(fluid)[[id_col]]
          conc_fluid <- XML::xmlValue(fluid)

          # define single entry
          c(
            pli,
            pl_info$barcode_plate,
            pl_info$plate_dim,
            row_well,
            col_well,
            pl_info$assay_vol_conv,
            pl_info$desired_unit,
            id_fluid,
            conc_fluid,
            conc_unit
          )
        }, character(length(trt_cols)))

        t(res) # transpose to correctly match columns
      })

      do.call(rbind, wl)
    })

    df_trt <- data.table::data.table(do.call(rbind, pl))
    colnames(df_trt) <- trt_cols
    df_trt
  }


get_conversion_factor <- function(from, to = get_muL()) {
  if (to != get_muL()) {
    stop(sprintf("conversion to unit '%s' not supported", to))
  }

  muL <- get_muL()
  switch(from,
         "nL" = 1e-3,
         muL = 1,
         "mL" = 1e3,
         stop(sprintf("unsupported conversion factor: '%s'", from))
  )
}


convert_units <- function(x, from, to) {
  conversion_factor <- get_conversion_factor(from, to)
  as.double(x) * conversion_factor
}


#########
# Gnum
#########

parse_D300_metadata_file <- function(metadata_file) {
  if (tools::file_ext(metadata_file) %in% c("xls", "xlsx")) {
    D300_Gnum_sheets <- readxl::excel_sheets(metadata_file)
    nsheets <- length(D300_Gnum_sheets)

    # Assertions.
    assertthat::assert_that(is.character(metadata_file), msg = "'metadata_file' must be a character vector")
    assertthat::assert_that(assertthat::is.readable(metadata_file), msg = "'metadata_file' must be a readable path")

    if (nsheets != 1L) {
      futile.logger::flog.error("only one data sheet is supported, found '%s' sheets in '%s'",
                                nsheets, metadata_file)
    }

    metadata <- read_excel_to_dt(metadata_file,
                                 sheet = D300_Gnum_sheets[[1]],
                                 col_names = TRUE)
  } else {
    metadata <- data.table::fread(metadata_file, header = FALSE)
  }
  metadata
}


#########
# Utils
#########

fill_NA <- function(x, from, with) {
  idx <- is.na(x[[from]])
  if (any(idx)) {
    data.table::set(x, j = from, value = x[which(idx), ..with])
  }
  x
}

get_muL <- function() {
  # microLiter avoiding the use of non-ASCII characters for R CMD check
  paste0(rawToChar(as.raw(c(194, 181))), "L")
}

context("import_doses")
td3 <- get_test_D300_data()

test_that("parse_D300_xml", {

  # valid output returned for the D300 96 well plate example
  fs <- td3[["f_96w"]]
  dose_df <- parse_D300_xml(fs[["d300"]])
  ref_dose_df <- data.table::setDT(qs2::qs_read(fs[["ref_d300"]]))
  data.table::setorder(dose_df, Row, Col, D300_Plate_N)
  data.table::setorder(ref_dose_df, Row, Col, D300_Plate_N)
  expect_identical(dose_df, ref_dose_df)

  # valid output returned for the D300 384 well plate example
  fs2 <- td3[["f_384w"]]
  dose_df <- parse_D300_xml(fs2[["d300"]])
  ref_dose_df <- data.table::setDT(qs2::qs_read(fs2[["ref_d300"]]))
  data.table::setorder(dose_df, Row, Col, D300_Plate_N)
  data.table::setorder(ref_dose_df, Row, Col, D300_Plate_N)
  expect_identical(dose_df, ref_dose_df)

  # expected error(s) returned
  err_msg1 <- "'D300_file' must be a readable path"
  expect_error(parse_D300_xml("/non/existent_file"), err_msg1)

})

test_that("parse_D300_metadata_file works as expected", {

  fs <- td3[["f_96w"]]
  Gnum_96w_file <- parse_D300_metadata_file(fs$Gnum)
  ref_Gnum_96w_file <- qs2::qs_read(fs$ref_Gnum)
  expect_equal(Gnum_96w_file, ref_Gnum_96w_file)

  fs2 <- td3[["f_384w"]]
  Gnum_384w_file <- parse_D300_metadata_file(fs2$Gnum)
  ref_Gnum_384w_file <- qs2::qs_read(fs2$ref_Gnum)
  expect_equal(Gnum_384w_file, ref_Gnum_384w_file)
})

test_that("import_D300 works with standard metadata", {

  on.exit({
    lapply(names(td3), function(x) {
      unlink(td3[[x]][["dest_path"]], recursive = TRUE)
    })
  })

  #for 96 and 384 well plates
  for (k in seq_along(td3)) {

    # validate output files for D300 examples
    fs <- td3[[k]]
    dest_path <- fs$dest_path
    ref_path <- fs$ref_output_path
    D300_file <- fs$d300
    Gnum_file <- fs$Gnum

    # create directory if not existing
    if (!file.exists(dest_path)) {
      dir.create(dest_path, recursive = TRUE)
    }

    # run import_D300
    import_D300(D300_file, dest_path, Gnum_file)

    # test every output file against reference file
    fs_files <- list.files(path = dest_path)
    idx <- c(1, length(fs_files)) # test with first and last file
    for (i in idx) {
      output_file_path <- file.path(dest_path, fs_files[i])
      ref_file_path <- file.path(ref_path, fs_files[i])
      #load sheets
      output_sheets <- readxl::excel_sheets(output_file_path)
      ref_sheets <- readxl::excel_sheets(ref_file_path)
      #test sheet names are identical
      expect_equal(output_sheets, ref_sheets)
      #test content of sheets is identical
      for (j in seq_along(output_sheets)) {
        output_sheet <- read_excel_to_dt(output_file_path,
                                         sheet = output_sheets[[j]],
                                         col_names = FALSE)
        ref_sheet <- read_excel_to_dt(ref_file_path,
                                      sheet = ref_sheets[[j]],
                                      col_names = FALSE)
        untreated_tags <- gDRutils::get_env_identifiers("untreated_tag")
        ref_sheet[, names(ref_sheet) := lapply(.SD, function(x) {
          if (is.character(x)) {
            gsub(untreated_tags[[2]], untreated_tags[[1]], x)
          } else {
            x
          }
        }), .SDcols = names(ref_sheet)]
        expect_equal(output_sheet, ref_sheet)
      }
    }
  }
})


test_that("get_conversion_factor works as expected", {
  expect_error(get_conversion_factor("nL", "mL"), regexp = "conversion to unit 'mL' not supported")
  expect_error(get_conversion_factor("L", "µL"), regexp = "unsupported conversion factor: 'L'")

  expect_equal(get_conversion_factor("nL", "µL"), 1e-3)
})

test_that("convert_units works as expected", {
  expect_equal(convert_units(1000, from = "mL", to = "µL"), 1000000)
})

#######
# Utils
#######

test_that("fill_NA works as expected", {
  n <- 5
  df <- data.table::data.table(a = rep(NA, n), b = seq(n))
  obs <- fill_NA(df, "a", "b")
  expect_equal(obs$a, df$b)

  obs2 <- fill_NA(df, "b", "a")
  expect_equal(obs2$b, df$b)
  expect_equal(obs2$a, df$a)
})

prism_data <- system.file("testdata/prism_sa.csv", package = "gDRimport")
prism_data_2 <- system.file("testdata/prism_sa_2.csv", package = "gDRimport")
prism_data_3 <- system.file("testdata/prism_combo.csv", package = "gDRimport")
prism_data_path <- system.file("testdata/prism_collapsed_LOGFC.csv", package = "gDRimport")
cell_line_data_path <- system.file("testdata/prism_cell_lines.csv", package = "gDRimport")
treatment_data_path <- system.file("testdata/prism_treatment.csv", package = "gDRimport")
prism_meta <- system.file("testdata/prism_model.csv", package = "gDRimport")

test_that("prism level5 single-agent data can be processed into gDR input format ", {
  df_prism <- purrr::quietly(convert_LEVEL5_prism_to_gDR_input)(prism_data, prism_meta)
  expect_is(df_prism$result, "data.table")
  expect_equal(dim(df_prism$result), c(4, 12))
  expect_equal(names(df_prism$result), c("clid", "CellLineName", "Tissue", "parental_identifier", "subtype",
                                         "ReferenceDivisionTime", "Duration", "ReadoutValue", "BackgroundValue",
                                         "Gnumber", "Concentration", "masked"))
  expect_equal(df_prism$result$Duration, c(120, 240, 120, 240))
  expect_true(all(gDRutils::get_env_identifiers(c("drug", "cellline"),
                                                simplify = FALSE) %in% names(df_prism$result)))

  # testing format of clid, CellLineName and Tissue column
  expect_equal(df_prism$result$clid, rep("some_clid", 4))
  expect_equal(
    df_prism$result$Tissue,
    rep("Breast", 4))

  df_prism_unknown <- purrr::quietly(convert_LEVEL5_prism_to_gDR_input)(prism_data_2, prism_meta)
  expect_is(df_prism_unknown$result, "data.table")
  expect_equal(dim(df_prism_unknown$result), c(4, 12))

  # testing format of clid, CellLineName and Tissue column
  expect_equal(df_prism_unknown$result$clid, df_prism_unknown$result$CellLineName)
  expect_equal(df_prism_unknown$result$Tissue, rep("unknown", 4))
})

test_that("prism level5 combo data can be processed into gDR input format ", {
  df_prism <- purrr::quietly(convert_LEVEL5_prism_to_gDR_input)(prism_data_3, prism_meta)
  expect_is(df_prism$result, "data.table")
  expect_equal(dim(df_prism$result), c(2, 14))
  expect_equal(names(df_prism$result), c("clid", "CellLineName", "Tissue", "parental_identifier", "subtype",
                                         "ReferenceDivisionTime", "Duration", "ReadoutValue", "BackgroundValue",
                                         "Gnumber", "Gnumber_2", "Concentration", "Concentration_2", "masked"))
  expect_true(all(gDRutils::get_env_identifiers(c("drug", "drug2", "cellline"),
                                                simplify = FALSE) %in% names(df_prism$result)))

  # testing format of clid, CellLineName and Tissue column
  expect_equal(df_prism$result$clid, rep("some_clid", 2))
  expect_equal(
    df_prism$result$Tissue,
    rep("Breast", 2))
})

test_that("prism level6  data can be processed into gDR format ", {
  df_prism <- purrr::quietly(convert_LEVEL6_prism_to_gDR_input)(prism_data_path,
                                                                cell_line_data_path,
                                                                treatment_data_path,
                                                                prism_meta)
  expect_is(df_prism$result, "data.table")
  expect_equal(dim(df_prism$result), c(3, 13))
  expect_equal(names(df_prism$result), c("clid", "Gnumber", "DrugName", "drug_moa", "Duration", "Concentration",
                                         "ReadoutValue", "masked", "CellLineName", "Tissue", "parental_identifier",
                                         "subtype", "ReferenceDivisionTime"))
  expect_true(all(gDRutils::get_env_identifiers(c("drug", "cellline"),
                                                simplify = FALSE) %in% names(df_prism$result)))
  expect_true(all(is.na(df_prism$result$Duration)))
  expect_type(df_prism$result$Duration, "double")

  # testing format of clid, CellLineName and Tissue column
  expect_equal(df_prism$result$clid, df_prism$result$CellLineName)
  expect_equal(
    df_prism$result$Tissue,
    rep("Breast", 3))
})

#/bin/sh

echo "Executing $0"
echo "Environment: ${rp_env}"
echo "Working directory: `pwd`"
echo "Working directory contains: `ls | tr '\n' ' '`"

# exit when any command fails
set -e

CORE_DIR=/mnt/vol
echo ">>>>> OWNER OF THE REPOSITORY"
# Fix for the access permissions issues with 'R CMD' when running Docker image with 'USER 1000'
sudo chown -R `id -u` $CORE_DIR

echo ">>>>>>>> RUNNING CHECK"
Rscript -e "gDRstyle::checkPackage('gDRimport' , '/mnt/vol')"

---
title: "gDRimport"
author: "gDR team"
output: BiocStyle::html_document
vignette: >
  %\VignetteIndexEntry{gDRimport}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r init-import, echo=FALSE}
library(gDRimport)
log_level <- futile.logger::flog.threshold("ERROR")
```

# Overview

The `gDRimport` package is a part of the gDR suite. It helps to prepare raw drug response data for downstream processing. It mainly contains helper functions for importing/loading/validating dose response data provided in different file formats.

# Use Cases

## Test Data
There are currently four test datasets that can be used to see what's the expected input data for the gDRimport.

```{r load-test-datasets}
# primary test data
td1 <- get_test_data()
summary(td1)
td1

# test data in Tecan format
td2 <- get_test_Tecan_data()
summary(td2)

# test data in D300 format
td3 <- get_test_D300_data()
summary(td3)

# test data obtained from EnVision
td4 <- get_test_EnVision_data()
summary(td4)
```

## Load data
The `load_data` is the key function. It wraps `load_manifest`, `load_templates` and `load_results` functions and supports different file formats.

```{r execute-data-loading}
ml <- load_manifest(manifest_path(td1))
summary(ml)

t_df <- load_templates(template_path(td1))
summary(t_df)

r_df <- suppressMessages(load_results(result_path(td1)))
summary(r_df)

l_tbl <-
  suppressMessages(
    load_data(manifest_path(td1), template_path(td1), result_path(td1)))
summary(l_tbl)
```

# PRISM

PRISM, the Multiplexed cancer cell line screening platform, facilitates rapid screening of a broad spectrum of drugs across more than 900 human cancer cell line models, employing a high-throughput, multiplexed approach. Publicly available PRISM data can be downloaded from the DepMap website ([DepMap](https://depmap.org/portal/download/all/)).

The `gDRimport` package provides support for processing PRISM data at two levels: LEVEL5 and LEVEL6.

* LEVEL5 Data: This format encapsulates all information about drugs, cell lines, and viability within a single file. To process LEVEL5 PRISM data, you can use the `convert_LEVEL5_prism_to_gDR_input()` function. This function not only transforms and cleans the data but also executes the gDR pipeline for further analysis.

* LEVEL6 Data: In LEVEL6, PRISM data is distributed across three separate files:

prism_data: containing collapsed log fold change data for viability assays.
cell_line_data: providing information about cell lines.
treatment_data: containing treatment data.

Processing LEVEL6 PRISM data can be accomplished using the `convert_LEVEL6_prism_to_gDR_input()` function, which requires paths to these three files as input arguments.

### Processing LEVEL5 PRISM Data

To process LEVEL5 PRISM data, you can use the following function:

```{r process-prism-level5, echo=TRUE, eval=FALSE}
convert_LEVEL5_prism_to_gDR_input("path_to_file")
```

Replace "path_to_file" with the actual path to your LEVEL5 PRISM data file. This function will handle the transformation, cleaning, and execution of the gDR pipeline automatically.


### Processing LEVEL6 PRISM Data

To process LEVEL6 PRISM data, you can use the following function:

```{r process-prism-level6, echo=TRUE, eval=FALSE}
convert_LEVEL6_prism_to_gDR_input("prism_data_path", "cell_line_data_path", "treatment_data_path")
```

Replace "prism_data_path", "cell_line_data_path", and "treatment_data_path" with the respective paths to your LEVEL6 PRISM data files.


## Package installation
The function `installAllDeps` assists in installing package dependencies.

# SessionInfo {-}

```{r session-info}
sessionInfo()
```

---
title: "Converting a gDR-generated MultiAssayExperiment object into a PharmacoSet"
author:
- name: Jermiah Joseph
  email: jermiah.joseph@uhn.ca
- name: Bartosz Czech
  email: bartosz.w.czech@gmail.com
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Converting a gDR-generated MultiAssayExperiment object into a PharmacoSet}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r knitr-options, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE
)
```

```{r load-libraries, include=FALSE}
library(PharmacoGx)
library(gDRimport)
library(MultiAssayExperiment)
```



## Introduction

The `PharmacoGx` package is a popular tool from the Bioconductor project within the field of bioinformatics and computational biology.
Whereas the `gDR` package is a valuable tool that provides the framework for the import, processing, analysism and visualization of
high-dimensional drug response data, the `PharmacoGx` package provides functionality to containerize processed drug response and genomic data
to perform pharmacogenomic analyses. The data structure used by `PharmacoGx` is described as a `PharamcoSet` class, inherited from `CoreGx::CoreSet`.
Inspired by the `SummarizedExperiment` class, a new data structure class called the `TreatmentResponseExperiment` was developed to store drug response
data within the `PharmacoSet`.

In collaboration with the BHKLab, we have developed functionality to convert between the `gDR`-generated `MultiAssayExperiment (MAE)` object and the `PharmacoSet` object
(see the ConvertingPharmacoSetToGDR vignette to convert from a PharmacoSet to gDR object).


# Loading the gDR-generated MAE object
This workflow assumes that you have already generated a `MAE` object using the `gDR` package. We will load in a `MAE` object generated using
the data from the [KW. Song et al, 2022](https://doi.org/10.1158/2159-8290.CD-21-0072) study. This `MAE` contains two drugs (`treatments`) and 7 cell lines (`samples`).
``` {r load-mae-data}
(mae <- qs2::qs_read(
    system.file("extdata", "kyung2022mae", "MAE_E2.qs2", package = "gDRimport")
  )
)
```

# Converting the MAE object into a PharmacoSet object
First we will view the experiments and assays within the MultiAssayExperiment object.
``` {r inspect-mae-assays}
gDRutils::MAEpply(mae, assays)
```

We will now convert the `MAE` object into a `PharmacoSet` object. We need to provide a name for the data object, in this example
we will use the name of the study.
``` {r convert-mae-to-pset, warning=FALSE}
pset <- convert_MAE_to_PSet(mae, pset_name="Kyung2022")
```

We can now view the `PharmacoSet` object.
``` {r display-pset-object}
pset
```

We can view information about the drugs and cell lines in the `PharmacoSet` object using the `treatment` and `sample` slots:

``` {r inspect-metadata}
head(treatmentInfo(pset))

head(sampleInfo(pset))

```


# TreatmentResponseExperiment Object
The `TreatmentResponseExperiment` is a new class that was developed to store drug response data within the `PharmacoSet` object.
This vignette will cover some basics about this class, however for more information, please refer to the `PharmacoSet` vignette
[TreatmentResponseExperiment](https://bioconductor.org/packages/release/bioc/vignettes/CoreGx/inst/doc/TreatmentResponseExperiment.html)****.

## Row and Column Names
```{r inspect-tre-rows, echo=TRUE}
tre <- treatmentResponse(pset)
head(rownames(tre))
head(rowData(tre))
```

```{r inspect-tre-cols, echo=TRUE}
head(colnames(tre))
head(colData(tre))
```

## `data.table` Subsetting

In addition to regex queries, a `TreatmentResponseExperiment` object supports
arbitrarily complex subset queries using the `data.table` API. To access this
API, you will need to use the `.` function, which allows you to pass raw R
expressions to be evaluated inside the `i` and `j` arguments for
`dataTable[i, j]`.

For example if we want to subset to rows where the cell line is "CL131891_EFM-19_Breast_EFM-19_unknown_64" and drug is "G02967907_GDC-0077_PI3K-A_168".

```{r subset-tre-data-table, echo=TRUE}
tre[
  .(treatmentid=="G02967907_GDC-0077_PI3K-A_168"), # query on row
  .(sampleid=="CL131891_EFM-19_Breast_EFM-19_unknown_64") # query on column
]
```

We can also invert matches or subset on other columns in `rowData` or `colData`:

```{r invert-subset-tre, echo=TRUE}
tre[
  .(treatmentid!="G02967907_GDC-0077_PI3K-A_168"), # query on row
  .(sampleid!="CL131891_EFM-19_Breast_EFM-19_unknown_64") # query on column
]
```


## assays
The assays can be viewed through the `assays` function. Each assay has a column "data_type" which references which `SummarizedExperiment` from
the `MAE` the data corresponds with.

``` {r view-tre-assays}
lapply(assays(tre), head)
```


``` {r view-tre-assay-names}
assayNames(tre)
```


``` {r inspect-metrics-assay}
head(assay(tre, "Metrics"),3)
```

# References
1. Song, K. W., Edgar, K. A., Hanan, E. J., Hafner, M., Oeh, J., Merchant, M., ... & Friedman, L. S. (2022).
   RTK-dependent inducible degradation of mutant PI3K drives GDC-0077 (Inavolisib) efficacy. Cancer discovery, 12(1), 204-219.


```{r session-info}
sessionInfo()
```

---
title: "Converting PharmacoSet Drug Response Data into gDR object"
author:
- name: Jermiah Joseph
  email: jermiah.joseph@uhn.ca
- name: Bartosz Czech
  email: bartosz.w.czech@gmail.com
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Converting PharmacoSet Drug Response Data into gDR object}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


```{r knitr-init, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r load-deps}
library(PharmacoGx)
library(gDRimport)
```


# Overview
The `gDRimport` package is a part of the gDR suite. It helps to prepare raw drug response data for downstream processing.
It mainly contains helper functions for importing/loading/validating dose response data provided from different scanner sources.
In collaboration with the BHKLab, `gDRimport` also provides functions that can convert a `PharmacoGx::PharamcoSet` object into a gDR object.
With this functionality, users familiar with the gDR suite of packages and methods can utilize the publically available, curated datasets from the PharmacoGx database.
The main step in this process is to extract the drug dose-response data from the PharmacoSets and transform them into a `data.table` that can be used as input for the `gDRcore::runDrugResponseProcessingPipeline`.


# Loading a PharmacoSet (PSet)
Whereas a user might already have a pharmacoset loaded in their R session, if they wish to obtain a different pharmacoset or use the same script in the future we provide a helper function to do so.
It helps to have a user directory in which to store all pharmacosets, and by passing this directory into the function as a parameter, the function will also check to see if the PSet exists in the user-defined directory.
This is to ensure that the PSet is not being re-downloaded if it already has.
```{r get-pset-example, eval = FALSE}
pset <- getPSet("Tavor_2020")
pset
```

```{r load-mocked-psets, include = FALSE}
read_mocked_PSets <- function(canonical = FALSE) {
  qs2::qs_read(
    system.file("extdata", "data_for_unittests", "PSets.qs2", package = "gDRimport")
  )
}
pset <- testthat::with_mocked_bindings(
    availablePSets = read_mocked_PSets,
    .package = "PharmacoGx",
    suppressMessages(getPSet(
      "Tavor_2020", psetDir = system.file("extdata/pset", package = "gDRimport")
    ))
)
pset
```

# Converting PharmacoSet to data.table for gDR pipeline
`PharamcoSets` hold data pertaining to the cell lines (@sample slot), drugs (@treatment slot), and dose response experiments (@treatmentResponse slot).
The dose response data is stored in a `treatmentResponseExperiment` object and the function `gDRimport::convert_pset_to_df` extracts this information to
build a `data.table` that can be used as input to the gDR pipeline.

```{r convert-pset-to-dt}
# Store treatment response data in df_
dt <- convert_pset_to_df(pharmacoset = pset)
str(dt)
```

# Subsetting to extract relevant information
Most canonical `PharmacoSets` have data pertaining to many cell lines and their response to many drugs (drug-combination data is available in some but its conversion to gDR is not currently supported).
As such, in the interest of time and resources, it may be useful to subset the data before providing it as input for the gDR pipeline.

```{r subset-by-cell-line}
# example subset using only 1 cell line
subset_cl <- dt$Clid[1]
x <- dt[Clid == subset_cl]
x
```

# Running drug response pipeline with data
The subsetted data can now be used as input for the `gDRcore::runDrugResponseProcessingPipeline()`. The output of this function is a `MultiAssayExperiment` object which can be accessed with `gDRutils::convert_se_assay_to_dt()`

```{r run-gdr-pipeline, eval = FALSE}
# RUN DRUG RESPONSE PROCESSING PIPELINE
se <- gDRcore::runDrugResponseProcessingPipeline(x)
se
```

```{r convert-se-to-dt, eval = FALSE}
# Convert Summarized Experiments to data.table
# Available SEs : "RawTreatred", "Controls", "Normalized", "Averaged", "Metrics"

str(gDRutils::convert_se_assay_to_dt(se[[1]], "Averaged"))
str(gDRutils::convert_se_assay_to_dt(se[[1]], "Metrics"))
```


```{r reset-identifiers, include = FALSE}
gDRutils::reset_env_identifiers()
```


# SessionInfo {-}

```{r session-info}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{validate_template_xlsx}
\alias{validate_template_xlsx}
\title{Validate template xlsx data}
\usage{
validate_template_xlsx(template_file, template_filename, template_sheets, idx)
}
\arguments{
\item{template_file}{character, file path(s) to template(s)}

\item{template_filename}{character, file name(s)}

\item{template_sheets}{template sheet names}

\item{idx}{template file index}
}
\value{
\code{NULL} invisibly.
}
\description{
Validate template xlsx data
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pset_to_gdrDF.R
\name{.removeNegatives}
\alias{.removeNegatives}
\title{Remove negative viabilities}
\usage{
.removeNegatives(dataset)
}
\value{
data.table with positive values in column \code{ReadoutValue}
}
\description{
Remove negative viabilities
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_manifest}
\alias{load_manifest}
\title{Load manifest}
\usage{
load_manifest(manifest_file)
}
\arguments{
\item{manifest_file}{character, file path(s) to manifest(s)}
}
\value{
list with manifest data.table and headers
}
\description{
This functions loads and checks the manifest file(s)
}
\examples{
 td <- get_test_data()
 ml <- load_manifest(manifest_path(td))

}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MAE_to_PSet.R
\name{convert_MAE_to_PSet}
\alias{convert_MAE_to_PSet}
\title{Convert MultiAssayExperiment to TreatmentResponseExperiment}
\usage{
convert_MAE_to_PSet(mae, pset_name)
}
\arguments{
\item{mae}{A MultiAssayExperiment object generated by gDR.}

\item{pset_name}{A character string specifying the name of the resulting PharmacoSet object.}
}
\value{
A PharmacoSet object.
}
\description{
This function converts a MultiAssayExperiment generated by gDR into a TreatmentResponseExperiment for use in the
PharmacoGx package. The resulting PharmacoSet can be used for pharmacogenomic analysis of drug response.
}
\examples{
# Convert a MultiAssayExperiment object to a PharmacoSet object
m <- 20
n <- 10
rnames <- LETTERS[1:m]
cnames <- letters[1:n]
ref_gr_value <-  matrix(runif(m * n), nrow = m, ncol = n, dimnames = list(rnames, cnames))
se <- SummarizedExperiment::SummarizedExperiment(assays = list(RefGRvalue = ref_gr_value),
                                                 rowData = S4Vectors::DataFrame(rnames),
                                                 colData = S4Vectors::DataFrame(cnames))
mae <- MultiAssayExperiment::MultiAssayExperiment(experiments = list("single-agent" = se))
convert_MAE_to_PSet(mae, "my_pset")

}
\keyword{pset_conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prism_to_gdrDF.R
\name{convert_LEVEL5_prism_to_gDR_input}
\alias{convert_LEVEL5_prism_to_gDR_input}
\title{Load, convert and process the level 5 PRISM data into a gDR input}
\usage{
convert_LEVEL5_prism_to_gDR_input(
  prism_data_path,
  meta_data_path,
  readout_min = 1.03
)
}
\arguments{
\item{prism_data_path}{path to PRISM LEVEL5 csv file with data}

\item{meta_data_path}{path to metadata file describing all cancer models/cell lines
which are referenced by a dataset contained within the DepMap portal}

\item{readout_min}{minimum ReadoutValue}
}
\value{
\code{data.table} object with input data for gDR pipeline
}
\description{
Load, convert and process the level 5 PRISM data into a gDR input
}
\examples{
 prism_data <- system.file("testdata/prism_sa.csv", package = "gDRimport")
 prism_meta <- system.file("testdata/prism_model.csv", package = "gDRimport")
 convert_LEVEL5_prism_to_gDR_input(prism_data, prism_meta)

}
\keyword{prism_conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{read_EnVision_delim}
\alias{read_EnVision_delim}
\title{Read EnVision delimited text files}
\usage{
read_EnVision_delim(file, nrows = 10000, seps = c(",", "\\t"))
}
\arguments{
\item{file}{string to path of input file from EnVision scanner}

\item{nrows}{maximum number of file rows to be processed}

\item{seps}{potential field separators of the input file}
}
\value{
a list containing the data table, n_col, n_row, and if is edited
}
\description{
This function reads file from the EnVision Workstation
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correction.R
\name{get_xl_sheets}
\alias{get_xl_sheets}
\title{Get Excel sheets}
\usage{
get_xl_sheets(files)
}
\arguments{
\item{files}{charvec with file paths}
}
\value{
named list where names are the excel filenames
and the values are the sheets within each file
}
\description{
get sheets for given set of XLS files
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{read_in_results_Tecan}
\alias{read_in_results_Tecan}
\title{read in Tecan data}
\usage{
read_in_results_Tecan(results_file, results_sheets, headers)
}
\arguments{
\item{results_file}{string, file path to a result file}

\item{results_sheets}{template sheet names}

\item{headers}{list of headers identified in the manifest}
}
\value{
data.table derived from Tecan data
}
\description{
read in Tecan data
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{check_metadata_field_names}
\alias{check_metadata_field_names}
\title{Check metadata field names}
\usage{
check_metadata_field_names(corrected_names, df_name)
}
\arguments{
\item{corrected_names}{a charvec with corrected colnames of df}

\item{df_name}{a name of data.table ("" by default)}
}
\value{
a charvec with corrected colnames of df
}
\description{
Check metadata field names
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{detect_file_format}
\alias{detect_file_format}
\title{Detect format of results data}
\usage{
detect_file_format(results_file)
}
\arguments{
\item{results_file}{path to results data}
}
\value{
string of the detected file format
}
\description{
Detect format of results data
}
\examples{
td2 <- get_test_Tecan_data()
detect_file_format(td2$r_files[1])

}
\keyword{utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correction.R
\name{fix_typos_with_reference}
\alias{fix_typos_with_reference}
\title{Fix typos using reference data}
\usage{
fix_typos_with_reference(
  data,
  ref,
  method = c("exact", "grepl", "adist"),
  fix_underscores = FALSE
)
}
\arguments{
\item{data}{list of charvec(s) or charvec with data}

\item{ref}{charvec with reference data}

\item{method}{charvec type of the method to be used
'exact' is used to find  identical entries from 'ref' in the
data (after corrections and uppercase'ing)
'grepl' is used to find entries from 'ref' that might be
somehow pre- or post- fixed}

\item{fix_underscores}{logical flag fix the issues with underscores in data identfiers?}
}
\value{
list or charvec with corrected data
}
\description{
Fix typos using reference data
Evaluate given list of ids and try to update them
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{read_EnVision_xlsx}
\alias{read_EnVision_xlsx}
\title{Read in single xlsx data from EnVision}
\usage{
read_EnVision_xlsx(results_file, results_sheet)
}
\arguments{
\item{results_file}{character, file path(s) to results file(s)}

\item{results_sheet}{results sheet names}
}
\value{
data.table with results data
}
\description{
Read in single xlsx data from EnVision
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correction.R
\name{mgrepl}
\alias{mgrepl}
\title{grep wrapper to support multiple patterns}
\usage{
mgrepl(patterns, x, do_unlist = TRUE, ...)
}
\arguments{
\item{patterns}{charvec with patterns to be checked}

\item{x}{charvec with data}

\item{do_unlist}{logical_flag unlist the final results?}

\item{...}{additional argument}
}
\value{
list of charvec with grep output
}
\description{
grep wrapper to support multiple patterns
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_templates_tsv}
\alias{load_templates_tsv}
\title{Load templates from tsv}
\usage{
load_templates_tsv(template_file, template_filename = NULL)
}
\arguments{
\item{template_file}{character, file path(s) to template(s)}

\item{template_filename}{character, file name(s)}
}
\value{
data.table with template data
}
\description{
This functions loads and checks the template file(s)
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{get_df_from_raw_unedited_EnVision_df}
\alias{get_df_from_raw_unedited_EnVision_df}
\title{Get final results (as a data.table)
from raw unedited EnVision data.table}
\usage{
get_df_from_raw_unedited_EnVision_df(df, n_row, n_col, barcode_col)
}
\arguments{
\item{df}{raw data.table}

\item{n_row}{number of rows}

\item{n_col}{number of columns}

\item{barcode_col}{column number for barcode data}
}
\value{
data.table derived from EnVision data
}
\description{
Get final results (as a data.table)
from raw unedited EnVision data.table
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correction.R
\name{get_expected_template_sheets}
\alias{get_expected_template_sheets}
\title{Get names of the sheets expected in templates xlsx}
\usage{
get_expected_template_sheets(type = c("all", "core", "optional"))
}
\arguments{
\item{type}{charvec type of the sheets}
}
\value{
string with type of the sheets
}
\description{
Get names of the sheets expected in templates xlsx
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_results_EnVision}
\alias{load_results_EnVision}
\title{Load EnVision results from xlsx}
\usage{
load_results_EnVision(results_file, headers = gDRutils::get_env_identifiers())
}
\arguments{
\item{results_file}{character vector containing file path(s) to results file(s)}

\item{headers}{list of headers identified in the manifest}
}
\value{
data.table with results data
}
\description{
This functions loads and checks the results file(s)
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{get_EnVision_properties}
\alias{get_EnVision_properties}
\title{Get properties of EnVision data}
\usage{
get_EnVision_properties(results.list, fname)
}
\arguments{
\item{results.list}{list with EnVision data}

\item{fname}{name of the input file}
}
\value{
list with EnVision propertiesa
}
\description{
This function return properties of EnVision data
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gdr_test_data_class.R
\docType{methods}
\name{manifest_path}
\alias{manifest_path}
\alias{manifest_path,gdr_test_data-method}
\title{Method manifest_path}
\usage{
manifest_path(x)

\S4method{manifest_path}{gdr_test_data}(x)
}
\arguments{
\item{x}{object class gdr_test_data}
}
\value{
value of slot \code{manifest_path}
}
\description{
Method for object gdr_test_data - access to slot \code{manifest_path}
}
\examples{
 td <- get_test_data()
 manifest_file_path <- manifest_path(td)

}
\keyword{methods}
\keyword{test_data_class}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testdata.R
\name{get_test_D300_data}
\alias{get_test_D300_data}
\title{get test D300 data}
\usage{
get_test_D300_data()
}
\value{
list with with input data (manifest/template/result paths)
and related reference data (qs2 file paths)
}
\description{
get test D300 data
}
\examples{
get_test_D300_data()
}
\keyword{test_data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{get_plate_info_from_template_xlsx}
\alias{get_plate_info_from_template_xlsx}
\title{Get plate info from template xlsx}
\usage{
get_plate_info_from_template_xlsx(template_file, Gnumber_idx, idx)
}
\arguments{
\item{template_file}{character, file path(s) to template(s)}

\item{Gnumber_idx}{index with Gnumber data}

\item{idx}{template file index}
}
\value{
list with plate info
}
\description{
Get plate info from template xlsx
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{get_df_from_raw_edited_EnVision_df}
\alias{get_df_from_raw_edited_EnVision_df}
\title{Get final results (as a data.table)
from raw edited EnVision data.table}
\usage{
get_df_from_raw_edited_EnVision_df(
  df,
  barcode_idx,
  barcode_col,
  n_row,
  n_col,
  fname,
  sheet_name,
  headers
)
}
\arguments{
\item{df}{raw data.table}

\item{barcode_idx}{numeric vector with barcode indices}

\item{barcode_col}{column number for barcode data}

\item{n_row}{number of rows}

\item{n_col}{number of columns}

\item{fname}{file name}

\item{sheet_name}{name of the Excel sheet}

\item{headers}{list with the headersa}
}
\value{
data.table derived from EnVision data
}
\description{
Get final results (as a data.table)
from raw edited EnVision data.table
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testdata.R
\name{get_test_data}
\alias{get_test_data}
\title{get primary test data}
\usage{
get_test_data()
}
\value{
object class "gdr_test_data" with with input data (manifest/template/result paths)
and related reference data (qs2 file paths)
}
\description{
get primary test data
}
\examples{
get_test_data()

}
\keyword{test_data_class}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pset_to_gdrDF.R
\name{.extractDoseResponse}
\alias{.extractDoseResponse}
\title{Get dose and viability readouts and melt into large data table}
\usage{
.extractDoseResponse(pset)
}
\value{
data.table with dose-response data
}
\description{
Get dose and viability readouts and melt into large data table
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pset_to_gdrDF.R
\name{setEnvForPSet}
\alias{setEnvForPSet}
\title{Adjust environment variables to meet gDR standards}
\usage{
setEnvForPSet()
}
\value{
\code{NULL}
}
\description{
Adjust environment variables to meet gDR standards
}
\examples{
setEnvForPSet()
gDRutils::reset_env_identifiers()

}
\keyword{pset_conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{standardize_record_values}
\alias{standardize_record_values}
\title{standardize_record_values}
\usage{
standardize_record_values(x, dictionary = DICTIONARY)
}
\arguments{
\item{x}{a named array}

\item{dictionary}{a named array}
}
\value{
a named array with updated names
}
\description{
map values to a dictionary
}
\examples{
standardize_record_values(c("Vehicle", "vehcle"))

}
\keyword{utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{read_ref_data}
\alias{read_ref_data}
\title{read_ref_data}
\usage{
read_ref_data(inDir, prefix = "ref")
}
\arguments{
\item{inDir}{a directory path of reference data}

\item{prefix}{a prefix of reference file names ('ref' by default)}
}
\value{
a list of reference data
}
\description{
Read reference data
}
\keyword{utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{read_in_manifest_file}
\alias{read_in_manifest_file}
\title{read manifest files}
\usage{
read_in_manifest_file(manifest_file, available_formats)
}
\arguments{
\item{manifest_file}{character, file path(s) to manifest(s)}

\item{available_formats}{charvec with available file formats}
}
\value{
a data.table with manifest data
}
\description{
read manifest files
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{read_excel_to_dt}
\alias{read_excel_to_dt}
\title{Read excel file and transorm it into data.table object}
\usage{
read_excel_to_dt(path, ...)
}
\arguments{
\item{path}{path to excel file}

\item{...}{other arguments that should be passed into \code{readxl::read_excel}}
}
\value{
data.table object with read excel file
}
\description{
Read excel file and transorm it into data.table object
}
\examples{
datasets <- readxl::readxl_example("datasets.xlsx")
read_excel_to_dt(datasets)
}
\keyword{utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{check_metadata_against_spaces}
\alias{check_metadata_against_spaces}
\title{Check metadata against spaces}
\usage{
check_metadata_against_spaces(corrected_names, df_name)
}
\arguments{
\item{corrected_names}{a charvec with corrected colnames of df}

\item{df_name}{a name of data.table ("" by default)}
}
\value{
a charvec with corrected colnames of df
}
\description{
Check metadata against spaces
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{check_metadata_req_col_names}
\alias{check_metadata_req_col_names}
\title{Check metadata for required column names}
\usage{
check_metadata_req_col_names(col_df, df_name, df_type)
}
\arguments{
\item{col_df}{a charvec with corrected colnames of df}

\item{df_name}{a name of data.table ("" by default)}

\item{df_type}{a type of a data.table (NULL by default)}
}
\value{
\code{NULL} invisibly.
}
\description{
Check metadata for required column names
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/D300.R
\name{parse_D300_xml}
\alias{parse_D300_xml}
\title{Parse D300}
\usage{
parse_D300_xml(D300_file)
}
\arguments{
\item{D300_file}{string, file path to D300 .tdd file}
}
\value{
data.table representing input \code{D300_file}.
}
\description{
This function parses a D300 *.tdd file (XML format) into a data.table
}
\examples{
td3 <- get_test_D300_data()
fs <- td3[["f_96w"]]
dose_df <- parse_D300_xml(fs[["d300"]])
}
\keyword{D300}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{enhance_raw_edited_EnVision_df}
\alias{enhance_raw_edited_EnVision_df}
\title{Enhance raw edited EnVision data.table}
\usage{
enhance_raw_edited_EnVision_df(df, barcode_col, headers)
}
\arguments{
\item{df}{raw data.table}

\item{barcode_col}{column number for barcode data}

\item{headers}{list with the headersa}
}
\value{
data.table derived from EnVision data
}
\description{
Enhance raw edited EnVision data.table
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pset_to_gdrDF.R
\name{.createPseudoData}
\alias{.createPseudoData}
\title{Add in pseudo-data for duration and cell reference division time}
\usage{
.createPseudoData(dt)
}
\value{
data.table
}
\description{
Add in pseudo-data for duration and cell reference division time
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_results_tsv}
\alias{load_results_tsv}
\title{Load results from tsv}
\usage{
load_results_tsv(results_file, headers)
}
\arguments{
\item{results_file}{character, file path(s) to template(s)}

\item{headers}{list of headers identified in the manifest}
}
\value{
data.table with results data
}
\description{
This functions loads and checks the results file(s)
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gdr_test_data_class.R
\docType{methods}
\name{result_path}
\alias{result_path}
\alias{result_path,gdr_test_data-method}
\title{Method result_path}
\usage{
result_path(x)

\S4method{result_path}{gdr_test_data}(x)
}
\arguments{
\item{x}{object class gdr_test_data}
}
\value{
value of slot \code{result_path}
}
\description{
Method for object gdr_test_data - access to slot \code{result_path}
}
\examples{
 td <- get_test_data()
 result_file_path <- result_path(td)

}
\keyword{methods}
\keyword{test_data_class}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_data}
\alias{load_data}
\title{Load data}
\usage{
load_data(
  manifest_file,
  df_template_files,
  results_file,
  instrument = "EnVision"
)
}
\arguments{
\item{manifest_file}{character, file path(s) to manifest(s)}

\item{df_template_files}{data.table, with datapaths and names of results file(s)
or character with file path of templates file(s)}

\item{results_file}{data.table, with datapaths and names of results file(s)
or character with file path of results file(s)}

\item{instrument}{character}
}
\value{
a list with three data.tables for manifest/treatment and results
}
\description{
This functions loads and checks the data file(s)
}
\examples{
 td <- get_test_data()
 l_tbl <- load_data(manifest_path(td), template_path(td), result_path(td))

}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correction.R
\name{correct_template_sheets}
\alias{correct_template_sheets}
\title{Correct names of the template sheets (if required)}
\usage{
correct_template_sheets(tfiles)
}
\arguments{
\item{tfiles}{charvec with paths to template files}
}
\value{
charvec with paths to corrected sheet names
}
\description{
Correct names of the template sheets (if required)
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{.fill_empty_wells}
\alias{.fill_empty_wells}
\title{Correct plates with not fully filled readout values}
\usage{
.fill_empty_wells(
  df,
  plate_rows,
  data_rows,
  exp_row,
  exp_col,
  numeric_regex = "^\\\\d+$"
)
}
\value{
data.table with corrected plates data
}
\description{
Correct plates with not fully filled readout values
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{read_in_EnVision_file}
\alias{read_in_EnVision_file}
\title{Read EnVision file}
\usage{
read_in_EnVision_file(file, nrows, seps)
}
\arguments{
\item{file}{input file from EnVision}

\item{nrows}{maximum number of file rows to be processed}

\item{seps}{potential field separators of the input file}
}
\value{
list with one element per EnVisoin input file
}
\description{
This function reads file from the EnVision Workstation
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{check_metadata_names}
\alias{check_metadata_names}
\title{check_metadata_names}
\usage{
check_metadata_names(col_df, df_name = "", df_type = NULL)
}
\arguments{
\item{col_df}{a character with colnames of df}

\item{df_name}{a name of data.table ("" by default)}

\item{df_type}{a type of a data.table (NULL by default)}
}
\value{
a charvec with corrected colnames of df
}
\description{
Check whether all metadata names are correct
}
\examples{
 td <- get_test_data()
 m_file <- manifest_path(td)
 m_data <- read_excel_to_dt(m_file)
 result <- check_metadata_names(col_df = colnames(m_data))

}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{.get_plate_size}
\alias{.get_plate_size}
\title{Get plate size}
\usage{
.get_plate_size(df)
}
\value{
charvec with plate dims
}
\description{
Get plate size
}
\details{
All plate sizes assume 1.5x nrows = ncolumns.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/D300.R
\name{import_D300}
\alias{import_D300}
\title{Import D300}
\usage{
import_D300(D300_file, destination_path, metadata_file = NULL, day0 = FALSE)
}
\arguments{
\item{D300_file}{character, file path to D300 file}

\item{destination_path}{character, path to folder where template
files will be generated}

\item{metadata_file}{character, file path to file with mapping from D300 names to Gnumbers. Defaults to NULL.}

\item{day0}{logical, if TRUE, creates a template file for Day 0 data filled with vehicles in addition to
the standard plates. Defaults to FALSE.}
}
\value{
Create one Excel file per plate. Each sheet in each plate file describes
the drugs and corrresponding concentrations of what was tested in each well.
}
\description{
This functions takes a D300 file and generates corresponding template files
}
\details{
For example, wells treated with 2 drugs in combination will result in 4 sheets per plate.
\itemize{
\item{Sheet 1: Drug 1}
\item{Sheet 2: Conc of Drug 1}
\item{Sheet 3: Drug 2}
\item{Sheet 4: Conc of Drug 2}
}
}
\examples{
td3 <- get_test_D300_data()[["f_96w"]]
o_path <- file.path(tempdir(), "td3")
dir.create(o_path)
import_D300(td3$d300, o_path, td3$Gnum)
list.files(o_path)
unlink(o_path, recursive = TRUE)

}
\keyword{D300}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pset_to_gdrDF.R
\name{getPSet}
\alias{getPSet}
\title{Get PharmacoSet}
\usage{
getPSet(
  pset_name,
  psetDir = getwd(),
  canonical = FALSE,
  timeout = 600,
  use_local_PSets_list = FALSE
)
}
\arguments{
\item{pset_name}{string with the name of the PharmacoSet}

\item{psetDir}{string with the temporary directory for the PharmacoSet}

\item{canonical}{logical flag indicating if the PSet canonical}

\item{timeout}{maximum number of seconds allowed for PSet download}

\item{use_local_PSets_list}{logical flag if PSets list should be used from
local. If FALSE PSets list will be taken from web.}
}
\value{
PharmacoSet object
}
\description{
Get PharmacoSet
}
\examples{
suppressMessages(getPSet(
  "Tavor_2020",
  psetDir = system.file("extdata/pset", package = "gDRimport"),
  use_local_PSets_list = TRUE
))

}
\keyword{pset_conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_results_Incucyte}
\alias{load_results_Incucyte}
\title{Load incucyte results from plain text}
\usage{
load_results_Incucyte(results_file, headers = gDRutils::get_env_identifiers())
}
\arguments{
\item{results_file}{list of strings: file paths to result paths from individual plates}

\item{headers}{list of headers identified in the manifest}
}
\value{
data.table derived from Incucyte data
}
\description{
This functions loads incucyte time-course cell count file
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{read_in_tsv_template_files}
\alias{read_in_tsv_template_files}
\title{read in tsv template files}
\usage{
read_in_tsv_template_files(template_file, template_filename, templates)
}
\arguments{
\item{template_file}{character, file path(s) to template(s)}

\item{template_filename}{character, file name(s)}

\item{templates}{list with templates data}
}
\value{
data.table with templates data
}
\description{
read in tsv template files
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testdata.R
\name{get_test_Tecan_data}
\alias{get_test_Tecan_data}
\title{get test Tecan data}
\usage{
get_test_Tecan_data()
}
\value{
list with with input data (manifest/template/result paths)
and related reference data (qs2 file paths)
}
\description{
get test Tecan data
}
\examples{
get_test_Tecan_data()
}
\keyword{test_data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{read_in_template_xlsx}
\alias{read_in_template_xlsx}
\title{Read in xlsx template files}
\usage{
read_in_template_xlsx(template_file, template_filename, template_sheets)
}
\arguments{
\item{template_file}{character, file path(s) to template(s)}

\item{template_filename}{character, file name(s)}

\item{template_sheets}{template sheet names}
}
\value{
data.table with templates data
}
\description{
Read in xlsx template files
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correction.R
\name{are_template_sheets_valid}
\alias{are_template_sheets_valid}
\title{are template sheet valid?}
\usage{
are_template_sheets_valid(ts)
}
\arguments{
\item{ts}{list with (per file) template sheets}
}
\value{
logical flag
}
\description{
are template sheet valid?
}
\seealso{
get_xl_sheets
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_results}
\alias{load_results}
\title{Load results}
\usage{
load_results(
  df_results_files,
  instrument = "EnVision",
  headers = gDRutils::get_env_identifiers()
)
}
\arguments{
\item{df_results_files}{data.table, with datapaths and names of results file(s)
or character with file path of results file(s)}

\item{instrument}{character}

\item{headers}{list of headers identified in the manifest file}
}
\value{
data.table with results' data
}
\description{
This functions loads and checks the results file(s)
}
\examples{
 td <- get_test_data()
 r_df <- load_results(result_path(td))

}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pset_to_gdrDF.R
\name{convert_pset_to_df}
\alias{convert_pset_to_df}
\title{Convert a PharmacoSet to a data.table that is prepare for input into gDR pipeline}
\usage{
convert_pset_to_df(pharmacoset, run_parallel = TRUE, workers = 2L)
}
\arguments{
\item{pharmacoset}{PharmacoSet object}

\item{run_parallel}{logical, TRUE (default) if to run functions in Parallel, FALSE to run in serial}

\item{workers}{integer, number of workers defaults to 2L if run_parallel is TRUE}
}
\value{
data.table of PharmacoSet's dose response data with column names aligned with gDR standard
}
\description{
Convert a PharmacoSet to a data.table that is prepare for input into gDR pipeline
}
\examples{
pset <- suppressMessages(getPSet(
  "Tavor_2020",
  psetDir = system.file("extdata/pset", package = "gDRimport"),
  use_local_PSets_list = TRUE
))
dt <- convert_pset_to_df(pset)
gDRutils::reset_env_identifiers()

}
\author{
Jermiah Joseph -- collaboration with BHKLab
}
\keyword{pset_conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{read_in_result_files}
\alias{read_in_result_files}
\title{Read in results files}
\usage{
read_in_result_files(results_file, results_filename, headers)
}
\arguments{
\item{results_file}{data.table, with datapaths and names of results file(s)
or character with file path of results file(s)}

\item{results_filename}{charvect with file names}

\item{headers}{list of headers identified in the result files}
}
\value{
data.table with results data
}
\description{
Read in results files
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MAE_to_PSet.R
\name{.extract_or_create_assay}
\alias{.extract_or_create_assay}
\title{Extracts an assay from a SummarizedExperiment object or creates a new one if it does not exist}
\usage{
.extract_or_create_assay(SE, assay_name)
}
\arguments{
\item{SE}{A SummarizedExperiment object}

\item{assay_name}{A character string specifying the name of the assay to extract or create}
}
\value{
A SummarizedExperiment object with the specified assay
}
\description{
This function takes a SummarizedExperiment object and an assay name as input.
If the specified assay already exists in the SummarizedExperiment object, it is returned.
Otherwise, a new assay with the specified name is created and added to the SummarizedExperiment object.
The new assay is initialized with NA values.
This is useful for when multiple Summarized Experiments in a given MAE do not have the same assays.
And it is necessary to have the same assays in all Summarized Experiments in order to convert the MAE to a PSet.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{.standardize_untreated_values}
\alias{.standardize_untreated_values}
\title{Standardize untreated values to ignore cases}
\usage{
.standardize_untreated_values(df)
}
\value{
data.table with standardized untreated values
}
\description{
Standardize untreated values to ignore cases
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exceptions.R
\name{get_exception_data}
\alias{get_exception_data}
\title{get exception data}
\usage{
get_exception_data(status_code = NULL)
}
\arguments{
\item{status_code}{A numeric value}
}
\value{
A data.table row with exception data or all exceptions
}
\description{
get exception data
}
\examples{
get_exception_data(1)
get_exception_data()
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gdr_test_data_class.R
\docType{methods}
\name{template_path}
\alias{template_path}
\alias{template_path,gdr_test_data-method}
\title{Method template_path}
\usage{
template_path(x)

\S4method{template_path}{gdr_test_data}(x)
}
\arguments{
\item{x}{object class gdr_test_data}
}
\value{
value of slot \code{template_path}
}
\description{
Method for object gdr_test_data - access to slot \code{template_path}
}
\examples{
 td <- get_test_data()
 template_file_path <- template_path(td)

}
\keyword{methods}
\keyword{test_data_class}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{get_excel_sheet_names}
\alias{get_excel_sheet_names}
\title{get Excel sheets names for a charvec of files
for non-Excel files return 0}
\usage{
get_excel_sheet_names(fls)
}
\arguments{
\item{fls}{charvec with file pathsa}
}
\value{
list with one element per file with
sheet names or 0 (for non-Excel file)
}
\description{
get Excel sheets names for a charvec of files
for non-Excel files return 0
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_results_Tecan}
\alias{load_results_Tecan}
\title{Load tecan results from xlsx}
\usage{
load_results_Tecan(results_file, headers = gDRutils::get_env_identifiers())
}
\arguments{
\item{results_file}{string, file path to a result file}

\item{headers}{list of headers identified in the manifest}
}
\value{
data.table derived from Tecan data
}
\description{
This functions loads and checks the results file
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_results_EnVision_new}
\alias{load_results_EnVision_new}
\title{Load results from EnVision_new (CSV and XLSX)}
\usage{
load_results_EnVision_new(
  results_file,
  headers = gDRutils::get_env_identifiers()
)
}
\arguments{
\item{results_file}{character, file path(s) to result file(s)}

\item{headers}{list of headers identified in the manifest}
}
\value{
data.table with results data
}
\description{
This functions loads and checks the results file(s) from a new Envision instrument
in the CSV or XLSX format. Supports multiple plates in a single file or multiple
sheets in an Excel file by robustly checking the file structure.
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correction.R
\name{.check_against_single_template_sheet}
\alias{.check_against_single_template_sheet}
\title{Evaluate if template file with single sheet is present,
if the name of the sheet is correct and if it can be fixed}
\usage{
.check_against_single_template_sheet(ts)
}
\arguments{
\item{ts}{list with template sheets info}
}
\value{
logical flag
}
\description{
get sheets for given set of XLS files
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prism_to_gdrDF.R
\name{convert_LEVEL6_prism_to_gDR_input}
\alias{convert_LEVEL6_prism_to_gDR_input}
\title{Load, convert and process the level 6 PRISM data into a gDR input}
\usage{
convert_LEVEL6_prism_to_gDR_input(
  prism_data_path,
  cell_line_data_path,
  treatment_data_path,
  meta_data_path,
  readout_min = 1.03
)
}
\arguments{
\item{prism_data_path}{path to PRISM LEVEL6 csv file with collapsed
log fold change data}

\item{cell_line_data_path}{path to cell line info data}

\item{treatment_data_path}{path to collapsed treatment info data}

\item{meta_data_path}{path to metadata file describing all cancer models/cell lines
which are referenced by a dataset contained within the DepMap portal}

\item{readout_min}{minimum ReadoutValue}
}
\value{
\code{data.table} object with input data for gDR pipeline
}
\description{
Load, convert and process the level 6 PRISM data into a gDR input
}
\examples{
 prism_data_path <- system.file("testdata/prism_collapsed_LOGFC.csv", package = "gDRimport")
 cell_line_data_path <- system.file("testdata/prism_cell_lines.csv", package = "gDRimport")
 treatment_data_path <- system.file("testdata/prism_treatment.csv", package = "gDRimport")
 prism_meta <- system.file("testdata/prism_model.csv", package = "gDRimport")
 convert_LEVEL6_prism_to_gDR_input(prism_data_path, cell_line_data_path, treatment_data_path, prism_meta)

}
\keyword{prism_conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{check_metadata_headers}
\alias{check_metadata_headers}
\title{Check whether metadata headers are correct and make fixes if needed}
\usage{
check_metadata_headers(corrected_names, df_name)
}
\arguments{
\item{corrected_names}{a charvec with corrected colnames of df}

\item{df_name}{a name of data.table ("" by default)}
}
\value{
a charvec with corrected colnames of df
}
\description{
Check whether metadata headers are correct and make fixes if needed
}
\keyword{correction_exception}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assert_utils.R
\name{is_readable_v}
\alias{is_readable_v}
\title{is_readable_v
Check if all paths in vector are readable}
\usage{
is_readable_v(paths)
}
\arguments{
\item{paths}{a character with path(s)}
}
\value{
\code{NULL} invisibly.
}
\description{
is_readable_v
Check if all paths in vector are readable
}
\examples{
td2 <- get_test_Tecan_data()
is_readable_v(td2$r_files)

}
\keyword{utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/D300.R
\name{save_drug_info_per_well}
\alias{save_drug_info_per_well}
\title{for each drug create a Gnumber and Concentration information for each well}
\usage{
save_drug_info_per_well(trt_info, trt_gnumber_conc, wb, idfs)
}
\arguments{
\item{trt_info}{list with treatment info}

\item{trt_gnumber_conc}{list with treatment data}

\item{wb}{pointer to xlsx workbook}

\item{idfs}{charvec with identifiers}
}
\value{
\code{NULL} invisibly.
}
\description{
for each drug create a Gnumber and Concentration information for each well
}
\keyword{D300}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{load_templates}
\alias{load_templates}
\title{Load templates}
\usage{
load_templates(df_template_files)
}
\arguments{
\item{df_template_files}{data.table, with datapaths and names of results file(s)
or character with file path of templates file(s)}
}
\value{
data.table with templates data
}
\description{
This functions loads and checks the template file(s)
}
\examples{
 td <- get_test_data()
 t_df <- load_templates(template_path(td))

}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{.check_file_structure}
\alias{.check_file_structure}
\title{Check the structure of raw data}
\usage{
.check_file_structure(
  df,
  filename,
  sheet_name,
  readout_offset,
  n_row,
  n_col,
  bcode_idx,
  bcode_col
)
}
\value{
\code{NULL} invisibly.
}
\description{
Check the structure of raw data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_files.R
\name{read_in_template_sheet_xlsx}
\alias{read_in_template_sheet_xlsx}
\title{Read in data from xlsx template sheet}
\usage{
read_in_template_sheet_xlsx(template_file, template_sheets, idx, plate_info)
}
\arguments{
\item{template_file}{character, file path(s) to template(s)}

\item{template_sheets}{template sheet names}

\item{idx}{template file index}

\item{plate_info}{list with plate info}
}
\value{
data.table with template data
}
\description{
Read in data from xlsx template sheet
}
\keyword{load_files}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gDRimport-package.R
\docType{package}
\name{gDRimport-package}
\alias{gDRimport}
\alias{gDRimport-package}
\title{gDRimport: Package for handling the import of dose-response data}
\value{
package help page
}
\description{
The package is a part of the gDR suite. It helps to prepare raw drug response data for downstream processing. It mainly contains helper functions for importing/loading/validating dose-response data provided in different file formats.
}
\note{
To learn more about functions start with \code{help(package = "gDRimport")}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/gdrplatform/gDRimport}
  \item \url{https://gdrplatform.github.io/gDRimport/}
  \item Report bugs at \url{https://github.com/gdrplatform/gDRimport/issues}
}

}
\author{
\strong{Maintainer}: Arkadiusz Gladki \email{gladki.arkadiusz@gmail.com} (\href{https://orcid.org/0000-0002-7059-6378}{ORCID})

Authors:
\itemize{
  \item Bartosz Czech (\href{https://orcid.org/0000-0002-9908-3007}{ORCID})
  \item Marc Hafner (\href{https://orcid.org/0000-0003-1337-7598}{ORCID})
  \item Sergiu Mocanu
  \item Dariusz Scigocki
  \item Allison Vuong
  \item Luca Gerosa (\href{https://orcid.org/0000-0001-6805-9410}{ORCID})
  \item Janina Smola
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gdr_test_data_class.R
\docType{class}
\name{gdr_test_data-class}
\alias{gdr_test_data-class}
\title{gDR Test Data object}
\value{
object class \code{gdr_test_data} with primary test data
}
\description{
Object class \code{gdr_test_data} is build by function \code{\link[=get_test_data]{get_test_data()}}
}
\section{Slots}{

\describe{
\item{\code{manifest_path}}{character, path to manifest file}

\item{\code{result_path}}{character, path(s) to results file}

\item{\code{template_path}}{character, path(s) to data.table with template data}

\item{\code{ref_m_df}}{character, data.table with manifest data}

\item{\code{ref_r1_r2}}{character, path to reference file with raw data for treated & untreated}

\item{\code{ref_r1}}{character, path to reference file with raw data for treated}

\item{\code{ref_t1_t2}}{character, path to reference template file with treated & untreated data}

\item{\code{ref_t1}}{character, path to reference template file with treated data}
}}

\keyword{classes}
\keyword{test_data_class}
